************************************************************************
*
*     * LMBMBS includes the following subroutines *
*
*     S   ACTVAR          Finding the index set of free and active
*                           variables at the generalized Cauchy point.
*     S   AGBFGS          Simplified subgradient aggregation with L-BFGS
*                           formula.
*     S   AGGSR1          Subgradient aggregation.
*     S   CP1ST           Computing the generalized Cauchy point and the
*                           search direction at the first iteration.
*     S   CPBFGS          Computing the generalized Cauchy point by the
*                            limited memory BFGS update formula.
*     S   CPSR1           Computing the generalized Cauchy point by the
*                           limited memory SR1 update formula.
*     S   DESTEP          Stepsize determination for descent steps.
*     S   DLBFGS          Updating the matrices and computing the search
*                           direction by the limited memory BFGS update.
*     S   DLSR1           Updating the matrices and computing the search
*                           direction by the limited memory SR1 update.    
*     S   DOBUN           Bundle selection.
*     S   FRMJ            Formation of the JJ' Cholesky factorization of
*                           matrix J = SM' SM / GAMMA + LC^(-1)L'.
*     S   FRMLEL          Formation of the LEL' factorization of the
*                           indefinite (BFGS) matrix.
*     S   GETIME          Execution time.
*     S   INDIC1          Initialization of indices.
*     S   LLS2            Line search using function values and
*                           subgradients.
*     S   NULSTP          Stepsize determination for null steps.
*     S   PRDBMV          Computation of the product of the 2mc x 2mc
*                           middle matrix in the compact L-BFGS formula
*                           and a 2mc vector.
*     S   PRDKMV          Computation of the product of the inverse of
*                           2mc x 2mc matrix and and a 2mc vector.
*     S   PROJGR          Projection of the subgradient and calculation
*                           of Eucleidean norm.
*     S   PROJX           Projection of the initial X to the feasible
*                           region if necessary.
*     S   QINT            Quadratic interpolation for line search
*                           with one directional derivative.
*     S   RESTAR          Initialization.
*     S   RPRINT          Printout the results.
*     RF  SCLPAR          Calculation of the scaling parameter.
*     S   SSBFGS          Subspace minimization and direction finding
*                           using limited memory BFGS matrices.
*     S   SSSR1           Subspace minimization and direction finding
*                           using limited memory SR1 matrices.
*     S   TINIT           Calculation of the initial step size.
*     S   WPRINT          Printout the error and warning messages.
*
*
*     Napsu Karmitsa (2002 -2007, Last modified 2009)      
*
* 
************************************************************************
*
*     * SUBROUTINE CP1ST *
*
*      
*     * Purpose *
*      
*     Computation of generalized Cauchy point and the search direction 
*     at the first iteration.
*
*      
*     * Calling sequence *
*     
*     CALL CP1ST(N,X,XCP,XL,XU,IB,IT,G,DC,T,SMALL)
*
*      
*     * Parameters *
*
*     II  N               Number of variables.
*     RI  X(N)            Vector of variables. 
*     RO  XCP(N)          Generalized Cauchy point.
*     RI  XL(N)           Lower bounds for variables. 
*     RI  XU(N)           Upper bounds for variables. 
*     II  IB(N)           Type of bound constraints:
*                             0  - X(I) is unbounded,
*                             1  - X(I) has only a lower bound,
*                             2  - X(I) has both lower and upper bounds,
*                             3  - X(I) has only an upper bound. 
*     RI  G(N)            Current subgradient of the objective function.
*     RO  DC(N)           Cauchy direction P(X-T*G)-X.
*     RA  T(N)            Auxiliary array used to store the breakpoints.
*     IA  IT(N)           Index set of breakpoints:
*                             for I=1,...,NLEFT, IT(I) are the indices
*                                 of breakpoints that have not been
*                                 encountered,
*                             for i=NLEFT+1,...,NBREAK, IT(I) are the
*                                 indices of encountered breakpoints and
*                             for i=IFREE,...,N, IT(I) are the indices
*                                 of variables that have no bound
*                                 constraints along the search direction.
*     RI  SMALL           The smallest positive number such that
*                           1.0 + SMALL > 1.0.
*     
*
*     * Subprograms used *
*      
*     S   HPSRT           Heapsort algorithm.
*     S   SCSUM           Sum of a vector and the scaled vector.
*     S   XDIFFY          Difference of two vectors.
*
*      
      
      SUBROUTINE CP1ST(N,X,XCP,XL,XU,IB,IT,G,DC,T,SMALL)
      

*     Scalar Arguments
      INTEGER N
      DOUBLE PRECISION SMALL
      
*     Array Arguments
      INTEGER IB(*),IT(*)
      DOUBLE PRECISION G(*),X(*),XCP(*),XL(*),XU(*),DC(*),T(*)
      
*     Local Scalars
      INTEGER I,NBREAK,IFREE,ITMIN,NINT,NLEFT,ISBP
      INTEGER J,IBOUND
      DOUBLE PRECISION F1,F2,TMIN,DLTMIN,TOLD,DELTAT,GISBP,GISBP2,
     &     ZISBP
      DOUBLE PRECISION DFREE,SMALL2
      DOUBLE PRECISION ALPHA

*     Intrinsic Functions
      INTRINSIC MAX,ABS,SQRT
      INTRINSIC MIN

*     External Subroutines
      EXTERNAL HPSRT,SCSUM,XDIFFY



*
*     Initialization
*

      NBREAK = 0
      NLEFT  = 0
      IFREE  = N+1
      F1     = 0.0D+00
      F2     = 0.0D+00
      TMIN   = 0.0D+00
      ITMIN  = 0
      IBOUND = 0   

      SMALL2 = SMALL
c      SMALL2 = SQRT(SMALL)


*
*     Computation of the Cauchy direction DC, the first derivative
*     F1= -(DC)'*DC, and the breakpoints T in each coordinate
*     direction. Identification of the smallest breakpoint TMIN. 
*      

      DO 10 I=1,N

         IF (IB(I) .NE. 0) THEN
            IF (IB(I).LE.2 .AND. ABS(G(I)) .LE. SMALL2 .AND. 
     &           X(I)-XL(I) .LE. SMALL2) THEN
               DC(I) = 0.0D+00
               XCP(I) = X(I)
               GOTO 10
            END IF

            IF (IB(I).GE.2 .AND. ABS(G(I)) .LE. SMALL2 .AND. 
     &           XU(I)-X(I) .LE. SMALL2) THEN
               DC(I) = 0.0D+00
               XCP(I) = X(I)
               GOTO 10
            END IF
         END IF

         IF (IB(I).NE.0 .AND. IB(I).LE.2 .AND. G(I) .GT. 0.0D+00) THEN


*
*     Breakpoint in lower bound
*

            T(NBREAK+1) = (X(I) - XL(I))/G(I)
            
            IF (T(NBREAK+1) .LE. SMALL2) THEN
               DC(I) = 0.0D+00

            ELSE
               NBREAK = NBREAK + 1
               IT(NBREAK) = I
               DC(I) = -G(I)
               F1 = F1 - DC(I)*DC(I)
            

*     
*     Determination of the smallest breakpoint
*     
          
               IF (NBREAK .EQ. 1 .OR. T(NBREAK) .LT. TMIN) THEN
                  TMIN=T(NBREAK)
                  ITMIN=NBREAK
               END IF
            END IF
      

         ELSE IF (IB(I) .GE. 2 .AND. G(I) .LT. 0.0D+00) THEN


*
*     Breakpoint in upper bound
*               
 
            T(NBREAK+1) = (X(I) - XU(I))/G(I)
            
            IF (T(NBREAK+1) .LE. SMALL2) THEN

               DC(I) = 0.0D+00

            ELSE
               NBREAK = NBREAK + 1
               IT(NBREAK) = I
               DC(I) = -G(I)
               F1 = F1 - DC(I)*DC(I)


*     
*     Determination of the smallest breakpoint
*               

               IF (NBREAK .EQ. 1 .OR. T(NBREAK) .LT. TMIN) THEN
                  TMIN=T(NBREAK)
                  ITMIN=NBREAK
               END IF
            END IF
            
         ELSE


*
*     No breakpoint 
*

            IFREE = IFREE - 1
            IT(IFREE) = I
            DC(I) = -G(I)
            F1 = F1 - DC(I)*DC(I)

            IF(ABS(G(I)) .LE. 0.0D+00) IBOUND=-1

         END IF


*     
*     Initialization of Cauchy point
*

         XCP(I) = X(I)
           
 10   CONTINUE


*
*     The indices corresponding to free variables are located in
*     IT(1),...,IT(NBREAK) and IT(IFREE),...,IT(N). The smallest of the
*     NBREAK breakpoints is in T(ITMIN)=TMIN.
*

*      
*     DC is a zero vector. Return with the initial XCP.
*

      NLEFT = NBREAK


      IF (NBREAK .EQ. 0 .AND. IBOUND .EQ. 0) GOTO 999


*
*     Calculation of derivative F2 = DC'*DC, 
*

      F2 =  -F1
      DLTMIN = 1.0D+00

      TOLD = 0.0D+00
      NINT = 1
      

*
*     There is no breakpoints, locate the GCP and return. 
*

      IF (NBREAK .EQ. 0) GOTO 888


*
*     Begining of the loop.
*      

 100  CONTINUE


      IF (NINT .EQ. 1) THEN


*     the smallest breakpoint is TMIN=T(ITMIN) and its index is
*     in IT(ITMIN)

         ISBP = IT(ITMIN)

      ELSE
   

*     the smallest breakpoint is chosen by heapsort algorithm

         IF (NINT .EQ. 2) THEN


*     Remove the already used smallest breakpoint before heapsort call.
            
            IF (ITMIN .NE. NBREAK) THEN
               T(ITMIN) = T(NBREAK)
               IT(ITMIN) = IT(NBREAK)
            END IF


*     Heapsort with initialization of the heap.

            CALL HPSRT(NLEFT,T,IT,0)

         ELSE


*     Heapsort with updating the heap.

            CALL HPSRT(NLEFT,T,IT,1)

         END IF


         TMIN = T(NLEFT)
         ISBP = IT(NLEFT)

      END IF


      DELTAT = TMIN - TOLD
      
      IF (DELTAT .LT. 0.0D+00) THEN
         PRINT*,' no nyt voit jo hirtt ihtes.'
      END IF


*
*     Minimizer is within this interval, locate the GCP and return.
*

      IF (DLTMIN .LT. DELTAT) GOTO 888


*      
*     Examination of subsequent seqment
*

      NLEFT = NLEFT - 1
      NINT = NINT + 1
      GISBP = G(ISBP)
      DC(ISBP) = 0.0D+00

      IF (GISBP .LT. 0.0D+00) THEN
         XCP(ISBP) = XU(ISBP)
         ZISBP = XU(ISBP) - X(ISBP)
      ELSE 
         XCP(ISBP) = XL(ISBP)
         ZISBP = XL(ISBP) - X(ISBP)
      END IF


*         
*     All  variables are fixed, return with current XCP as GCP.
*         

      IF (NLEFT .EQ. 0 .AND. NBREAK .EQ. N) GOTO 999


*      
*     Update F1 and F2
*      

      GISBP2 = GISBP*GISBP

      F1 = F1 + DELTAT*F2 + GISBP2 + GISBP*ZISBP
      F2 = F2 - GISBP2


*
*     DC is a zero vector. Return with the current XCP.
*

      IF (F2 .LE. SMALL2) GOTO 999

      DLTMIN = -F1/F2


*
*     Repeat the loop for unsearched intervals.          
*

      TOLD = TMIN

      IF (NLEFT .GT. 0) GOTO 100
      IF (IBOUND .EQ. 0) DLTMIN = 0.0D+00   


*     
*     End of the loop.
*


 888  CONTINUE

      DLTMIN = MAX(0.0D+00,DLTMIN)
      TOLD = TOLD + DLTMIN


*      
*     Update free variables 
* 

      DO 20 I=1,NLEFT
         XCP(IT(I))=X(IT(I))+TOLD*DC(IT(I))
 20   CONTINUE

      DO 30 I=IFREE,N
         XCP(IT(I))=X(IT(I))+TOLD*DC(IT(I))
 30   CONTINUE

     
*
*     Calculate the search direction for the first iteration.
*

 999  CONTINUE

  
*     The indices corresponding to free variables are located in
*     IT(1),...,IT(NLEFT) and IT(IFREE),...,IT(N). The 
*     considered breakpoints are in IT(NLEFT+1),...,IT(NBREAK).
*     Otherwise, DC(I) = 0.

      DO 40 I=NLEFT+1,NBREAK 
         DC(IT(I))=XCP(IT(I))-X(IT(I))
 40   CONTINUE


*     
*     Backtrack to the feasible region if X + DC violates bounds
*     (among free variables).
*

      ALPHA = 1.0D+00
      DO 50 I=1,NLEFT
         J=IT(I)
         DFREE=DC(J)
         IF (IB(J) .NE. 0) THEN
            IF (IB(J) .LE. 2 .AND. X(J) + DFREE .LT. XL(J)) THEN

*     Otherwise the lower bound is never met
c               IF (X(J)-XCP(J)+DFREE .LT. 0.0D+00) THEN
               IF (ABS(X(J)-XCP(J)+DFREE) .GT. SMALL) THEN
                  ALPHA=MIN(ALPHA,(XL(J)-XCP(J))/(X(J)-XCP(J)+DFREE))
               END IF
 
            ELSE IF (IB(J) .GE. 2 .AND. X(J) + DFREE  .GT. XU(J)) THEN

*     Otherwise the upper bound is never met
               IF (ABS(X(J)-XCP(J)+DFREE) .GT. SMALL) THEN
                  ALPHA=MIN(ALPHA,(XU(J)-XCP(J))/(X(J)-XCP(J)+DFREE))
               END IF

            END IF
         END IF
 50   CONTINUE

      IF (ALPHA .LT. 1.0D+00) THEN

         DO 70 I=1,NLEFT
            J=IT(I)
            DC(J) = XCP(J)-X(J)+ALPHA*(X(J)-XCP(J)+DC(J))
 70      CONTINUE

         DO 80 I=IFREE,N
            J=IT(I)
            DC(J) = XCP(J)-X(J)+ALPHA*(X(J)-XCP(J)+DC(J))
 80      CONTINUE

      END IF

      RETURN
      END




************************************************************************
*
*     * SUBROUTINE DLBFGS *
*
*      
*     * Purpose *
*      
*     Matrix updating and computation of the search direction by the
*     limited memory BFGS update.
*
*      
*     * Calling sequence *
*     
*     CALL DLBFGS(N,NACT,IACT,ITYPE,IB,MC,MCC,INEW,IBFGS,IFLAG,X,XCP,
*    &     XL,XU,D,G,GP,S,U,SM,UM,RM,LM,UMTUM,SMTSM,CDIAG,SMTGP,UMTGP,
*    &     KM,GAMMA,TMPMC1,TMPMC2,TMPMC3,TMPMC4,TMPMC5,TMPMC6,TMPMC7,
*    &     TMPMC8,TMPMAT,TMPN1,SMALL,ISCALE,IERR)
*
*      
*     * Parameters *
*
*     II  N               Number of variables.
*     IO  NACT            Number of active variables.
*     II  MC              Declared number of stored corrections.
*     IU  MCC             Current number of Depositories used.
*     IU  INEW            Index for circular arrays.
*     IO  IBFGS           Index of the type of BFGS update:
*                             1  - BFGS update: the corrections are
*                                    stored.
*                             2  - BFGS update: the corrections are not
*                                    stored.
*                             3  - BFGS update is skipped.
*     IU  IFLAG           Index for adaptive version:
*                             0  - Maximum number of stored corrections
*                                    has not been changed at previous
*                                    iteration.
*                             1  - Maximum number of stored corrections
*                                    has been changed at previous
*                                    iteration.
*     IO  IACT(N)         Index set of active and free variables:
*                             for I=1,...,NACT, IACT(I) are the indices
*                                 of active variables (-J, if
*                                 XCP(J)=XL(J) and +J, otherwise)
*                             for i=NACT+1,...,N, IACT(I) are the indices
*                                 of free variables.
*     II  ITYPE           Type of problem:
*                             0  - problem is unbounded,
*                             1  - constrained problem.
*     II  IB(N)           Type of bound constraints:
*                             0  - X(I) is unbounded,
*                             1  - X(I) has only a lower bound,
*                             2  - X(I) has both lower and upper bounds,
*                             3  - X(I) has only an upper bound. 
*     RI  X(N)            Vector of variables.
*     RO  XCP(N)          Generalized Cauchy point.
*     RI  XL(N)           Lower bounds for variables. 
*     RI  XU(N)           Upper bounds for variables. 
*     RU  D(N)            Search direction.
*     RI  G(N)            Current subgradient of the objective
*                           function.
*     RI  GP(N)           Previous subgradient of the objective
*                           function.
*     RI  S(N)            Difference of current and previous variables.
*     RI  U(N)            Difference of current and previous
*                           subgradients.
*     RU  SM(N*(MCC))     Matrix whose columns are stored corrections.
*     RU  UM(N*(MCC))     Matrix whose columns are stored subgradient
*                           differences.
*     RU  LM((MCC)*(MCC+1)/2) Lower triangular matrix stored rowwise
*                           in the one-dimensional array.
*     RU  RM((MCC)*(MCC+1)/2) Upper triangular matrix stored columnwise
*                              in the one-dimensional array.
*     RU  SMTSM((MCC)*(MCC+1)/2)  Matrix SMTSM = SM'*SM.
*     RU  UMTUM((MCC)*(MCC+1)/2)  Matrix UMTUM = UM'*UM.
*     RU  CDIAG(MCC)      Diagonal matrix.
*     RU  SMTGP(MCC)      Vector SMTGP = SM'*GP.
*     RU  UMTGP(MCC)      Vector UMTGP = UM'*GP.
*     RA  KM((2MCC+1)*(MCC)) Auxiliary 2*MCC matrix.
*     RU  GAMMA           Scaling parameter.
*     RA  TMPMAT((MCC)*(MCC+1)/2) Auxiliary MCC matrix.
*     RA  TMPMC#(MCC)     Auxiliary arrays; # = 1,...,8.
*     RA  TMPN1(N)        Auxiliary array:
*                           On input: Previous aggregate subgradient.
*     RI  SMALL           Small positive value.
*     II  ISCALE          Selection of the scaling:
*                             0  - Scaling at every iteration
*                                  with STU/UTU.
*                             1  - Scaling at every iteration
*                                  with STS/STU.
*                             2  - Interval scaling with STU/UTU.
*                             3  - Interval scaling with STS/STU.
*                             4  - Preliminary scaling with STU/UTU.
*                             5  - Preliminary scaling with STS/STU.
*                             6  - No scaling.      
*     IO  IERR            Error indicador: 
*                             0  - Everything is ok.
*                            -1  - Error in CHOFA.
*                            -2  - Error in LINEQ.
*                            -3  - Error in TRLIEQ.
*                            -4  - Error in FRMLEL.
*                            -5  - Error in PRDKMV.
*                            -8  - Error in CPBFGS.
*                           -10  - Warning: MCC=0 and IBFGS=3.
*                           -11  - Warning: MCC=0 and ISR1=3.
*     
*     
*     * Local variables *
*
*     I   MN              Current number of stored corrections.
*     I   IOLD            Index of the oldest correction.
*     R   STU             STU = S'*U. 
*     R   STS             STS = S'*S. 
*     I   IFLAG2          Index for adaptive version.
*
*
*     * Subprograms used *
*      
*     RF  VDOT            Dot product of two vectors.
*     RF  SCLPAR          Calculation of the scaling parameter.
*      
*     S   ACTVAR          Finding the index set of free and active
*                           variables  at the generalized Cauchy point.
*     S   COPY            Copying of vector.
*     S   COPY2           Copying of two vectors.
*     S   COPY4           Copying of four vectors.
*     S   CPBFGS          Computing the generalized Cauchy point by the
*                           limited memory BFGS update formula.
*     S   INDIC1          Initialization of indices.
*     S   RWAXV2          Multiplication of two rowwise stored dense 
*                           rectangular matrices A and B by vectors X 
*                           and Y.
*     S   XDIFFY          Difference of two vectors.
*     S   VNEG            Copying of a vector with change of the sign
*     S   SSBFGS          Subspace minimization and direction finding
*                           using limited memory BFGS matrices.
*     
*
*     The variable and subgradient differences and all the MCC vectors are 
*     stored in a circular order controlled by the parameter point INEW.
*
*      
      
      SUBROUTINE DLBFGS(N,NACT,IACT,ITYPE,IB,MC,MCC,INEW,IBFGS,IFLAG,
     &     X,XCP,XL,XU,D,G,GP,S,U,SM,UM,RM,LM,UMTUM,SMTSM,CDIAG,
     &     SMTGP,UMTGP,KM,GAMMA,TMPMC1,TMPMC2,TMPMC3,TMPMC4,TMPMC5,
     &     TMPMC6,TMPMC7,TMPMC8,TMPMAT,TMPN1,SMALL,ISCALE,IERR)

      
*     Scalar Arguments
      INTEGER N,NACT,MC,MCC,INEW,IBFGS,IFLAG,ISCALE,ITYPE,IERR
      DOUBLE PRECISION GAMMA,SMALL

*     Array Arguments
      INTEGER IACT(*),IB(*)
      DOUBLE PRECISION X(*),XCP(*),XL(*),XU(*),D(*),G(*),GP(*),S(*),
     &     U(*),SM(*),UM(*),RM(*),LM(*),UMTUM(*),SMTSM(*),CDIAG(*),
     &     SMTGP(*),UMTGP(*),KM(*),TMPMC1(*),TMPMC2(*),TMPMC3(*),
     &     TMPMC4(*),TMPMC5(*),TMPMC6(*),TMPMC7(*),TMPMC8(*),TMPMAT(*),
     &     TMPN1(*)

*     Local Scalars
      INTEGER I,J,K,MN,IOLD,IFLAG2
      DOUBLE PRECISION STU,STS

*     External Functions
      DOUBLE PRECISION VDOT,SCLPAR
      EXTERNAL VDOT,SCLPAR

*     Intrinsic Functions
      INTRINSIC SQRT

*     External Subroutines
      EXTERNAL ACTVAR,COPY,COPY2,COPY4,CPBFGS,INDIC1,XDIFFY,VNEG,
     &     RWAXV2,SSBFGS



*
*     Initialization
*

      IERR = 0
      IBFGS = 0
      IFLAG2 = 0

      STU = VDOT(N,S,U)
      STS = VDOT(N,S,S)

      
*
*     Positive definiteness
*

      IF (STU .GT. SMALL) THEN
         
         IF (-VDOT(N,D,U)-VDOT(N,TMPN1,S) .LT. -SMALL) THEN

            
*     
*     Update matrices
*         

            IBFGS = 1


*
*     Initialization of indices.
*            

            CALL INDIC1(MC,MCC,MN,INEW,IOLD,IFLAG,IFLAG2,IBFGS)

            
*     
*     Update SM and UM
*

            CALL COPY2(N,S,SM((INEW-1)*N+1),U,UM((INEW-1)*N+1))

            
*     
*     Computation of SM'*G, UM'*G, SM'*S, and UM'*S
*

            IF (INEW .GE. MN) THEN
               CALL RWAXV2(N,MN,SM((INEW-MN)*N+1),
     &              UM((INEW-MN)*N+1),G,G,TMPMC1(IOLD),TMPMC2(IOLD))
               CALL RWAXV2(N,MN,SM((INEW-MN)*N+1),
     &              UM((INEW-MN)*N+1),S,S,TMPMC5(IOLD),TMPMC6(IOLD))

            ELSE
               CALL RWAXV2(N,INEW,SM,UM,G,G,TMPMC1,TMPMC2)
               CALL RWAXV2(N,INEW,SM,UM,S,S,TMPMC5,TMPMC6)
               CALL RWAXV2(N,MN-INEW,SM((IOLD-1)*N+1),
     &              UM((IOLD-1)*N+1),G,G,TMPMC1(IOLD),TMPMC2(IOLD))
               CALL RWAXV2(N,MN-INEW,SM((IOLD-1)*N+1),
     &              UM((IOLD-1)*N+1),S,S,TMPMC5(IOLD),TMPMC6(IOLD))
            END IF               
            
            
*         
*     Computation of SM'*U and UM'*U
*

            IF (INEW .GE. MN) THEN
               DO 30 I=IOLD,INEW-1
                  TMPMC3(I)=TMPMC1(I) - SMTGP(I)
                  SMTGP(I)=TMPMC1(I)
                  TMPMC4(I)=TMPMC2(I) - UMTGP(I)
                  UMTGP(I)=TMPMC2(I)
 30            CONTINUE

            ELSE
               DO 10 I=1,INEW-1
                  TMPMC3(I)=TMPMC1(I) - SMTGP(I)
                  SMTGP(I)=TMPMC1(I)
                  TMPMC4(I)=TMPMC2(I) - UMTGP(I)
                  UMTGP(I)=TMPMC2(I)
 10            CONTINUE

               DO 20 I=IOLD,MN+1
                  TMPMC3(I)=TMPMC1(I) - SMTGP(I)
                  SMTGP(I)=TMPMC1(I)
                  TMPMC4(I)=TMPMC2(I) - UMTGP(I)
                  UMTGP(I)=TMPMC2(I)
 20            CONTINUE
            END IF
            
            TMPMC3(INEW)=TMPMC1(INEW) - VDOT(N,S,GP)
            SMTGP(INEW)=TMPMC1(INEW)
            TMPMC4(INEW)=TMPMC2(INEW) - VDOT(N,U,GP)
            UMTGP(INEW)=TMPMC2(INEW)

            
*         
*     Update RM, LM, SMTSM, and UMTUM.
*

            IF (MCC .GE. MC .AND. IFLAG2 .NE. 1) THEN

*     Shift the old values.

               DO 40 I=1,MN-1
                  J=(I-1)*I/2+1
                  K=I*(I+1)/2+2
                  CALL COPY4(I,RM(K),RM(J),LM(K),LM(J),SMTSM(K),
     &                 SMTSM(J),UMTUM(K),UMTUM(J))
 40            CONTINUE
            END IF

*     Add the new values.

            IF (INEW .GE. MN) THEN
               CALL COPY4(MN,TMPMC3(IOLD),RM((MN-1)*MN/2+1),
     &              TMPMC6(IOLD),LM((MN-1)*MN/2+1),
     &              TMPMC5(IOLD),SMTSM((MN-1)*MN/2+1),
     &              TMPMC4(IOLD),UMTUM((MN-1)*MN/2+1))

            ELSE
               CALL COPY4(MN-INEW,TMPMC3(IOLD)
     &              ,RM((MN-1)*MN/2+1)
     &              ,TMPMC6(IOLD),LM((MN-1)*MN/2+1)
     &              ,TMPMC5(IOLD),SMTSM((MN-1)*MN/2+1)
     &              ,TMPMC4(IOLD),UMTUM((MN-1)*MN/2+1))
               
               CALL COPY4(INEW,TMPMC3,
     &              RM((MN-1)*MN/2+MN-INEW+1),
     &              TMPMC6,LM((MN-1)*MN/2+MN-INEW+1),
     &              TMPMC5,SMTSM((MN-1)*MN/2+MN-INEW+1),
     &              TMPMC4,UMTUM((MN-1)*MN/2+MN-INEW+1))
            END IF

            LM((MN+1)*MN/2) = 0.0D+00


*
*     Update CDIAG
*

            CDIAG(INEW) = STU
            
            
*         
*     Computation of GAMMA
*

            GAMMA = SCLPAR(MCC,ISCALE,STS,STU,TMPMC4(INEW),SMALL)

         
            INEW = INEW + 1
            IF (INEW .GT. MC + 1) INEW = 1
            IF (IFLAG .EQ. 0 .AND. MCC .LT. MC + 1) MCC = MCC + 1

         ELSE

            
*     
*     BFGS update, corrections are not saved.
*     

            IBFGS = 2


*
*     Initialization of indices.
*            

            CALL INDIC1(MC,MCC,MN,INEW,IOLD,IFLAG,IFLAG2,IBFGS)


*     
*     Update SM and UM.
*     

            CALL COPY2(N,S,SM((INEW-1)*N+1),U,UM((INEW-1)*N+1))

            
*     
*     Computation of SM'*G, UM'*G, SM'*S, and UM'*S.
*

            CALL RWAXV2(N,MN,SM,UM,G,G,TMPMC1,TMPMC2)
            CALL RWAXV2(N,MN,SM,UM,S,S,TMPMC5,TMPMC6)

            
*         
*     Computation of SM'*U and UM'*U.
*

            IF (IOLD .NE. 1) THEN
               DO 50 I=1,INEW-1
                  TMPMC3(I)=TMPMC1(I) - SMTGP(I)
                  SMTGP(I)=TMPMC1(I)
                  TMPMC4(I)=TMPMC2(I) - UMTGP(I)
                  UMTGP(I)=TMPMC2(I)
 50            CONTINUE

               DO 60 I=IOLD,MN
                  TMPMC3(I)=TMPMC1(I) - SMTGP(I)
                  SMTGP(I)=TMPMC1(I)
                  TMPMC4(I)=TMPMC2(I) - UMTGP(I)
                  UMTGP(I)=TMPMC2(I)
 60            CONTINUE

            ELSE
               DO 70 I=1,MN-1
                  TMPMC3(I)=TMPMC1(I) - SMTGP(I)
                  SMTGP(I)=TMPMC1(I)
                  TMPMC4(I)=TMPMC2(I) - UMTGP(I)
                  UMTGP(I)=TMPMC2(I)
 70            CONTINUE
               
            END IF

            TMPMC3(INEW)=TMPMC1(INEW) - VDOT(N,S,GP)
            SMTGP(INEW)=TMPMC1(INEW)
            TMPMC4(INEW)=TMPMC2(INEW) - VDOT(N,U,GP)
            UMTGP(INEW)=TMPMC2(INEW)

            
*         
*     Fill out RM, LM, SMTSM, and UMTUM
*

            IF (IOLD .NE. 1) THEN
               
               CALL COPY4(MN-INEW,TMPMC3(IOLD)
     &              ,RM((MN-1)*MN/2+1)
     &              ,TMPMC6(IOLD),LM((MN-1)*MN/2+1)
     &              ,TMPMC5(IOLD),SMTSM((MN-1)*MN/2+1)
     &              ,TMPMC4(IOLD),UMTUM((MN-1)*MN/2+1))

               CALL COPY4(INEW,TMPMC3,RM((MN-1)*MN/2+MN-INEW+1)
     &              ,TMPMC6,LM((MN-1)*MN/2+MN-INEW+1)
     &              ,TMPMC5,SMTSM((MN-1)*MN/2+MN-INEW+1)
     &              ,TMPMC4,UMTUM((MN-1)*MN/2+MN-INEW+1))
            
            ELSE

               CALL COPY4(MN,TMPMC3,RM((MN-1)*MN/2+1)
     &              ,TMPMC6,LM((MN-1)*MN/2+1)
     &              ,TMPMC5,SMTSM((MN-1)*MN/2+1)
     &              ,TMPMC4,UMTUM((MN-1)*MN/2+1))
            END IF

            LM((MN+1)*MN/2) = 0.0D+00


*
*     Update CDIAG
*

            CDIAG(INEW) = STU
            
            
*         
*     Computation of GAMMA
*

            GAMMA = SCLPAR(MCC,ISCALE,STS,STU,TMPMC4(INEW),SMALL)

               
         END IF

      ELSE

         
*     
*     BFGS update is skipped
*     

         IBFGS = 3


         IF (MCC .EQ. 0) THEN
            IF (ITYPE .EQ. 1) THEN
               IERR = -10
            ELSE
               IFLAG = 0
               CALL VNEG(N,G,D)
            END IF
            RETURN
         END IF

         
*
*     Initialization of indices.
*            

         CALL INDIC1(MC,MCC,MN,INEW,IOLD,IFLAG,IFLAG2,IBFGS)


*         
*     Computation of GAMMA
*
         
         IF (ISCALE .GE. 4) GAMMA = 1.0D+00

         
*         
*     Computation of SM'*G and UM'*G
*
         
         IF (IOLD .EQ. 1 .OR. IOLD .EQ. 2) THEN
            CALL RWAXV2(N,MN,SM((IOLD-1)*N+1),UM((IOLD-1)*N+1),G,G,
     &           SMTGP(IOLD),UMTGP(IOLD))
            
         ELSE
            CALL RWAXV2(N,INEW-1,SM,UM,G,G,SMTGP,UMTGP)
            CALL RWAXV2(N,MN-INEW+1,SM((IOLD-1)*N+1),
     &           UM((IOLD-1)*N+1),G,G,SMTGP(IOLD),UMTGP(IOLD))
         END IF

      END IF


*
*     Computation of Cauchy point.
*

      IF (ITYPE .EQ. 1) THEN
         
         CALL CPBFGS(N,MC,MN,IOLD,IBFGS,IFLAG,X,XCP,XL,XU,IB,IACT,G,
     &        D,TMPN1,SMTGP,UMTGP,SMTSM,SM,UM,LM,CDIAG,TMPMAT,TMPMC1,
     &        TMPMC2,TMPMC3,TMPMC4,TMPMC5,TMPMC6,TMPMC7,TMPMC8,GAMMA,
     &        SMALL,IERR)
         IF (IERR .NE. 0) RETURN


*
*     Determination of the active variables.
*

         CALL ACTVAR(N,NACT,IACT,IB,XCP,XL,XU,SMALL)

      END IF


*
*     Subspace minimization and direction finding.
*

      IF (NACT .NE. N) THEN


*     
*     If there are no free variables, then skip the subspace
*     minimization D=XCP-X.
*     

         CALL SSBFGS(N,NACT,MC,INEW,IOLD,MN,IBFGS,IFLAG,X,XL,XU,XCP,
     &        G,D,SM,UM,TMPN1,RM,UMTUM,CDIAG,SMTGP,UMTGP,KM,TMPMC1,
     &        TMPMC2,TMPMC3,TMPMC4,TMPMC5,GAMMA,ITYPE,IACT,IB,IERR,
     &        SMALL)
         IF (IERR .NE. 0) RETURN

      ELSE

         CALL XDIFFY(N,XCP,X,D)
      END IF

      RETURN
      END



************************************************************************
*
*     * SUBROUTINE CPBFGS *
*
*      
*     * Purpose *
*      
*     Computation of generalized Cauchy point by the limited memory
*     BFGS update.
*
*      
*     * Calling sequence *
*     
*      CALL CPBFGS(N,MC,MN,IOLD,IBFGS,IFLAG,X,XCP,XL,XU,IB,IT,
*     &     G,DC,T,SMTGP,UMTGP,SMTSM,SM,UM,LM,CDIAG,JT,PAUX1,PAUX2,
*     &     CAUX1,CAUX2,V1,V2,W1,W2,GAMMA,SMALL,IERR)
*
*      
*     * Parameters *
*     
*     II  N               Number of variables.
*     II  MC              Declared Number of stored corrections.
*     II  MN              Current number of stored corrections.
*     II  IOLD            Index for circular arrays.
*     II  IBFGS           Index of the type of BFGS update:
*                             1  - BFGS update: the corrections are
*                                    stored.
*                             2  - BFGS update: the corrections are not
*                                    stored.
*                             3  - BFGS update is skipped.
*     II  IFLAG           Index for adaptive version:
*                             0  - Maximum number of stored corrections
*                                    has not been changed at previous
*                                    iteration.
*                             1  - Maximum number of stored corrections
*                                    has been changed at previous
*                                    iteration.
*     RI  X(N)            Vector of variables. 
*     RO  XCP(N)          Generalized Cauchy point.
*     RI  XL(N)           Lower bounds for variables. 
*     RI  XU(N)           Upper bounds for variables. 
*     II  IB(N)           Type of bound constraints:
*                             0  - X(I) is unbounded,
*                             1  - X(I) has only a lower bound,
*                             2  - X(I) has both lower and upper bounds,
*                             3  - X(I) has only an upper bound. 
*     RI  G(N)            Current subgradient of the objective function.
*     RI  SM(N*(MN+1))    Matrix whose columns are stored corrections.
*     RI  UM(N*(MN+1))    Matrix whose columns are stored subgradient
*                           differences.
*     RI  SMTGP(MN+1)     Vector SMTGP = SM'*GP.
*     RI  UMTGP(MN+1)     Vector UMTGP = UM'*GP.
*     RI  SMTSM((MN+1)*MN/2)  Matrix SMTGP = SM'*SM.
*     RI  LM((MN+1)*MN/2) Lower triangular matrix stored rowwise
*                           in the one-dimensional array.
*     RI  CDIAG(MN+1)     Diagonal matrix.
*     RA  JT((MN+1)*MN/2) The upper triangular matrix J' that is the 
*                           Cholesky factorization of
*                           1/GAMMA * SM'*SM + L*CDIAG^(-1)*L'.  
*     RI  GAMMA           Scaling parameter.
*     RA  DC(N)           Auxiliary array used to store Cauchy direction.
*     RA  T(N)            Auxiliary array used to store the breakpoints.
*     IA  IT(N)           Index set of breakpoints:
*                             for I=1,...,NLEFT, IT(I) are the indices
*                                 of breakpoints that have not been
*                                 encountered,
*                             for i=NLEFT+1,...,NBREAK, IT(I) are the
*                                 indices of encountered breakpoints and
*                             for i=IFREE,...,N, IT(I) are the indices
*                                 of variables that have no bound
*                                 constraints along the search direction.
*     RA  PAUX1(MN)       Auxiliary array used to store the vector UM'*DC.
*     RA  PAUX2(MN)       Auxiliary array used to store the vector
*                           1/GAMMA * SM'*DC.
*     RA  CAUX1(MN)       Auxiliary array used to store the vector
*                           UM'*(XCP-X).
*     RA  CAUX2(MN)       Auxiliary array used to store the vector
*                           1/GAMMA * SM'*(XCP-X).
*     RA  V1(MN)          Auxiliary array.
*     RA  V2(MN)          Auxiliary array.
*     RA  W1(MN)          Auxiliary array.
*     RA  W2(MN)          Auxiliary array.
*     RI  SMALL           Small positive value.
*     IO  IERR            Error indicador: 
*                             0  - Everything is ok.
*                            -1  - Error in CHOFA (in FRMJ).
*                            -2  - Error in LINEQ.
*                            -3  - Error in TRLIEQ (in PRDBMV).
*                            -4  - Error in FRMLEL.
*                            -5  - Error in PRDKMV.
*                            -8  - Error in CPBFGS, a nonpositive
*                                  definite matrix detected.
*                           -10  - Warning: MCC=0 and IBFGS=3.
*                           -11  - Warning: MCC=0 and ISR1=3.
*     
*
*     * Subprograms used *
*      
*     RF  VDOT            Dot product of two vectors.
*     S   FRMJ            Formation of the J*J' Cholesky factorization 
*                           of matrix JT = SM'*SM/GAMMA + LM*C^(-1)*LM'.
*     S   HPSRT           Heapsort algorithm.
*     S   PRDBMV          The product of the 2MC X 2MC middle matrix in
*                           the compact L-BFGS formula and a 2MC vector
*     S   SCSUM           Sum of a vector and the scaled vector.
*     
*
*     The variable and subgradient differences and all the input
*     (MN+1)-vectors are stored in a circular order controlled by the
*     parameter point IOLD.
*
*
      
      SUBROUTINE CPBFGS(N,MC,MN,IOLD,IBFGS,IFLAG,X,XCP,XL,XU,IB,IT,
     &     G,DC,T,SMTGP,UMTGP,SMTSM,SM,UM,LM,CDIAG,JT,PAUX1,PAUX2,CAUX1,
     &     CAUX2,V1,V2,W1,W2,GAMMA,SMALL,IERR)

      
*     Scalar Arguments
      INTEGER N,MC,MN,IOLD,IBFGS,IFLAG,IERR
      DOUBLE PRECISION GAMMA,SMALL
      
*     Array Arguments
      INTEGER IB(*),IT(*)
      DOUBLE PRECISION G(*),X(*),XCP(*),XL(*),XU(*),DC(*),T(*),SMTGP(*),
     &     UMTGP(*),SM(*),UM(*),LM(*),CDIAG(*),SMTSM(*),JT(*),
     &     PAUX1(*),PAUX2(*),CAUX1(*),CAUX2(*),V1(*),V2(*),W1(*),W2(*)

*     Local Scalars
      INTEGER I,J,L,NBREAK,IFREE,ITMIN,NINT,ISBP,NLEFT,MCC,IBOUND
      DOUBLE PRECISION F1,F2,TMIN,DLTMIN,TOLD,DELTAT,GISBP,GISBP2,ZISBP,
     &     WMC,WMP,WMW,SMALL2
      
*     External Functions
      DOUBLE PRECISION VDOT
      EXTERNAL VDOT

*     Intrinsic Functions
      INTRINSIC MAX,ABS,SQRT

*     External Subroutines
      EXTERNAL FRMJ,HPSRT,SCSUM,PRDBMV


      
*
*     Initialization
*      

      NBREAK = 0
      NLEFT  = 0
      IFREE  = N+1
      F1     = 0.0D+00
      F2     = 0.0D+00
      TMIN   = 0.0D+00
      ITMIN  = 0
      IBOUND = 0
      IERR   = 0

      SMALL2 = SMALL
c      SMALL2 = SQRT(SMALL)
     
      
      IF (IOLD .EQ. 1 .OR. IBFGS .EQ. 2) THEN
         MCC = MN
      ELSE IF (IFLAG .EQ. 0) THEN
         MCC = MC + 1
      ELSE
         MCC = MC
      END IF
 
      
*     
*     Temporarily set PAUX = [PAUX1 PAUX2]' = -[UM  SM]'*G
*     and CAUX = [CAUX1 CAUX2]' = [UM  1/GAMMA*SM]'*(XCP-X) = 0
*      
      
      DO 10 I=1,MN
         L = I + IOLD - 1
         IF (L .GT. MCC) L=L-MCC
         PAUX1(I) = -UMTGP(L)
         PAUX2(I) = -SMTGP(L)

         CAUX1(I)= 0.0D+00
         CAUX2(I)= 0.0D+00
 10   CONTINUE

      
*
*     Computation of the Cauchy direction DC, the first derivative
*     F1= -(DC)'*DC, and the breakpoints T in each coordinate
*     direction. Identification of the smallest breakpoint TMIN. 
*      

      DO 20 I=1,N

         IF (IB(I) .NE. 0) THEN
            IF (IB(I).LE.2 .AND. ABS(G(I)) .LE. SMALL2 .AND. 
     &           X(I)-XL(I) .LE. SMALL2) THEN
               DC(I) = 0.0D+00
               XCP(I) = X(I)
               GOTO 20
            END IF

            IF (IB(I).GE.2 .AND. ABS(G(I)) .LE. SMALL2 .AND. 
     &           XU(I)-X(I) .LE. SMALL2) THEN
               DC(I) = 0.0D+00
               XCP(I) = X(I)
               GOTO 20
            END IF
         END IF

         IF (IB(I).NE.0 .AND. IB(I).LE.2 .AND. G(I).GT.0.0D+00) THEN


*
*     Breakpoint in lower bound
*

            T(NBREAK+1) = (X(I) - XL(I))/G(I)
            
            IF (T(NBREAK+1) .LE. SMALL2) THEN
               DC(I) = 0.0D+00


*     
*     Correction to PAUX = [PAUX1 PAUX2]' = [UM  1/GAMMA*SM]'*DC 
*

               DO 30 J=1,MN
                  L = J + IOLD - 1
                  IF (L .GT. MCC) L=L-MCC
                  PAUX1(J) = PAUX1(J)+UM(N*(L-1)+I)*G(I)
                  PAUX2(J) = PAUX2(J)+SM(N*(L-1)+I)*G(I)
 30            CONTINUE
               
            ELSE

               NBREAK = NBREAK + 1
               IT(NBREAK) = I
               DC(I) = -G(I)
               F1 = F1 - DC(I)*DC(I)


*     
*     Determination of the smallest breakpoint
*               

               IF (NBREAK .EQ. 1 .OR. T(NBREAK) .LT. TMIN) THEN
                  TMIN=T(NBREAK)
                  ITMIN=NBREAK
               END IF
            END IF


         ELSE IF (IB(I) .GE. 2 .AND. G(I) .LT. 0.0D+00) THEN


*
*     Breakpoint in upper bound
*               

            T(NBREAK+1) = (X(I) - XU(I))/G(I)

            IF (T(NBREAK+1) .LE. SMALL2) THEN
            
               DC(I) = 0.0D+00


*     
*     Correction to PAUX = [PAUX1 PAUX2]' = [UM  SM]'*DC 
*

               DO 40 J=1,MN
                  L = J + IOLD - 1
                  IF (L .GT. MCC) L=L-MCC
                  PAUX1(J) = PAUX1(J)+UM(N*(L-1)+I)*G(I)
                  PAUX2(J) = PAUX2(J)+SM(N*(L-1)+I)*G(I)
 40            CONTINUE
               
            ELSE

               NBREAK = NBREAK + 1
               IT(NBREAK) = I
               DC(I) = -G(I)
               F1 = F1 - DC(I)*DC(I)


*
*     Determination of the smallest breakpoint
*               

               IF (NBREAK .EQ. 1 .OR. T(NBREAK) .LT. TMIN) THEN
                  TMIN=T(NBREAK)
                  ITMIN=NBREAK
               END IF
            END IF
            
         ELSE


*
*     No breakpoint 
*

            IFREE = IFREE - 1
            IT(IFREE) = I
            DC(I) = -G(I)
            F1 = F1 - DC(I)*DC(I)
            
            IF (ABS(G(I)) .GT. 0.0D+00) IBOUND=-1

         END IF

         
*     
*     Initialization of Cauchy point
*

         XCP(I) = X(I)

 20   CONTINUE


*     
*     Correction to PAUX2 = 1/GAMMA*SM'*DC 
*

      IF (GAMMA .NE. 1.0D+00) THEN

         DO 70 J=1,MN
            PAUX2(J) = PAUX2(J)/GAMMA
 70      CONTINUE

      END IF


*
*     The indices corresponding to free variables are located in
*     IT(1),...,IT(NBREAK) and IT(IFREE),...,IT(N). The smallest of the
*     NBREAK breakpoints is in T(ITMIN)=TMIN.
*
      
*
*     DC is a zero vector. Return with the initial XCP.
*

      IF (NBREAK .EQ. 0 .AND. IBOUND .EQ. 0) GOTO 999


*
*     Initialization of derivative F2 = DC'*BM*DC, where BM is 
*     representated by the compact limited memory BFGS formula.
*

      F2 =  -F1/GAMMA


*
*     The product of the 2MN x 2MN middle matrix in the compact
*     L-BFGS formula and the 2MN vector P; The product is returned
*     in V.
*

      CALL FRMJ(MN,MCC,IOLD,SMTSM,JT,LM,CDIAG,GAMMA,SMALL,IERR)
      IF (IERR .NE. 0) RETURN

      CALL PRDBMV(MN,MCC,IOLD,PAUX1,PAUX2,V1,V2,CDIAG,LM,JT,SMALL,
     &     IERR)
      IF (IERR .NE. 0) RETURN


      F2 = F2 - VDOT(MN,PAUX1,V1) - VDOT(MN,PAUX2,V2)
      

      IF (F2 .LE. SMALL) THEN
         IERR = -8
         RETURN
      END IF


      DLTMIN = -F1/F2

      TOLD = 0.0D+00
      NINT = 1
 
      NLEFT = NBREAK

     
*      
*     There is no breakpoints, locate the GCP and return. 
*

      IF (NBREAK .EQ. 0) GOTO 888

      
*
*     Begining of the loop.
*      

 100  CONTINUE


      IF (NINT .EQ. 1) THEN


*     the smallest breakpoint is TMIN=T(ITMIN) and its index is
*     in IT(ITMIN)

         ISBP = IT(ITMIN)

      ELSE


*     the smallest breakpoint is chosen by heapsort algorithm

         IF (NINT .EQ. 2) THEN


*     Remove the already used smallest breakpoint before heapsort call.
            

            IF (ITMIN .NE. NBREAK) THEN
               T(ITMIN) = T(NBREAK)
               IT(ITMIN) = IT(NBREAK)
            END IF


*     Heapsort with initialization of the heap.

            CALL HPSRT(NLEFT,T,IT,0)

         ELSE


*     Heapsort with updating the heap.

            CALL HPSRT(NLEFT,T,IT,1)

         END IF


         TMIN = T(NLEFT)
         ISBP = IT(NLEFT)

      END IF


      DELTAT = TMIN - TOLD

      IF (DELTAT .LT. 0.0D+00) THEN
         PRINT*,' no nyt voit jo hirtt ihtes.'
      END IF
      

*
*     Minimizer is within this interval, locate the GCP and return.
*

      IF (DLTMIN .LT. DELTAT) GOTO 888


*      
*     Examination of subsequent seqment
*

      NLEFT = NLEFT - 1
      NINT = NINT + 1
      GISBP = G(ISBP)
      DC(ISBP) = 0.0D+00

      IF (GISBP .LT. 0.0D+00) THEN
         XCP(ISBP) = XU(ISBP)
         ZISBP = XU(ISBP) - X(ISBP)
      ELSE 
         XCP(ISBP) = XL(ISBP)
         ZISBP = XL(ISBP) - X(ISBP)
      END IF


*         
*     All  variables are fixed, return with current XCP as GCP.
*         

      IF (NLEFT .EQ. 0 .AND. NBREAK .EQ. N) GOTO 999


*      
*     Update F1 and F2
*      

      GISBP2=GISBP*GISBP


*
*     Initial updating of F1 and F2
*      

      F1 = F1 + DELTAT*F2 + GISBP2 + GISBP*ZISBP/GAMMA
      F2 = F2 - GISBP2/GAMMA


*         
*     Update CAUX1= CAUX1 + DELTAT*PAUX1 and CAUX2 = CAUX2 + DELTAT*PAUX2.
*         

      CALL SCSUM(MN,DELTAT,PAUX1,CAUX1,CAUX1)
      CALL SCSUM(MN,DELTAT,PAUX2,CAUX2,CAUX2)


*     
*     Selection of W, the ISBP:th row of [UM  1/GAMMA*SM]. 
*
         
      DO 50 J=1,MN
         L = J + IOLD - 1
         IF (L .GT. MCC) L=L-MCC
         W1(J)= UM(N*(L-1)+ISBP)
         W2(J)= SM(N*(L-1)+ISBP)/GAMMA
 50   CONTINUE


*
*     Computation of products WMC, WMP, and WMW
*
      
      CALL PRDBMV(MN,MCC,IOLD,W1,W2,V1,V2,CDIAG,LM,JT,SMALL,IERR)
      IF (IERR .NE. 0) RETURN

      WMC = VDOT(MN,CAUX1,V1) + VDOT(MN,CAUX2,V2) 
      WMP = VDOT(MN,PAUX1,V1) + VDOT(MN,PAUX2,V2)
      WMW = VDOT(MN,W1,V1) + VDOT(MN,W2,V2)


*         
*     Update PAUX1 = PAUX1 + GISBP*W1 and PAUX2 = PAUX2 + GISBP*W2. 
*

      CALL SCSUM(MN,GISBP,W1,PAUX1,PAUX1)
      CALL SCSUM(MN,GISBP,W2,PAUX2,PAUX2)
         

*
*     Complete updating of F1 and F2
*      

      F1 = F1 - GISBP*WMC
      F2 = F2 - 2.0D+00*GISBP*WMP - GISBP2*WMW


      IF (F2 .LE. SMALL) THEN
         IERR = -8
         RETURN
      END IF


      DLTMIN = -F1/F2


*
*     Repeat the loop for unsearched intervals.          
*

      TOLD = TMIN
      IF (NLEFT .GT. 0) GOTO 100
      IF (IBOUND .EQ.0) DLTMIN = 0.0D+00   


*     
*     End of the loop.
*


 888  CONTINUE

      DLTMIN = MAX(0.0D+00,DLTMIN)
      TOLD = TOLD + DLTMIN


*      
*     Update free variables 
* 
     
      DO 60 I=1,NLEFT
         XCP(IT(I))=X(IT(I))+TOLD*DC(IT(I))
 60   CONTINUE
      DO 80 I=IFREE,N
         XCP(IT(I))=X(IT(I))+TOLD*DC(IT(I))
 80   CONTINUE


 999  CONTINUE

     
      RETURN
      END
      

 
************************************************************************
*
*     * SUBROUTINE SSBFGS *
*
*      
*     * Purpose *
*      
*     Subspace minimization and computation of the search direction D
*     by the limited memory BFGS update.
*
*      
*     * Calling sequence *
*     
*     CALL SSBFGS(N,NACT,MC,INEW,IOLD,MN,IBFGS,IFLAG,X,XL,XU,XCP,G,D,
*    &     SM,UM,TMPN1,RM,UMTUM,CDIAG,SMTGP,UMTGP,KM,TMPMC1,TMPMC2,
*    &     TMPMC3,TMPMC4,TMPMC5,GAMMA,ITYPE,IACT,IB,IERR,SMALL)
*     
*      
*     * Parameters *
*
*     II  N               Number of variables.
*     II  NACT            Number of active variables.
*     II  MC              Declared number of stored corrections.
*     II  MN              Current number of stored corrections.
*     II  INEW            Index for circular arrays.
*     II  IOLD            Index of the oldest correction.
*     II  IBFGS           Index of the type of BFGS update:
*                             1  - BFGS update: the corrections are
*                                    stored.
*                             2  - BFGS update: the corrections are not
*                                    stored.
*                             3  - BFGS update is skipped.
*     II  IFLAG           Index for adaptive version:
*                             0  - Maximum number of stored corrections
*                                    has not been changed at previous
*                                    iteration.
*                             1  - Maximum number of stored corrections
*                                    has been changed at previous
*                                    iteration.
*     II  ITYPE           Type of problem:
*                             0  - problem is unbounded,
*                             1  - constrained problem.
*     II  IACT(N)         Index set of active and free variables:
*                             for I=1,...,NACT, IACT(I) are the indices
*                                 of active variables (-J, if
*                                 XCP(J)=XL(J) and +J, otherwise) and
*                             for i=NACT+1,...,N, IACT(I) are the indices
*                                 of free variables.
*     II  IB(N)           Type of bound constraints:
*                             0  - X(I) is unbounded,
*                             1  - X(I) has only a lower bound,
*                             2  - X(I) has both lower and upper bounds,
*                             3  - X(I) has only an upper bound. 
*     RI  X(N)            Vector of variables. 
*     RI  XCP(N)          Generalized Chauchy point.
*     RI  XL(N)           Lower bounds for variables. 
*     RI  XU(N)           Upper bounds for variables. 
*     RO  D(N)            Search direction.
*     RI  G(N)            Current subgradient of the objective
*                           function.
*     RI  SM(N*(MC+1))    Matrix whose columns are stored corrections.
*     RI  UM(N*(MC+1))    Matrix whose columns are stored subgradient
*                           differences.
*     RI  RM((MC+2)*(MC+1)/2)  Upper triangular matrix stored columnwise
*                               in the one-dimensional array.
*     RI  UMTUM((MC+2)*(MC+1)/2)  Matrix UMTUM = UM' * UM.
*     RI  C(MC+1)         Diagonal matrix.
*     RI  SMTGP(MC+1)     Vector SMTGP = SM' * GP.
*     RI  UMTGP(MC+1)     Vector UMTGP = UM' * GP.
*     RI  GAMMA           Scaling parameter.
*     RI  SMALL           Small positive value.
*     RA  TMPMC#(MC+1)    Auxiliary arrays; # = 1,...,5.
*     RA  TMPN1(N)        Auxiliary array.
*     RA  KM((2*MC+3)*(MC+1)) Auxiliary 2*MC matrix.
*     RI  SMALL           Small positive value.
*     IO  IERR            Error indicador: 
*                             0  - Everything is ok.
*                            -1  - Error in CHOFA.
*                            -2  - Error in LINEQ.
*                            -3  - Error in TRLIEQ.
*                            -4  - Error in FRMLEL.
*                            -5  - Error in PRDKMV.
*                           -10  - Warning: MCC=0 and IBFGS=3.
*                           -11  - Warning: MCC=0 and ISR1=3.
*     
*
*
*     * Subprograms used *
*      
*     S   FRMLEL          Computation of Cholesky factorization of
*                         indefinite matrix KM.
*     S   PRDKMV          Computation of the product of the inverse of
*                         2mc x 2mc matrix KM and and a 2mc vector P.
*     S   RECMAX          Multiplication of a vector by a dense
*                           rectangular matrix.
*     S   SCALEX          Scaling a vector.
*     S   SCDIFF          Difference of the scaled vector and a vector.
*     S   SCSUM           Sum of a vector and the scaled vector.
*     S   SYMAX           Multiplication of a dense symmetric matrix
*                           by a vector.
*     S   TRLIEQ          Solving X from linear equation L*X=Y or
*                           L'*X=Y.
*     S   VXDIAG          Multiplication of a vector and a diagonal
*                           matrix.
*     S   XDIFFY          Difference of two vectors.
*     S   XSUMY           Sum of two vectors.
*
*
*     All the (MC+1)-vectors are stored in a circular order controlled by
*     the parameter point IOLD.
*
*      
      
      SUBROUTINE SSBFGS(N,NACT,MC,INEW,IOLD,MN,IBFGS,IFLAG,
     &     X,XL,XU,XCP,G,D,SM,UM,TMPN1,RM,UMTUM,CDIAG,SMTGP,UMTGP,KM,
     &     TMPMC1,TMPMC2,TMPMC3,TMPMC4,TMPMC5,GAMMA,
     &     ITYPE,IACT,IB,IERR,SMALL)


*     Scalar Arguments
      INTEGER N,NACT,MC,INEW,IOLD,MN,IBFGS,IFLAG,ITYPE,IERR
      DOUBLE PRECISION GAMMA,SMALL
      
*     Array Arguments
      INTEGER IACT(*),IB(*)
      DOUBLE PRECISION X(*),XL(*),XU(*),XCP(*),D(*),G(*),
     &     SM(*),UM(*),RM(*),UMTUM(*),CDIAG(*),
     &     SMTGP(*),UMTGP(*),KM(*),TMPMC1(*),TMPMC2(*),
     &     TMPMC3(*),TMPMC4(*),TMPMC5(*),TMPN1(*)      

*     Local Scalars
      INTEGER I,J,IFREE,IND,MCC
      DOUBLE PRECISION ALPHA,DFREE,SIGNUM,TMP

*     Intrinsic Functions
      INTRINSIC ABS,SIGN,MIN,DBLE

*     External Subroutines
      EXTERNAL RECMAX,XDIFFY,XSUMY,SCALEX,SYMAX,SCSUM,SCDIFF,
     &     PRDKMV,FRMLEL,TRLIEQ,VXDIAG


      
*
*     Initialization
*      

      IERR = 0
      
      IF (IOLD .EQ. 1 .OR. IBFGS .EQ. 2) THEN
         MCC = MN
      ELSE IF (IFLAG .EQ. 0) THEN
         MCC = MC + 1
      ELSE
         MCC = MC
      END IF

      
*
*     Computation of two intermediate vectors TMPMC1 and TMPMC2
*

      CALL TRLIEQ(MN,MCC,IOLD,RM,TMPMC1,SMTGP,1,SMALL,IERR)
      IF (IERR .NE. 0) RETURN

      CALL SYMAX(MN,MCC,IOLD,UMTUM,TMPMC1,TMPMC3)
      CALL VXDIAG(MCC,CDIAG,TMPMC1,TMPMC2)
      CALL SCSUM(MCC,GAMMA,TMPMC3,TMPMC2,TMPMC2)
      CALL SCSUM(MCC,-GAMMA,UMTGP,TMPMC2,TMPMC3)
      CALL TRLIEQ(MN,MCC,IOLD,RM,TMPMC2,TMPMC3,0,SMALL,IERR)
      IF (IERR .NE. 0) RETURN

      
      IF (NACT .EQ. 0) THEN


*
*     Computation of the search direction D
*
         
         IF (IOLD .EQ. 1 .OR. IBFGS .EQ. 2) THEN
            CALL RECMAX(N,MN,UM,TMPMC1,D)
            
         ELSE
            CALL RECMAX(N,INEW-1,UM,TMPMC1,D)
            CALL RECMAX(N,MN-INEW+1,UM((IOLD-1)*N+1),TMPMC1(IOLD)
     $           ,TMPN1)
            CALL XSUMY(N,D,TMPN1,D)
         END IF

         CALL XDIFFY(N,D,G,D)
         
         IF (IOLD .EQ. 1 .OR. IBFGS .EQ. 2) THEN
            CALL RECMAX(N,MN,SM,TMPMC2,TMPN1)
            CALL SCDIFF(N,GAMMA,D,TMPN1,D)
            
         ELSE
            CALL RECMAX(N,INEW-1,SM,TMPMC2,TMPN1)
            CALL SCDIFF(N,GAMMA,D,TMPN1,D)
            CALL RECMAX(N,MN-INEW+1,SM((IOLD-1)*N+1),TMPMC2(IOLD)
     $           ,TMPN1)
            CALL XDIFFY(N,D,TMPN1,D)
         END IF
         
      ELSE


*
*     Computation of the intermediate vector TMPN1= -A'DM G - A'(XCP-X)
*     

         DO 10 I=1,NACT
            TMPN1(I)=0.0D+00
            IND=ABS(IACT(I))
            SIGNUM = SIGN(1.0D+00,DBLE(IACT(I)))
            D(I) = SIGNUM*(XCP(IND)-X(IND))
               
            IF (IOLD .EQ. 1 .OR. IBFGS .EQ. 2) THEN
               DO 11 J = 1, MN
                  TMPN1(I)=TMPN1(I) + SIGNUM*(SM((J-1)*N+IND)*TMPMC2(J)
     &                 - GAMMA*UM((J-1)*N+IND)*TMPMC1(J))
 11            CONTINUE
            ELSE 
               
               DO 12 J = 1, INEW-1
                  TMPN1(I)=TMPN1(I) + SIGNUM*(SM((J-1)*N+IND)*TMPMC2(J)
     &                 - GAMMA*UM((J-1)*N+IND)*TMPMC1(J))
 12            CONTINUE
               DO 13 J = IOLD, MCC
                  TMPN1(I)=TMPN1(I) + SIGNUM*(SM((J-1)*N+IND)*TMPMC2(J)
     &                 - GAMMA*UM((J-1)*N+IND)*TMPMC1(J))
 13            CONTINUE
               
            END IF
            TMPN1(I) = -SIGNUM*GAMMA*G(IND) - TMPN1(I) - D(I) 
 10      CONTINUE


*
*     Computation of Lagrange multipiliers TMPN1 = (A'DM A)^{-1} TMPN1.
*         
     
*     [GAMMA*UM SM]' A TMPN1 (Note the reorder of blocks).
            
         IF (IOLD .EQ. 1 .OR. IBFGS .EQ. 2) THEN
            DO 20 J = 1, MN
               TMPMC3(J)= 0.0D+00
               TMPMC4(J)= 0.0D+00

               DO 21 I=1,NACT
                  IND=ABS(IACT(I))
                  SIGNUM = SIGN(1.0D+00,DBLE(IACT(I)))
                  TMPMC3(J)=TMPMC3(J)+SIGNUM*GAMMA*UM((J-1)*N+IND)
     &                 *TMPN1(I)
                  TMPMC4(J)=TMPMC4(J)+SIGNUM*SM((J-1)*N+IND)*TMPN1(I)
 21            CONTINUE
 20         CONTINUE
            
         ELSE 
            DO 30 J = 1, INEW-1
               TMPMC3(J)= 0.0D+00
               TMPMC4(J)= 0.0D+00
               DO 31 I=1,NACT
                  IND=ABS(IACT(I))
                  SIGNUM = SIGN(1.0D+00,DBLE(IACT(I)))
                  TMPMC3(J)=TMPMC3(J)+SIGNUM*GAMMA*UM((J-1)*N+IND)
     &                 *TMPN1(I)
                  TMPMC4(J)=TMPMC4(J)+SIGNUM*SM((J-1)*N+IND)*TMPN1(I)
 31            CONTINUE
 30         CONTINUE
            DO 40 J = IOLD, MCC
               TMPMC3(J)= 0.0D+00
               TMPMC4(J)= 0.0D+00
               DO 41 I=1,NACT
                  IND=ABS(IACT(I))
                  SIGNUM = SIGN(1.0D+00,DBLE(IACT(I)))
                  TMPMC3(J)=TMPMC3(J)+SIGNUM*GAMMA*UM((J-1)*N+IND)
     &                 *TMPN1(I)
                  TMPMC4(J)=TMPMC4(J)+SIGNUM*SM((J-1)*N+IND)*TMPN1(I)
 41            CONTINUE
 40         CONTINUE
               
         END IF


*     Computation of [ TMPMC3 TMPMC4 ] = KM^{-1} [ TMPMC3 TMPMC4 ]

         CALL FRMLEL(N,MN,MCC,NACT,IOLD,KM,IACT,UM,SM,CDIAG,GAMMA,
     &        SMALL,IERR)
         IF (IERR .NE. 0) RETURN
         
         CALL PRDKMV(MN,MCC,IOLD,TMPMC3,TMPMC4,TMPMC3,TMPMC4,KM,
     &        SMALL,IERR)
         IF (IERR .NE. 0) RETURN
         
         DO 50 I=1,NACT
            D(I)=0.0D+00
            IND=ABS(IACT(I))
            SIGNUM = SIGN(1.0D+00,DBLE(IACT(I)))
            IF (IOLD .EQ. 1 .OR. IBFGS .EQ. 2) THEN
               DO 51 J = 1, MN
                  D(I)=D(I) + SIGNUM*(SM((J-1)*N+IND)*TMPMC4(J)+
     &                 GAMMA*UM((J-1)*N+IND)*TMPMC3(J))
 51            CONTINUE

            ELSE 
               
               DO 52 J = 1, INEW-1
                  D(I)=D(I) + SIGNUM*(SM((J-1)*N+IND)*TMPMC4(J)+
     &                 GAMMA*UM((J-1)*N+IND)*TMPMC3(J))
 52            CONTINUE
               DO 53 J = IOLD, MCC
                  D(I)=D(I) + SIGNUM*(SM((J-1)*N+IND)*TMPMC4(J)+
     &                 GAMMA*UM((J-1)*N+IND)*TMPMC3(J))
 53            CONTINUE
               
            END IF

            TMPN1(I) = (TMPN1(I) - D(I)/GAMMA) /GAMMA
 50      CONTINUE
      
      
*     
*     Computation of the search direction D = -DM(A TMPN1 + G)
*         

*     [SM GAMMA*UM]' A TMPN1
            
         IF (IOLD .EQ. 1 .OR. IBFGS .EQ. 2) THEN
            DO 60 J = 1, MN
               TMPMC3(J)= 0.0D+00
               TMPMC4(J)= 0.0D+00
               DO 61 I=1,NACT
                  IND=ABS(IACT(I))
                  SIGNUM = SIGN(1.0D+00,DBLE(IACT(I)))
                  TMPMC3(J)=TMPMC3(J)+SIGNUM*SM((J-1)*N+IND)*TMPN1(I)
                  TMPMC4(J)=TMPMC4(J)+SIGNUM*GAMMA*UM((J-1)*N+IND)
     &                 *TMPN1(I)
 61            CONTINUE
 60         CONTINUE
         ELSE 
               
            DO 70 J = 1, INEW-1
               TMPMC3(J)= 0.0D+00
               TMPMC4(J)= 0.0D+00
               DO 71 I=1,NACT
                  IND=ABS(IACT(I))
                  SIGNUM = SIGN(1.0D+00,DBLE(IACT(I)))
                  TMPMC3(J)=TMPMC3(J)+SIGNUM*SM((J-1)*N+IND)*TMPN1(I)
                  TMPMC4(J)=TMPMC4(J)+SIGNUM*GAMMA*UM((J-1)*N+IND)
     &                 *TMPN1(I)
 71            CONTINUE
 70         CONTINUE
            DO 80 J = IOLD, MCC
               TMPMC3(J)= 0.0D+00
               TMPMC4(J)= 0.0D+00
               DO 81 I=1,NACT
                  IND=ABS(IACT(I))
                  SIGNUM = SIGN(1.0D+00,DBLE(IACT(I)))
                  TMPMC3(J)=TMPMC3(J)+SIGNUM*SM((J-1)*N+IND)*TMPN1(I)
                  TMPMC4(J)=TMPMC4(J)+SIGNUM*GAMMA*UM((J-1)*N+IND)
     &                 *TMPN1(I)
 81            CONTINUE
 80         CONTINUE
            
         END IF
            

*     Computation of two intermediate vectors TMPMC3 and TMPMC4
*     such that
*     
*         [ TMPMC3 ] = N * [ TMPMC3 ] + [ TMPMC2 ],
*         [ TMPMC4 ]       [ TMPMC4 ]   [-TMPMC1 ]
*
*     where N is the middle matrix in BFGS formula.

         CALL TRLIEQ(MN,MCC,IOLD,RM,TMPMC3,TMPMC3,1,SMALL,IERR)
         IF (IERR .NE. 0) RETURN
         
         CALL SYMAX(MN,MCC,IOLD,UMTUM,TMPMC3,TMPMC5)
         CALL SCDIFF(MCC,GAMMA,TMPMC5,TMPMC4,TMPMC4)
         CALL VXDIAG(MCC,CDIAG,TMPMC3,TMPMC5)
         CALL XSUMY(MCC,TMPMC5,TMPMC4,TMPMC4)
         
         CALL TRLIEQ(MN,MCC,IOLD,RM,TMPMC4,TMPMC4,0,SMALL,IERR)
         IF (IERR .NE. 0) RETURN

         IF (IOLD .EQ. 1 .OR. IBFGS .EQ. 2) THEN

            DO 90 I=1,MN
               TMP = TMPMC3(I)
               TMPMC3(I)= TMPMC4(I) + TMPMC2(I)
               TMPMC4(I)= -TMP - TMPMC1(I)
 90         CONTINUE
         ELSE 

            DO 100 I= 1, INEW-1
               TMP = TMPMC3(I)
               TMPMC3(I)= TMPMC4(I) + TMPMC2(I)
               TMPMC4(I)= -TMP - TMPMC1(I)
 100        CONTINUE
            DO 110 I= IOLD, MCC
               TMP = TMPMC3(I)
               TMPMC3(I)= TMPMC4(I) + TMPMC2(I)
               TMPMC4(I)= -TMP - TMPMC1(I)
 110        CONTINUE
         END IF

         CALL SCALEX(N,-GAMMA,G,D)

         DO 120 I=1,NACT
            IND=ABS(IACT(I))
            SIGNUM = SIGN(1.0D+00,DBLE(IACT(I)))
            
            D(IND) = D(IND) - GAMMA*SIGNUM*TMPN1(I)
 120     CONTINUE
         
         IF (IOLD .EQ. 1 .OR. IBFGS .EQ. 2) THEN
            CALL RECMAX(N,MN,SM,TMPMC3,TMPN1)
            CALL XDIFFY(N,D,TMPN1,D)
            CALL RECMAX(N,MN,UM,TMPMC4,TMPN1)
            CALL SCSUM(N,-GAMMA,TMPN1,D,D)
        
         ELSE 
            CALL RECMAX(N,INEW-1,SM,TMPMC3,TMPN1)
            CALL XDIFFY(N,D,TMPN1,D)
            CALL RECMAX(N,MN-INEW+1,SM((IOLD-1)*N+1),TMPMC3(IOLD)
     &           ,TMPN1)
            CALL XDIFFY(N,D,TMPN1,D)
            
            CALL RECMAX(N,INEW-1,UM,TMPMC4,TMPN1)
            CALL SCSUM(N,-GAMMA,TMPN1,D,D)
            CALL RECMAX(N,MN-INEW+1,UM((IOLD-1)*N+1),TMPMC4(IOLD)
     &           ,TMPN1)
            CALL SCSUM(N,-GAMMA,TMPN1,D,D)
         END IF

      END IF

      
*
*     Backtrack to the feasible region if X + D violates bounds
*     (among free variables).
*

      IF (ITYPE .EQ. 1) THEN

         ALPHA = 1.0D+00
         DO 200 I = NACT+1,N
            IFREE=IACT(I)
            DFREE=D(IFREE)
            IF (IB(IFREE) .NE. 0) THEN
               IF (IB(IFREE) .LE. 2 .AND. 
     &              X(IFREE) + DFREE .LT. XL(IFREE)) THEN

*     Otherwise the lower bound is never met
 
                  IF (ABS(X(IFREE)-XCP(IFREE)+DFREE) .GT. SMALL) THEN
                     ALPHA=MIN(ALPHA,(XL(IFREE)-XCP(IFREE))/
     &                    (X(IFREE)-XCP(IFREE)+DFREE))
                  END IF

               ELSE IF (IB(IFREE) .GE. 2 .AND. 
     &                 X(IFREE) + DFREE  .GT. XU(IFREE)) THEN

*     Otherwise the upper bound is never met
 
                  IF (ABS(X(IFREE)-XCP(IFREE)+DFREE) .GT. SMALL) THEN
                     ALPHA=MIN(ALPHA,(XU(IFREE)-XCP(IFREE))/
     &                    (X(IFREE)-XCP(IFREE)+DFREE))
                  END IF
                  
               END IF
               
            END IF
 200     CONTINUE

         
         IF (ALPHA .LT. 1.0D+00) THEN
            DO 210 I=NACT+1,N
               IFREE=IACT(I)
               D(IFREE) = XCP(IFREE) - X(IFREE) +
     &              ALPHA*(X(IFREE) - XCP(IFREE) + D(IFREE))
 210        CONTINUE
         END IF

         ALPHA=0.0D+00
         J=0
         DO 220 I=1,NACT
            IND=ABS(IACT(I))
            D(IND)=XCP(IND)-X(IND)

 220     CONTINUE
            
      END IF

 
      RETURN
      END



************************************************************************
*
*     * SUBROUTINE FRMJ *
*
*      
*     * Purpose *
*      
*     Formation of the JJ' Cholesky factorization of the matrix
*      
*          JT = SM'*SM/GAMMA + LM*C^(-1)*LM'.
*
*      
*     * Calling sequence *
*     
*     CALL FRMJ(MN,MCC,IOLD,SMTSM,JT,LM,CDIAG,GAMMA,SMALL,IERR)
*
*      
*     * Parameters *
*
*     II  MN              Current number of corrections.
*     II  MCC             Current size of vectors. 
*     II  IOLD            Index of the oldest correction.
*     RO  JT((MN+1)*MN/2) The upper triangular matrix J' that is
*                             the Cholesky factorization of
*                             (SM'SM/GAMMA+LM CDIAG^(-1) LM').  
*     RI  LM((MN+1)*MN/2) Lower triangular matrix stored rowwise
*                             in the one-dimensional array.
*     RI  SMTSM((MN+1)*MN/2)  Matrix SMTSM = SM' * SM.
*     RI  CDIAG(MCC)      Diagonal matrix stored in circular order.
*     RI  GAMMA           Scaling parameter.
*     RI  SMALL           Small positive value.
*     IO  IERR            Error indicador: 
*                             0  - Everything is ok.
*                            -1  - Error: matrix is not positive
*                                  definite.
*
*
*     * Subprograms used *
*      
*     S   CHOFA           Cholesfy factorization.
*     
*     
      
      SUBROUTINE FRMJ(MN,MCC,IOLD,SMTSM,JT,LM,CDIAG,GAMMA,SMALL,IERR)

      
*     Scalar Arguments
      INTEGER MN,MCC,IOLD,IERR
      DOUBLE PRECISION GAMMA,SMALL
      
*     Array Arguments
      DOUBLE PRECISION SMTSM(*),JT(*),LM(*),CDIAG(*)

*     Local Scalars
      INTEGER I,IJ,JK,IK,K,J,JC
      
*     External Subroutines
      EXTERNAL CHOFA      


    
      IERR = 0
      IF (MN .EQ. 0) RETURN

      
*     Form JT = SM'*SM/GAMMA + LM*CDIAG^(-1)*LM'.

      IJ = 0
      DO 10 I = 1, MN
         IK = 0
         DO 20 K = 1,I
            IJ=IJ+1
            JT(IJ) = SMTSM(IJ)/GAMMA
            DO 30 J = 1, K
               JC=J+IOLD-1
               IF (JC .GT. MCC) JC=JC-MCC
               IK = IK+1
               JK = IJ+J-K
               JT(IJ) = JT(IJ) + LM(IK)*LM(JK)/CDIAG(JC)
 30         CONTINUE
 20      CONTINUE
 10   CONTINUE


*     Cholesky factorizion of JT = SM'SM/GAMMA + LM CDIAG^(-1)LM' to JJ'.
 
      CALL CHOFA(MN,JT,SMALL,IERR)

      RETURN
      END
 
     

************************************************************************
*
*     * SUBROUTINE PRDBMV *
*
*      
*     * Purpose *
*      
*     Computation of the product of the 2MN x 2MN middle matrix
*
*     [ -C^(1/2)   C^(-1/2)*LM'  ]^(-1) [  C^(1/2)      0  ]^(-1)
*     [  0         JT'           ]      [ -LM*C^(-1/2)  JT ] 
*      
*     in the compact L-BFGS formula and a 2MN vector P = [ P1 P2 ]'.
*     The product is returned on V = [ V1 V2 ]' .
*
*      
*     * Calling sequence *
*     
*     CALL PRDBMV(MN,MCC,IOLD,P1,P2,V1,V2,C,LM,JT,SMALL,IERR)
*
*      
*     * Parameters *
*     
*
*     II  MN              Current number of stored corrections.
*     II  MCC             Current number of depositories used.
*     II  IOLD            Index for circular arrays.
*     RI  P1(MCC)         Input vector.
*     RI  P2(MCC)         Input vector.
*     RO  V1(MCC)         Output vector.
*     RO  V2(MCC)         Output vector.
*     RI  LM((MN+1)*MN/2) Lower triangular matrix stored rowwise
*                           in the one-dimensional array.
*     RI  C(MCC)          Diagonal matrix.
*     RI  JT((MN+1)*MN/2) Upper triangular matrix J' that is the
*                         Cholesky factor of (1/GAMMA*SM'SM+LC^(-1)L').
*     RI  SMALL           Small positive value.
*     IU  IERR            Error indicador: 
*                             0  - Everything is ok.
*                            -3  - Error; 0 at diagonal.
*     
*     
*
*     All the input matrices are stored in a circular order
*     controlled by the parameter point iold.
*
*
      
      SUBROUTINE PRDBMV(MN,MCC,IOLD,P1,P2,V1,V2,C,LM,JT,SMALL,IERR)
 
     
*     Scalar Arguments
      INTEGER MN,MCC,IOLD,IERR
      DOUBLE PRECISION SMALL
      
*     Array Arguments
      DOUBLE PRECISION P1(*),P2(*),C(*),LM(*),V1(*),V2(*),JT(*)

*     Local Scalars
      INTEGER I,II,K,IK,J
      DOUBLE PRECISION TMP
      
*     Intrinsic Functions
      INTRINSIC SQRT


     
*
*     Initialization
*

      IERR = 0
      IF (MN .EQ. 0) RETURN


*
*     Solving [  C^(1/2)      0  ] [ v1 ] = [ p1 ]
*             [ -LM*C^(-1/2)  JT ] [ v2 ]   [ p2 ].


*     Solving C^(1/2) V1 = P1.
      
      DO 30 I = 1, MN
         J=I+IOLD-1
         IF (J .GT. MCC) J=J-MCC
         V1(I) = P1(I) / SQRT(C(J))
 30   CONTINUE 


*     
*     Solving JT V2 = P2+LM C^(-1)P1.
*      

      IK = 0
      V2(1) = P2(1)
      DO 10 I = 2, MN
         IK=IK+1
         TMP = 0.0D+00
         DO 20 K = 1, I - 1
            J=K+IOLD-1
            IF (J .GT. MCC) J=J-MCC
            IK = IK + 1
            TMP = TMP + LM(IK)*P1(K)/C(J)
 20      CONTINUE
      
         V2(I) = P2(I) + TMP
 10   CONTINUE  


*
*     Solving the triangular system.
*

*     Vector V2 is not in circular order.

      CALL TRLIEQ(MN,MN,1,JT,V2,V2,0,SMALL,IERR)

      IF (IERR .NE. 0) RETURN
 

*
*     Solving [ -C^(1/2)   C^(-1/2)*LM'  ] [ V1 ] = [ V1 ]
*             [  0         JT'           ] [ V2 ]   [ V2 ].

*      
*     Solving JT'V2=V2.
*      

      CALL TRLIEQ(MN,MN,1,JT,V2,V2,1,SMALL,IERR)
      IF (IERR .NE. 0) RETURN


*      
*     Computation of V1=-C^(-1/2)(V1-C^(-1/2)LM'V2)
*                      =-C^(-1/2)V1+C^(-1)LM'V2.
*

      DO 40 I = 1, MN
         J=I+IOLD-1
         IF (J .GT. MCC) J=J-MCC
         V1(I) = -V1(I) / SQRT(C(J))
 40   CONTINUE

      II = 1
      DO 50 I = 1, MN
         TMP = 0.0D+00
         J=I+IOLD-1
         IF (J .GT. MCC) J=J-MCC
         II = II + I + 1
         IK = II - 1
         DO 60 K = I+1, MN
            TMP = TMP + LM(IK)*V2(K)
            IK = IK + K
  60     CONTINUE
         TMP = TMP/C(J)
         V1(I) = V1(I) + TMP
  50  CONTINUE
      

      RETURN
      END


************************************************************************
*
*     * SUBROUTINE FRMLEL *
*
*      
*     * Purpose *
*      
*     Formation of the LEL' factorization of the indefinite matrix
*      
*          KM = [-CDIAG - GAMMA*U'ZZ'U     L_a' - R_z'  ]
*               [L_a - R_z                 S'AA'S/GAMMA ]
*     
*     where     E = [-I  0]
*                   [ 0  I]
*
*     and L_a is the strictly lower triangular part of S'AA'U and R_z is
*     the upper triangular part of S'ZZ'U.
*     
* 
*     * Calling sequence *
*     
*     CALL FRMLEL(N,MN,MCC,NACT,IOLD,KM,IACT,UM,SM,CDIAG,GAMMA,SMALL,
*    &      IERR)
*
*      
*     * Parameters *
*     
*     II  N               Number of variables.
*     II  MN              Current number of stored corrections.
*     II  MCC             Current size of vectors.
*     II  IOLD            Index for circular arrays.
*     II  NACT            Number of active variables.
*     II  IACT(N)         Index set of active and free variables:
*                             for I=1,...,NACT, IACT(I) are the indices
*                                 of active variables (-J, if
*                                 XCP(J)=XL(J) and +J, otherwise)
*                             for i=NACT+1,...,N, IACT(I) are the indices
*                                 of free variables.
*     RI  SM(N*MCC)       Matrix whose columns are stored corrections.
*     RI  UM(N*MCC)       Matrix whose columns are stored subgradient
*                           differences.
*     RI  GAMMA           Scaling parameter.
*     RI  CDIAG(MCC)      Diagonal matrix.
*     RO  KM((2MN+1)*MN)  LEL' factorization of indefinite matrix.
*     RI  SMALL           Small positive value.
*     IO  IERR            Error indicador: 
*                             0  - Everything is ok.
*                            -1  - Error in CHOFA.
*                            -3  - Error in TRLIEQ.
*                            -4  - Error: nonpositive value at diagonal.
*     
*     
*     * Subprograms used *
*
*     RF  VDOT            Dot product of two vectors.
*     S   CHOFA           Cholesfy factorization.
*     S   TRLIEQ          Solving X from linear equation L*X=Y or
*                           L'*X=Y.
*
*     
*     The input matrices SM, UM and CDIAG are stored in a circular order
*     controlled by the parameter point IOLD.
*
*
*     * Original method *
*      
*     C. Zhu, R.H. Byrd, P. Lu, J. Nocedal, `L-BFGS-B: a limited memory
*     FORTRAN code for solving bound constrained optimization problems',
*     Tech. Report, NAM-11, Northwestern University, 1994.
*
*
*     * Limited memory bundle version *
*      
*     Marjo Haarala (2005)
*

      
      SUBROUTINE FRMLEL(N,MN,MCC,NACT,IOLD,KM,IACT,UM,SM,CDIAG,GAMMA,
     &     SMALL,IERR)

*     Scalar Arguments
      INTEGER N,MN,MCC,NACT,IOLD,IERR
      DOUBLE PRECISION GAMMA,SMALL
      
*     Array Arguments
      INTEGER IACT(*)
      DOUBLE PRECISION KM(*),UM(*),SM(*),CDIAG(*)

*     Local Scalars
      INTEGER I,II,I1,IJ,IJ2,IJ3,J,JI,JJ,J1,K,LI,LJ,IND
      DOUBLE PRECISION SUM,TMP

*     External Functions
      DOUBLE PRECISION VDOT
      EXTERNAL VDOT
      
*     Intrinsic Functions
      INTRINSIC ABS,SQRT

*     External Subroutines
      EXTERNAL CHOFA,TRLIEQ
           
*
*     Initialization
*
      IERR = 0


*
*     Form the upper triangle
*     
*     KM = [CDIAG+GAMMA*U'ZZ'U   -L_a'+R_z'   ] 
*          [-L_a+R_z             S'AA'S/GAMMA ]
*

c     Note this could be done more efficiently

      IJ = 0
      IJ2 = MN*(MN+1)/2
      IJ3 = IJ2
      
      DO 10 I = 1,MN
         LI=I+IOLD-1
         IF (LI .GT. MCC) LI=LI-MCC

*     computation of -L_a'

         DO 11 J = 1,I-1
            LJ=J+IOLD-1
            IF (LJ .GT. MCC) LJ=LJ-MCC
            IJ3 = IJ3+1
            KM(IJ3)=0.0D+00
            DO 12 K = 1,NACT
               IND = ABS(IACT(K))
               KM(IJ3) = KM(IJ3) - SM((LI-1)*N+IND)*UM((LJ-1)*N+IND)
 12         CONTINUE
 11      CONTINUE

*     computation of R_z'

         DO 13 J = I,MN
            LJ=J+IOLD-1
            IF (LJ .GT. MCC) LJ=LJ-MCC
            IJ3 = IJ3+1
            KM(IJ3)=0.0D+00

            DO 14 K = NACT+1,N
               IND = IACT(K)
               KM(IJ3) = KM(IJ3) + SM((LI-1)*N+IND)*UM((LJ-1)*N+IND)
 14         CONTINUE

 13      CONTINUE

         IJ2=IJ2+MN
         DO 15 J=1,I
            LJ=J+IOLD-1
            IF (LJ .GT. MCC) LJ=LJ-MCC
            IJ=IJ+1
            IJ2=IJ2+1

*     computation of GAMMA*U'ZZ'U

            KM(IJ)=0.0D+00
            DO 16 K = NACT + 1, N
               IND = IACT(K)
               KM(IJ) = KM(IJ) + UM((LJ-1)*N+IND)*UM((LI-1)*N+IND)
 16         CONTINUE
            KM(IJ)= GAMMA*KM(IJ)

c     computation of S'AA'S/GAMMA

            KM(IJ2)=0.0D+00
            DO 17 K = 1,NACT
               IND = ABS(IACT(K))
               KM(IJ2) = KM(IJ2) + SM((LJ-1)*N+IND)*SM((LI-1)*N+IND)
 17         CONTINUE
            KM(IJ2)= KM(IJ2)/GAMMA
 15      CONTINUE

         IJ3=IJ3+I
 10   CONTINUE
      
*     computation of CDIAG+GAMMA*U'ZZ'U

      II=0
      DO 20 I=1,MN
         LI=I+IOLD-1
         IF (LI .GT. MCC) LI=LI-MCC
         II=II+I
         KM(II)=CDIAG(LI)+KM(II)
 20   CONTINUE


*
*     Form the upper triangle of
*     
*     KM=[ LL'              L^-1(-La'+Rz')                              ]
*        [(-L_a+R_z)L'^-1   S'AA'S/GAMMA+(L^-1(-La'+Rz'))'L^-1(-La'+Rz')]
*      

*     First Cholesky factor (1,1) block of KM to get LL'
*     with L' stored in the (1,1) upper triangle of KM.

      CALL CHOFA(MN,KM,SMALL,IERR)
      IF (IERR .NE. 0) RETURN

*     then form L^-1(-L_a'+R_z') in the (1,2) block.

      K = MN*(MN+1)/2+1
      DO 30 I = 1, MN
         CALL TRLIEQ(MN,MN,1,KM,KM(K),KM(K),0,SMALL,IERR)
         IF (IERR .NE. 0) RETURN
         K=K+MN+I
 30   CONTINUE


*     
*     Form S'AA'S/GAMMA + (L^-1(-L_a'+R_z'))'L^-1(-L_a'+R_z') in the
*     upper triangle of (2,2) block of KM.
*

      IJ2 = MN*(MN+1)/2
      JI = IJ2+1

      DO 40 I = 1,MN
         IJ2 = IJ2 + MN
         IJ = MN*(MN+1)/2+1
         DO 41 J = 1,I
            IJ2 = IJ2 + 1
            KM(IJ2) = KM(IJ2)+VDOT(MN,KM(IJ),KM(JI))
            IJ=IJ+J+MN
 41      CONTINUE
         JI=JI+I+MN
 40   CONTINUE


*      
*     Cholesky factorization of (2,2) block of KM.
*

      II= (MN+1)*(MN+2)/2
      I1= II

      DO 50 I=1,MN
         SUM=0.0D+00
         JI = I1
         JJ = (MN+1)*(MN+2)/2
         J1 = JJ
         DO 51 J=1, I-1
            TMP = KM(JI) - VDOT(J-1,KM(J1),KM(I1))
            TMP = TMP/KM(JJ)
            KM(JI)=TMP
            SUM=SUM+TMP*TMP
            JI = JI + 1
            J1 = JJ + MN + 1
            JJ = JJ + MN + 1 + J
 51      CONTINUE
         SUM = KM(II) - SUM

         IF (SUM .LE. 0.0D0) THEN
c         IF (SUM .LE. SMALL) THEN
            IERR = -4
            RETURN
         END IF
         KM(II) = SQRT(SUM)
         I1 = II + MN + 1
         II = II + MN + 1 + I
 50   CONTINUE

      
      RETURN
      END
      
      
************************************************************************
*     
*     * SUBROUTINE PRDKMV *
*
*      
*     * Purpose *
*      
*     Computation of the product of the inverse of 2MN x 2MN matrix
*     
*     KM = [-C - GAMMA*U'ZZ'U     L_a' - R_z'  ]
*          [L_a - R_z             S'AA'S/GAMMA ]
*      
*     and a 2MN vector [ P1 P2 ]. The indefinite matrix KM is given
*     using Cholesky factorization KM=L*E*L' obtained by the subroutine
*     FRMLEL.
*
*      
*     * Calling sequence *
*     
*     CALL PRDKMV(MN,MCC,IOLD,P1,P2,V1,V2,KM,SMALL,IERR)
*
*      
*     * Parameters *
*     
*
*     II  MN              Current number of corrections.
*     II  MCC             Current size of vectors. 
*     II  IOLD            Index of the oldest corrections.
*     RI  KM(MN*(MN+1))   Factorization KM=L*E*L' obtained by the
*                           subroutine FRMLEL.
*     RI  P1(MCC)         Input vector.
*     RI  P2(MCC)         Input vector
*     RO  V1(MCC)         Output vector.
*     RO  V2(MCC)         Output vector.
*     RI  SMALL           Small positive value.
*     IO  IERR            Error indicador: 
*                             0  - Everything is ok.
*                            -3  - Error in TRLIEQ.
*                            -5  - Error: 0 at diagonal.
*     
*     
*     * Subprograms used *
*
*     S   TRLIEQ          Solving X from linear equation L*X=Y or
*                           L'*X=Y.
*
*     
*     All the input vectors are stored in a circular order
*     controlled by the parameter point IOLD.
*
*
      
      SUBROUTINE PRDKMV(MN,MCC,IOLD,P1,P2,V1,V2,KM,SMALL,IERR)
      
*     Scalar Arguments
      INTEGER MN,MCC,IOLD,IERR
      DOUBLE PRECISION SMALL
      
*     Array Arguments
      DOUBLE PRECISION P1(*),V1(*),P2(*),V2(*),KM(*)

*     Local Scalars
      INTEGER I,II,IJ,L,J,K

*     External Subroutines
      EXTERNAL TRLIEQ

      IERR = 0

      IF (MN .EQ. 0) RETURN
*
*     Compute V=KM^(-1)P , where V=[V1' V2']', P=[P1' P2']' and KM=LEL'.
*

*     Solving V1 (=EL1'V1) from L1 V1 = P1, where L1 is the upper part of L
      CALL TRLIEQ(MN,MCC,IOLD,KM,V1,P1,0,SMALL,IERR)
      IF (IERR .NE. 0) RETURN
      
      
*     Solving the rest of the triangular system: V2=L**(-1)*P2.
      II = MN*(MN+1)/2
      IJ = II
      DO 10 I = 1,MN
         II = II + I + MN
         L=I+IOLD-1
         IF (L .GT. MCC) L=L-MCC

         IF (KM(II) .EQ. 0.0D+00) THEN
            IERR = -5
            RETURN
         END IF
         V2(L) = P2(L)
c         V2(L) = P2(L)/KM(II)
         
         DO 20 J = 1,MN
            IJ = IJ + 1
            K=J+IOLD-1
            IF (K .GT. MCC) K=K-MCC
            V2(L) = V2(L) - KM(IJ)*V1(K)
c            V2(L) = V2(L) - KM(IJ)*V1(K)/KM(II)
 20      CONTINUE
         DO 30 J = 1,I - 1
            IJ = IJ + 1
            K=J+IOLD-1
            IF (K .GT. MCC) K=K-MCC
            V2(L) = V2(L) - KM(IJ)*V2(K)
c            V2(L) = V2(L) - KM(IJ)*V2(K)/KM(II)
 30      CONTINUE
         IJ = IJ + 1
         V2(L) = V2(L)/KM(II)
 10   CONTINUE

      DO 40 I=1,MCC
         V1(I) = -V1(I)
 40   CONTINUE

*     Solving V2 from L2' V2 = V2, where L2 is the lower MN part of L 
      II = 2*MN * (2*MN+1)/2
      DO 50 I = MN,1,-1
         L=I+IOLD-1
         IF (L .GT. MCC) L=L-MCC
         IF (KM(II) .EQ. 0.0D+00) THEN
            IERR = -5
            RETURN
         END IF
         IJ = II 
         DO 60 J = I + 1,MN
            K=J+IOLD-1
            IF (K .GT. MCC) K=K-MCC
            IJ = IJ + MN + J - 1
            V2(L) = V2(L) - KM(IJ)*V2(K)
 60      CONTINUE
         V2(L)=V2(L)/KM(II)
         II = II - I - MN
 50   CONTINUE

      DO 70 I = MN,1,-1
         L=I+IOLD-1
         IF (L .GT. MCC) L=L-MCC
         IF (KM(II) .EQ. 0.0D+00) THEN
            IERR = -5
            RETURN
         END IF
         IJ = II 
         DO 80 J = I + 1,MN
            K=J+IOLD-1
            IF (K .GT. MCC) K=K-MCC
            IJ = IJ + J - 1
            V1(L) = V1(L) - KM(IJ)*V1(K)
 80      CONTINUE
         DO 90 J = 1,MN
            K=J+IOLD-1
            IF (K .GT. MCC) K=K-MCC
            IJ = IJ + MN + J - 1
            V1(L) = V1(L) - KM(IJ)*V2(K)
 90      CONTINUE
         V2(L)=V2(L)/KM(II)
         II = II - I
 70   CONTINUE

      RETURN
      END


************************************************************************
*
*     * SUBROUTINE DLSR1 *
*
*      
*     * Purpose *
*      
*     Matrix updating and computation of the search direction by the
*     limited memory SR1 update.
*
*      
*     * Calling sequence *
*     
*     CALL DLSR1(N,NACT,MC,MCC,INEW,ISR1,IFLAG,ITYPE,IB,IACT,
*    &     X,XCP,XL,XU,D,GP,GA,S,U,SM,UM,R,LM,UMTUM,SMTSM,C,SMTGP,
*    &     UMTGP,GAMMA,TMPMC1,TMPMC2,TMPMC3,TMPMC4,TMPMC5,TMPMC6,
*    &     TMPN1,TMPN2,TMPMAT,TMPMA2,NNK,SMALL,IPRINT,IERR)
*     
*     
*     * Parameters *
*
*     II  N               Number of variables.
*     IU  NACT            Number of active variables.
*     II  MC              Declared number of stored corrections.
*     IU  MCC             Current number of sepositories used.
*     IU  INEW            Index for circular arrays.
*     IO  ISR1            Index of the type of SR1 update.
*                           1  - SR1 update: the corrections are stored.
*                           3  - SR1 update is skipped.
*     IU  IFLAG           Index for adaptive version:
*                             0  - Maximum number of stored corrections
*                                    has not been changed at previous
*                                    iteration.
*                             1  - Maximum number of stored corrections
*                                    has been changed at previous
*                                    iteration.
*     II  ITYPE           Type of problem:
*                             0  - problem is unbounded,
*                             1  - constrained problem.
*     II  IB(N)           Type of bound constraints:
*                             0  - X(I) is unbounded,
*                             1  - X(I) has only a lower bound,
*                             2  - X(I) has both lower and upper bounds,
*                             3  - X(I) has only an upper bound. 
*     IO  IACT(N)         Index set of active and free variables:
*                             for I=1,...,NACT, IACT(I) are the indices
*                                 of active variables (-J, if
*                                 XCP(J)=XL(J) and +J, otherwise)
*                             for i=NACT+1,...,N, IACT(I) are the indices
*                                 of free variables.
*     RI  X(N)            Vector of variables.
*     RO  XCP(N)          Generalized Cauchy point.
*     RI  XL(N)           Lower bounds for variables. 
*     RI  XU(N)           Upper bounds for variables. 
*     RU  D(N)            Search direction.
*     RI  GP(N)           Basic subgradient of the objective function.
*     RI  GA(N)           Current aggregate subgradient.
*     RI  S(N)            Difference of auxiliary and current variables.
*     RI  U(N)            Difference of auxiliary and current
*                           subgradients.
*     RU  SM(N*(MC+1))    Matrix whose columns are stored corrections.
*     RU  UM(N*(MC+1))    Matrix whose columns are stored subgradient
*                           differences.
*     RU  R((MC+2)*(MC+1)/2)  Upper triangular matrix stored columnwise
*                               in the one-dimensional array.
*     RU  LM((MC+2)*(MC+1)/2) Lower triangular matrix stored rowwise in the
*                           one-dimensional array.
*     RU  UMTUM((MC+2)*(MC+1)/2)  Matrix UMTUM = UM' * UM.
*     RU  SMTSM((MC+2)*(MC+1)/2)  Matrix SMTSM = SM' * SM.
*     RU  C(MC+1)         Diagonal matrix.
*     RU  SMTGP(MC+1)     Vector SMTGP = SM'*GP.
*     RU  UMTGP(MC+1)     Vector UMTGP = UM'*GP.
*     RU  GAMMA           Scaling parameter.
*     RO  SMTGA(MC+1)     Vector SMTGA = SM'*GA.
*     RO  UMTGA(MC+1)     Vector UMTGA = UM'*GA.
*     RA  TMPMC#(MC+1)    Auxiliary arrays: # = 3,6.
*     RA  TMPN1(N)        Auxiliary array:
*                           On input: previous aggregate subgradient.
*     RA  TMPN2(N)        Auxiliary array.
*     RA  TMPMAT((MC+1)*(MC)/2)  Auxiliary matrix.
*     RA  TMPMA2((MC+1)*(MC)/2)  Auxiliary matrix.
*     II  NNK             Consecutive null steps counter.
*     RI  SMALL           Small positive value.
*     II  IPRINT          Printout specification.
*     IO  IERR            Error indicador: 
*                             0  - Everything is ok.
*                            -1  - Error in CHOFA.
*                            -2  - Error in LINEQ.
*                            -3  - Error in TRLIEQ.
*                            -4  - Error in FRMLEL.
*                            -5  - Error in PRDKMV.
*                            -7  - Error in CPSR1, a nonpositive
*                                  definite matrix detected.
*                           -10  - Warning: MCC=0 and IBFGS=3.
*                           -11  - Warning: MCC=0 and ISR1=3.
*     
*     
*     * Local variables *
*
*     I   MN              Current size of vectors.
*     I   IOLD            Index of the oldest corrections.
*     I   IFLAG2          Index for adaptive version.
*
*      
*     * Subprograms used *
*      
*     RF  VDOT            Dot product of two vectors.
*     S   ACTVAR          Finding the index set of free and active
*                           variables  at the generalized Cauchy point.
*     S   CALQ            Solving X from linear equation A*X=Y.
*     S   COPY            Copying of a vector.
*     S   COPY2           Copying of two vectors.
*     S   COPY4           Copying of four vectors.
*     S   CPSR1           Computing the generalized Cauchy point by
*                           limited memory SR1 update formula.
*     S   INDIC1          Initialization of indices.
*     S   RECMAX          Multiplication of a vector by a dense
*                           rectangular matrix.
*     S   RWAXV2          Multiplication of two rowwise stored dense 
*                           rectangular matrices A and B by vectors X 
*                           and Y.
*     S   XDIFFY          Difference of two vectors.
*     S   VNEG            Copying of a vector with change of the sign
*     S   SCALEX          Scaling a vector.
*     S   XSUMY           Sum of two vectors.
*     S   SCDIFF          Difference of the scaled vector and a vector.
*     S   SSSR1           Subspace minimization and computation of the
*                           search direction by the L-SR1 update.
*     
*
*     The variable and subgradient differences and all the (MC+1)-vectors
*     are stored in a circular order controlled by the parameter point
*     INEW.
*
*     
  
      SUBROUTINE DLSR1(N,NACT,MC,MCC,INEW,ISR1,IFLAG,ITYPE,IB,IACT,
     &     X,XCP,XL,XU,D,GP,GA,S,U,SM,UM,R,LM,UMTUM,SMTSM,C,SMTGP,
     &     UMTGP,GAMMA,SMTGA,UMTGA,TMPMC3,TMPMC4,TMPMC5,TMPMC6,TMPN1,
     &     TMPN2,TMPMAT,TMPMA2,NNK,SMALL,IPRINT,IERR)
      

*     Scalar Arguments
      INTEGER N,NACT,MC,MCC,INEW,ISR1,IFLAG,NNK,IPRINT,ITYPE,IERR
      DOUBLE PRECISION GAMMA,SMALL
      
*     Array Arguments
      INTEGER IB(*),IACT(*)
      DOUBLE PRECISION X(*),XCP(*),XL(*),XU(*),D(*),GP(*),GA(*),
     &     S(*),U(*),SM(*),UM(*),R(*),LM(*),UMTUM(*),SMTSM(*),C(*),
     &     SMTGP(*),UMTGP(*),SMTGA(*),UMTGA(*),TMPMC3(*),TMPMC4(*),
     &     TMPMC5(*),TMPMC6(*),TMPN1(*),TMPN2(*),TMPMAT(*),TMPMA2(*)


*     Local Scalars
      INTEGER I,J,K,MN,IOLD,IFLAG2,ICON
      DOUBLE PRECISION STU,GDOLDG,GDNEWG
      
*     External Functions
      DOUBLE PRECISION VDOT
      EXTERNAL VDOT

*     External Subroutines
      EXTERNAL RECMAX,RWAXV2,COPY,VNEG,XSUMY,XDIFFY,SCALEX,CALQ,
     &     SCDIFF,COPY2,COPY4,INDIC1,CPSR1,ACTVAR,SSSR1


*
*     Initialization
*


      IERR = 0
      ISR1 = 0
      IFLAG2 = 0
      ICON = 0


*
*     Initialization of indices
*      

      CALL INDIC1(MC,MCC,MN,INEW,IOLD,IFLAG,IFLAG2,3)
      

*         
*     Computation of GAMMA 
*

      GAMMA = 1.0D+00

      
*     
*     Computation of SM'*GA and UM'*GA
*

      IF (IOLD .EQ. 1 .OR. IOLD .EQ. 2) THEN
         CALL RWAXV2(N,MN,SM((IOLD-1)*N+1),UM((IOLD-1)*N+1),GA,GA,
     &        SMTGA(IOLD),UMTGA(IOLD))

      ELSE
         CALL RWAXV2(N,INEW-1,SM,UM,GA,GA,SMTGA,UMTGA)
         CALL RWAXV2(N,MN-INEW+1,SM((IOLD-1)*N+1),UM((IOLD-1)*N+1),
     &        GA,GA,SMTGA(IOLD),UMTGA(IOLD))
      END IF

      
*
*     Positive definiteness
*

      IF (-VDOT(N,D,U) - VDOT(N,TMPN1,S) .GE. -SMALL) THEN

      
*     
*     SR1 update is skipped
*     

         ISR1 = 3

         IF (MCC .EQ. 0) THEN
            IFLAG = 0
            IF (ITYPE .EQ. 1) THEN
               IERR = -11
            ELSE
               CALL VNEG(N,GA,D)
            END IF
            RETURN
         END IF
         
         GO TO 200

      END IF
      
      STU = VDOT(N,S,U)

      SMTGA(INEW) = VDOT(N,S,GA)
      UMTGA(INEW) = VDOT(N,U,GA)

      
*
*     Convergence conditions
*

      IF (NNK .EQ. 1 .OR. MCC .LT. MC) GO TO 100

      IF (IFLAG .EQ. 1 .AND. (INEW .EQ. 1 .OR. INEW .EQ. MC)) GO TO 100

      
      ICON = 1


*
*     Calculation of matrix (UMTUM-R-R'+C) from previous iteration
*      

      DO 10 I=1,MN*(MN+1)/2
         TMPMAT(I)= GAMMA * UMTUM(I) - R(I)
 10   CONTINUE

      
*     
*     Solving TMPMC4 from TMPMAT*TMPMC4= GAMMA*UM'*GA-SM'*GA
*

      IF (IOLD .EQ. 1) THEN
         CALL SCDIFF(MN,GAMMA,UMTGA,SMTGA,TMPMC5)
         CALL CALQ(MN,MN,IOLD,TMPMAT,TMPMC4,TMPMC5,SMALL,0,IERR)
         IF (IERR .NE. 0) RETURN

      ELSE
         CALL SCDIFF(MCC,GAMMA,UMTGA,SMTGA,TMPMC5)
         CALL CALQ(MN,MCC,IOLD,TMPMAT,TMPMC4,TMPMC5,SMALL,0,IERR)
         IF (IERR .NE. 0) RETURN

      END IF

      
*
*     Computation of GDOLDG = -GA'*DM_(K-1)*GA
*
      
      IF (IOLD .EQ. 1 .OR. IOLD .EQ. 2) THEN
         CALL SCALEX(MN,GAMMA,TMPMC4(IOLD),TMPMC3(IOLD))
         CALL RECMAX(N,MN,SM((IOLD-1)*N+1),TMPMC4(IOLD),TMPN1)
         CALL SCDIFF(N,-GAMMA,GA,TMPN1,TMPN2)
         CALL RECMAX(N,MN,UM((IOLD-1)*N+1),TMPMC3(IOLD),TMPN1)
         CALL XSUMY(N,TMPN2,TMPN1,TMPN2)

      ELSE
         CALL SCALEX(MCC,GAMMA,TMPMC4,TMPMC3)
         CALL RECMAX(N,INEW-1,SM,TMPMC4,TMPN1)
         CALL SCDIFF(N,-GAMMA,GA,TMPN1,TMPN2)
         CALL RECMAX(N,MN-INEW+1,SM((IOLD-1)*N+1),TMPMC4(IOLD),
     &        TMPN1)
         CALL XDIFFY(N,TMPN2,TMPN1,TMPN2)
         CALL RECMAX(N,INEW-1,UM,TMPMC3,TMPN1)
         CALL XSUMY(N,TMPN2,TMPN1,TMPN2)
         CALL RECMAX(N,MN-INEW+1,UM((IOLD-1)*N+1),TMPMC3(IOLD),
     &        TMPN1)
         CALL XSUMY(N,TMPN2,TMPN1,TMPN2)
      END IF

      
      GDOLDG = VDOT(N,GA,TMPN2)

      
*
*     Calculation of the new canditate for search direction
*     Updates are not necessarily saved
*

      IF (IFLAG .EQ. 0) THEN
c         MN = MC
         IOLD = INEW + 2
         IF (IOLD .GT. MC+1) IOLD = IOLD - MC - 1
         
      ELSE
c         MN = MC - 1
         IOLD = INEW + 2
         IF (IOLD .GT. MC) IOLD = IOLD - MC
         
      END IF
      
      
*     
*     Update SM and UM
*

      CALL COPY2(N,S,SM((INEW-1)*N+1),U,UM((INEW-1)*N+1))

            
*     
*     Computation of SM'*U and UM'*U
*
      IF (IOLD .EQ. 1 .OR. IOLD .EQ. 2) THEN
         CALL RWAXV2(N,MN-1,SM((IOLD-1)*N+1),UM((IOLD-1)*N+1),U,U,
     &        TMPMC3(IOLD),TMPMC4(IOLD))

      ELSE
         CALL RWAXV2(N,INEW-1,SM,UM,U,U,TMPMC3,TMPMC4)
         CALL RWAXV2(N,MN-INEW,SM((IOLD-1)*N+1),UM((IOLD-1)*N+1),U,U,
     &        TMPMC3(IOLD),TMPMC4(IOLD))
         
      END IF
       
      TMPMC3(INEW) = STU
      TMPMC4(INEW) = VDOT(N,U,U)


*
*     Calculation of matrix (UMTUM-R-R'+C) without updating
*     matrices R, UMTUM and C
*      

      DO 20 I=1,MN*(MN+1)/2
         TMPMAT(I)= GAMMA * UMTUM(I) - R(I)
 20   CONTINUE

      DO 30 I=1,MN-1
         J=(I-1)*I/2+1
         K=I*(I+1)/2+2
         CALL COPY(I,TMPMAT(K),TMPMAT(J))
 30   CONTINUE

      
      CALL SCDIFF(MN+1,GAMMA,TMPMC4,TMPMC3,TMPMC5)

      IF (INEW .GE. MN) THEN
         CALL COPY(MN,TMPMC5(IOLD),
     &        TMPMAT((MN-1)*MN/2+1))

      ELSE
         CALL COPY(MN-INEW,TMPMC5(IOLD),
     &        TMPMAT((MN-1)*MN/2+1))
         CALL COPY(INEW,TMPMC5,
     &        TMPMAT((MN-1)*MN/2+MN-INEW+1))
      END IF
            
      
      IF (IFLAG .EQ. 0) THEN
         CALL SCDIFF(MC+1,GAMMA,UMTGA,SMTGA,TMPMC5)
         CALL CALQ(MN,MC+1,IOLD,TMPMAT,TMPMC5,TMPMC5,SMALL,IPRINT,
     &        IERR)

      ELSE
         CALL SCDIFF(MC,GAMMA,UMTGA,SMTGA,TMPMC5)
         CALL CALQ(MN,MC,IOLD,TMPMAT,TMPMC5,TMPMC5,SMALL,IPRINT,
     &        IERR)

      END IF
      IF (IERR .NE. 0) RETURN

      
*
*     Calculation of the new canditate for search direction D = -DM_K*GA
*     and computation of GDNEWG = -GA'*DM_K*GA
*
      
      IF (IOLD .EQ. 1 .OR. IOLD .EQ. 2) THEN
         CALL SCALEX(MN,GAMMA,TMPMC5(IOLD),TMPMC6(IOLD))
         CALL RECMAX(N,MN,SM((IOLD-1)*N+1),TMPMC5(IOLD),TMPN1)
         CALL SCDIFF(N,-GAMMA,GA,TMPN1,D)
         CALL RECMAX(N,MN,UM((IOLD-1)*N+1),TMPMC6(IOLD),TMPN1)
         CALL XSUMY(N,D,TMPN1,D)
         
      ELSE
         CALL SCALEX(MN+1,GAMMA,TMPMC5,TMPMC6)
         CALL RECMAX(N,INEW,SM,TMPMC5,TMPN1)
         CALL SCDIFF(N,-GAMMA,GA,TMPN1,D)
         CALL RECMAX(N,MN-INEW,SM((IOLD-1)*N+1),TMPMC5(IOLD),
     &        TMPN1)
         CALL XDIFFY(N,D,TMPN1,D)
         CALL RECMAX(N,INEW,UM,TMPMC6,TMPN1)
         CALL XSUMY(N,D,TMPN1,D)
         CALL RECMAX(N,MN-INEW,UM((IOLD-1)*N+1),TMPMC6(IOLD),
     &        TMPN1)
         CALL XSUMY(N,D,TMPN1,D)
      END IF


      GDNEWG = VDOT(N,GA,D)


*
*     Checking the convergence conditions
*      

      IF (GDNEWG - GDOLDG .LT. 0.0D+00) THEN
         ISR1 = 3
         CALL COPY(N,TMPN2,D)
            
         IF (IFLAG .EQ. 0) THEN
            IOLD = INEW + 1
            IF (IOLD .GT. MC+1) IOLD = IOLD - MC - 1
         
         ELSE
            IOLD = INEW + 1
            IF (IOLD .GT. MC) IOLD = IOLD - MC
         
         END IF
      ELSE
         ISR1 = 1

         
*     
*     Update SM'*GP and UM'*GP
*     

         SMTGP(INEW) = VDOT(N,S,GP)
         UMTGP(INEW) = VDOT(N,U,GP)

            
*     
*     Computation of SM'*S, and UM'*S
*

         IF (INEW .GE. MN) THEN
            CALL RWAXV2(N,MN,SM((INEW-MN)*N+1),
     &           UM((INEW-MN)*N+1),S,S,TMPMC5(IOLD),TMPMC6(IOLD))

         ELSE
            CALL RWAXV2(N,INEW,SM,UM,S,S,TMPMC5,TMPMC6)
            CALL RWAXV2(N,MN-INEW,SM((IOLD-1)*N+1),UM((IOLD-1)*N+1),
     &           S,S,TMPMC5(IOLD),TMPMC6(IOLD))
         END IF               


*         
*     Update R, LM, SMTSM, and UMTUM
*

*     Shift the old values
         DO 40 I=1,MN-1
            J=(I-1)*I/2+1
            K=I*(I+1)/2+2
            CALL COPY4(I,R(K),R(J),LM(K),LM(J),SMTSM(K),SMTSM(J),
     &           UMTUM(K),UMTUM(J))
 40      CONTINUE

*     Add the new values

         IF (INEW .GE. MN) THEN
            CALL COPY4(MN,TMPMC3(IOLD),R((MN-1)*MN/2+1),
     &           TMPMC6(IOLD),LM((MN-1)*MN/2+1),
     &           TMPMC5(IOLD),SMTSM((MN-1)*MN/2+1),
     &           TMPMC4(IOLD),UMTUM((MN-1)*MN/2+1))

         ELSE
            CALL COPY4(MN-INEW,TMPMC3(IOLD),R((MN-1)*MN/2+1)
     &           ,TMPMC6(IOLD),LM((MN-1)*MN/2+1)
     &           ,TMPMC5(IOLD),SMTSM((MN-1)*MN/2+1)
     &           ,TMPMC4(IOLD),UMTUM((MN-1)*MN/2+1))
               
            CALL COPY4(INEW,TMPMC3,R((MN-1)*MN/2+MN-INEW+1),
     &           TMPMC6,LM((MN-1)*MN/2+MN-INEW+1),
     &           TMPMC5,SMTSM((MN-1)*MN/2+MN-INEW+1),
     &           TMPMC4,UMTUM((MN-1)*MN/2+MN-INEW+1))

         END IF
         LM((MN+1)*MN/2) = 0.0D+00
         
          
*
*     Update C
*

         C(INEW) = STU

         
         INEW = INEW + 1
         IF (INEW .GT. MC + 1) INEW = 1
         IF (IFLAG .EQ. 0 .AND. MCC .LT. MC + 1) MCC = MCC + 1

      END IF

      GO TO 200

      
 100  CONTINUE

      ISR1 = 1

      CALL INDIC1(MC,MCC,MN,INEW,IOLD,IFLAG,IFLAG2,1)        

      IF (IFLAG2 .EQ. 1 .AND. IOLD .EQ. 2) THEN
         SMTGA(INEW) = SMTGA(1)
         UMTGA(INEW) = UMTGA(1)
      END IF

            
*     
*     Update SM and UM
*

      CALL COPY2(N,S,SM((INEW-1)*N+1),U,UM((INEW-1)*N+1))

            
*         
*     Update SM'*GP and UM'*GP
*

      SMTGP(INEW) = VDOT(N,S,GP)
      UMTGP(INEW) = VDOT(N,U,GP)

      
*     
*     Computation of SM'*U, UM'*U, SM'*S, and UM'*S.
*

      IF (IOLD .EQ. 1 .OR. IOLD .EQ. 2) THEN
         CALL RWAXV2(N,MN-1,SM((IOLD-1)*N+1),UM((IOLD-1)*N+1),U,U,
     &        TMPMC3(IOLD),TMPMC4(IOLD))
         CALL RWAXV2(N,MN,SM((IOLD-1)*N+1),UM((IOLD-1)*N+1),S,S,
     &        TMPMC5(IOLD),TMPMC6(IOLD))

      ELSE
         CALL RWAXV2(N,INEW-1,SM,UM,U,U,TMPMC3,TMPMC4)
         CALL RWAXV2(N,INEW,SM,UM,S,S,TMPMC5,TMPMC6)
         CALL RWAXV2(N,MN-INEW,SM((IOLD-1)*N+1),UM((IOLD-1)*N+1),
     &        U,U,TMPMC3(IOLD),TMPMC4(IOLD))
         CALL RWAXV2(N,MN-INEW,SM((IOLD-1)*N+1),UM((IOLD-1)*N+1),
     &        S,S,TMPMC5(IOLD),TMPMC6(IOLD))
      END IF

      TMPMC3(INEW) = STU
      TMPMC4(INEW) = VDOT(N,U,U)

      
*         
*     Update R, LM, SMTSM, and UMTUM
*

      IF (MCC .GE. MC .AND. IFLAG2 .NE. 1) THEN
         

*     Shift the old values
         
         DO 110 I=1,MN-1
            J=(I-1)*I/2+1
            K=I*(I+1)/2+2
            CALL COPY4(I,R(K),R(J),LM(K),LM(J),SMTSM(K),SMTSM(J),
     &           UMTUM(K),UMTUM(J))
 110     CONTINUE
      END IF

*     Add the new values
      
      IF (INEW .GE. MN) THEN
         CALL COPY4(MN,TMPMC3(IOLD),R((MN-1)*MN/2+1),
     &        TMPMC6(IOLD),LM((MN-1)*MN/2+1),
     &        TMPMC5(IOLD),SMTSM((MN-1)*MN/2+1),
     &        TMPMC4(IOLD),UMTUM((MN-1)*MN/2+1))

      ELSE
         CALL COPY4(MN-INEW,TMPMC3(IOLD),R((MN-1)*MN/2+1)
     &        ,TMPMC6(IOLD),LM((MN-1)*MN/2+1)
     &        ,TMPMC5(IOLD),SMTSM((MN-1)*MN/2+1)
     &        ,TMPMC4(IOLD),UMTUM((MN-1)*MN/2+1))
               
         CALL COPY4(INEW,TMPMC3,R((MN-1)*MN/2+MN-INEW+1),
     &        TMPMC6,LM((MN-1)*MN/2+MN-INEW+1),
     &        TMPMC5,SMTSM((MN-1)*MN/2+MN-INEW+1),
     &        TMPMC4,UMTUM((MN-1)*MN/2+MN-INEW+1))
      END IF

      LM((MN+1)*MN/2) = 0.0D+00

      
*
*     Update C
*

      C(INEW) = STU

         
      INEW = INEW + 1
      IF (INEW .GT. MC + 1) INEW = 1
      IF (IFLAG .EQ. 0 .AND. MCC .LT. MC + 1) MCC = MCC + 1
      
      
 200  CONTINUE
      
      IF (ITYPE .EQ. 1) THEN


*     
*     Generalized Cauchy point.
*
      
         CALL CPSR1(N,MC,MN,IOLD,IFLAG,IPRINT,
     &        X,XCP,XL,XU,IB,IACT,GA,TMPN2,TMPN1,
     &        SM,UM,SMTGA,UMTGA,SMTSM,LM,C,TMPMAT,
     &        TMPMC3,TMPMC4,TMPMC5,TMPMC6,GAMMA,SMALL,IERR)
         IF (IERR .NE. 0) RETURN

*
*     Determination of the active variables.
*

         CALL ACTVAR(N,NACT,IACT,IB,XCP,XL,XU,SMALL)

      END IF


*
*     Subspace minimization and direction finding
*      

      IF (NACT .NE. N) THEN

         CALL SSSR1(N,NACT,MC,MN,INEW,IOLD,IFLAG,ICON,ITYPE,
     &        X,XL,XU,XCP,GA,D,
     &        SM,UM,R,UMTUM,SMTGA,UMTGA,TMPMC5,TMPMAT,TMPMA2,TMPMC3,
     &        TMPMC4,TMPN1,TMPN2,GAMMA,SMALL,IACT,IB,IPRINT,IERR)
         IF (IERR .NE. 0) RETURN

      ELSE

         CALL XDIFFY(N,XCP,X,D)
         
      END IF
      
      RETURN
      END
      
      
************************************************************************
*
*     * SUBROUTINE SSSR1 *
*
*      
*     * Purpose *
*      
*     Subspace minimization and computation of the search direction by
*     the limited memory SR1 update.
*
*      
*     * Calling sequence *
*     
*     CALL SSSR1(N,NACT,MC,MN,INEW,IOLD,IFLAG,ICON,ITYPE,
*    &     X,XL,XU,XCP,GA,D,SM,UM,R,UMTUM,SMTGA,UMTGA,MWTGA,TMPMAT,
*    &     TMPMA2,TMPMC3,TMPMC4,TMPN1,TMPN2,GAMMA,SMALL,IACT,IB,IPRINT
*    &    ,IERR)
*     
*      
*     * Parameters *
*
*     II  N               Number of variables.
*     II  NACT            Number of active variables.
*     II  MC              Declared number of stored corrections.
*     II  MN              Current number of stored corrections.
*     II  INEW            Index for circular arrays.
*     II  IOLD            Index of the oldest corrections.
*     II  IFLAG           Index for adaptive version:
*                           0  - Maximum number of stored corrections
*                                  has not been changed at previous
*                                  iteration.
*                           1  - Maximum number of stored corrections
*                                  has been changed at previous
*                                  iteration.
*     II  ICON            Parameter for convergence test:
*                           0  - additional test has not been executed,
*                           1  - additional test has been executed.
*     II  ITYPE           Type of problem:
*                           0  - problem is unbounded,
*                           1  - constrained problem.
*     II  IACT(N)         Index set of active and free variables:
*                           for I=1,...,NACT, IACT(I) are the indices
*                               of active variables (-J, if
*                               XCP(J)=XL(J) and +J, otherwise) and
*                           for I=NACT+1,...,N, IACT(I) are the indices
*                               of free variables.
*     II  IB(N)           Type of bound constraints:
*                           0  - X(I) is unbounded,
*                           1  - X(I) has only a lower bound,
*                           2  - X(I) has both lower and upper bounds,
*                           3  - X(I) has only an upper bound. 
*     RI  X(N)            Vector of variables. 
*     RI  XCP(N)          Generalized Chauchy point.
*     RI  XL(N)           Lower bounds for variables. 
*     RI  XU(N)           Upper bounds for variables. 
*     RI  GA(N)           Current aggregate subgradient of the objective
*                           function.
*     RU  D(N)            Search direction.
*                           If ICON=1, then, on input, D=-DM GA.
*     RI  SMTGA(MC+1)     Vector SMTGP = SM'*GA.
*     RI  UMTGA(MC+1)     Vector UMTGP = UM'*GA.
*     RA  MWTGA(MC+1)     Auxiliary vector.
*     RI  SM(N*(MC+1))    Matrix whose columns are stored corrections.
*     RI  UM(N*(MC+1))    Matrix whose columns are stored subgradient
*                           differences.
*     RI  R((MN+1)*MN/2)  Upper triangular matrix stored
*                           columnwise in the one-dimensional array.
*     RI  UMTUM((MN+1)*MN/2)  Matrix UMTUM = UM' * UM.
*     RA  TMPMAT((MN+1)*MN/2)  Auxiliary matrix.
*     RA  TMPMA2((MN+1)*MN/2)  Auxiliary matrix.
*     RA  TMPMC#(MC+1)    Auxiliary arrays; # = 3,4.
*     RA  TMPN#(N)        Auxiliary array; # = 1,2.
*     RI  GAMMA           Scaling parameter.
*     RI  SMALL           Small positive value.
*     II  IPRINT          Printout specification.
*     IO  IERR            Error indicador: 
*                             0  - Everything is ok.
*                            -2  - Error in LINEQ.
*
*
*     * Subprograms used *
*      
*     S   CALQ            Solving X from linear equation A*X=Y.
*     S   LINEQ           Solving X from linear equation A*X=Y, where
*                           A is given in factorized form.
*     S   RECMAX          Multiplication of a vector by a dense
*                           rectangular matrix.
*     S   XDIFFY          Difference of two vectors.
*     S   SCALEX          Scaling a vector.
*     S   XSUMY           Sum of two vectors.
*     S   SCDIFF          Difference of the scaled vector and a vector.
*     S   SCSUM           Sum of a vector and the scaled vector.
*
*     All the (MC+1)-vectors are stored in a circular order controlled by
*     the parameter point INEW.
*
*      
      
      SUBROUTINE SSSR1(N,NACT,MC,MN,INEW,IOLD,IFLAG,ICON,ITYPE,
     &     X,XL,XU,XCP,GA,D,
     &     SM,UM,R,UMTUM,SMTGA,UMTGA,MWTGA,TMPMAT,TMPMA2,TMPMC3,TMPMC4,
     &     TMPN1,TMPN2,GAMMA,SMALL,IACT,IB,IPRINT,IERR)
      
*     Scalar Arguments
      INTEGER N,NACT,MC,MN,INEW,IOLD,IFLAG,ICON,ITYPE,IPRINT,IERR
      DOUBLE PRECISION GAMMA,SMALL
      
*     Array Arguments
      INTEGER IACT(*),IB(*)
      DOUBLE PRECISION X(*),XL(*),XU(*),XCP(*),GA(*),D(*),
     &     SM(*),UM(*),SMTGA(*),UMTGA(*),MWTGA(*),R(*),UMTUM(*),
     &     TMPMAT(*),TMPMA2(*),TMPMC3(*),TMPMC4(*),TMPN1(*),TMPN2(*)
      
*     Local Scalars
      INTEGER I,IJ,J,K,LI,LJ,IFREE,MCC,IND
      DOUBLE PRECISION ALPHA,DFREE,SIGNUM 

*     Intrinsic Functions
      INTRINSIC ABS,SIGN,MIN,DBLE

*     External Subroutines
      EXTERNAL SCDIFF,CALQ,SCALEX,RECMAX,XSUMY,XDIFFY,LINEQ

*     If NACT=0 and ICON=1 then D= -DM*GA has already been calculated.
      IF (NACT .EQ. 0 .AND. ICON .EQ. 1) GOTO 600

      IF (IOLD .EQ. 1) THEN
         MCC = MN
      ELSE IF (IFLAG .EQ. 0) THEN
         MCC = MC + 1
      ELSE
         MCC = MC
      END IF

      
      IF (NACT .EQ. 0) THEN
         

*
*     Calculation of matrix TMPMAT = UMTUM - R - R' + C
*

         DO 10 I=1,MN*(MN+1)/2
            TMPMAT(I)= GAMMA * UMTUM(I) - R(I)
 10      CONTINUE


*     
*     Solving MWTGA from TMPMAT*MWTGA= GAMMA*UM'*GA - SM'*GA
*

         CALL SCDIFF(MCC,GAMMA,UMTGA,SMTGA,MWTGA)
         CALL CALQ(MN,MCC,IOLD,TMPMAT,MWTGA,MWTGA,SMALL,IPRINT,IERR)
         IF (IERR .NE. 0) RETURN

      
*
*     Computation of the search direction D
*

         IF (IOLD .EQ. 1 .OR. IOLD .EQ. 2) THEN
            CALL SCALEX(MN,GAMMA,MWTGA(IOLD),TMPMC3(IOLD))
            CALL RECMAX(N,MN,SM((IOLD-1)*N+1),MWTGA(IOLD),TMPN1)
            CALL SCDIFF(N,-GAMMA,GA,TMPN1,D)
            CALL RECMAX(N,MN,UM((IOLD-1)*N+1),TMPMC3(IOLD),TMPN1)
            CALL XSUMY(N,D,TMPN1,D)
         
         ELSE
            CALL SCALEX(MCC,GAMMA,MWTGA,TMPMC3)
            CALL RECMAX(N,INEW-1,SM,MWTGA,TMPN1)
            CALL SCDIFF(N,-GAMMA,GA,TMPN1,D)
            CALL RECMAX(N,MN-INEW+1,SM((IOLD-1)*N+1),MWTGA(IOLD),
     &           TMPN1)
            CALL XDIFFY(N,D,TMPN1,D)
            CALL RECMAX(N,INEW-1,UM,TMPMC3,TMPN1)
            CALL XSUMY(N,D,TMPN1,D)
            CALL RECMAX(N,MN-INEW+1,UM((IOLD-1)*N+1),TMPMC3(IOLD),
     &           TMPN1)
            CALL XSUMY(N,D,TMPN1,D)
         END IF
      
         
      ELSE
      

*
*     Computation of the intermediate vector TMPN1= -A'DM GA - A'(XCP-X)
*

         IF (ICON .EQ. 1) THEN
c     Note that D=-DM GA

            DO 100 I=1,NACT
               IND=ABS(IACT(I))
               SIGNUM = SIGN(1.0D+00,DBLE(IACT(I)))
               
               TMPN1(I) = SIGNUM*D(IND) - SIGNUM*(XCP(IND)-X(IND))
 100        CONTINUE
         ELSE


*     
*     Calculation of matrix TMPMAT = GAMMA*UMTUM - R - R' + C
*
            DO 200 I=1,MN*(MN+1)/2
               TMPMAT(I)= GAMMA * UMTUM(I) - R(I)
 200        CONTINUE

*     
*     Solving MWTGA from TMPMAT*MWTGA= GAMMA*UM'*GA - SM'*GA
*
            CALL SCDIFF(MCC,GAMMA,UMTGA,SMTGA,MWTGA)
            CALL CALQ(MN,MCC,IOLD,TMPMAT,MWTGA,MWTGA,SMALL
     &           ,IPRINT,IERR)
            IF (IERR .NE. 0) RETURN

            DO 210 I=1,NACT
               TMPN1(I)=0.0D+00
               IND=ABS(IACT(I))
               SIGNUM = SIGN(1.0D+00,DBLE(IACT(I)))
               TMPN2(I) = SIGNUM*(XCP(IND)-X(IND))

               IF (IOLD .EQ. 1) THEN
                  DO 211 J=1,MN
                     TMPN1(I) = TMPN1(I) + SIGNUM * MWTGA(J) * 
     &                    (GAMMA*UM((J-1)*N+IND) - SM((J-1)*N+IND))
 211              CONTINUE
               ELSE
                  DO 212 J = 1, INEW-1
                     TMPN1(I) = TMPN1(I) + SIGNUM * MWTGA(J) *
     &                    (GAMMA*UM((J-1)*N+IND) - SM((J-1)*N+IND))
 212              CONTINUE
                  DO 213 J = IOLD, MCC
                     TMPN1(I)=TMPN1(I) + SIGNUM * MWTGA(J) *
     &                    (GAMMA*UM((J-1)*N+IND) - SM((J-1)*N+IND))
 213              CONTINUE
                  
               END IF
               
               TMPN1(I) = -SIGNUM*GAMMA*GA(IND) + TMPN1(I) - TMPN2(I)
 210        CONTINUE
         END IF
         

*
*     Computation of Lagrange multipiliers TMPN1 = (A'DM A)^{-1} TMPN1.
*
         
*     TMPMC3=(GAMMA*UM - SM)' A TMPN1 

         IF (IOLD .EQ. 1) THEN
            DO 220 J = 1, MN
               TMPMC3(J)= 0.0D+00
               DO 221 I=1,NACT
                  IND=ABS(IACT(I))
                  SIGNUM = SIGN(1.0D+00,DBLE(IACT(I)))
                  TMPMC3(J)=TMPMC3(J)+SIGNUM*(GAMMA*UM((J-1)*N+IND)-
     &                 SM((J-1)*N+IND))*TMPN1(I)
 221           CONTINUE
 220        CONTINUE

         ELSE 

            DO 230 J = 1, INEW-1
               TMPMC3(J)= 0.0D+00
               DO 231 I=1,NACT
                  IND=ABS(IACT(I))
                  SIGNUM = SIGN(1.0D+00,DBLE(IACT(I)))
                  TMPMC3(J)=TMPMC3(J)+SIGNUM*(GAMMA*UM((J-1)*N+IND)-
     &                 SM((J-1)*N+IND))*TMPN1(I)
 231           CONTINUE
 230        CONTINUE
            DO 240 J = IOLD, MCC
               TMPMC3(J)= 0.0D+00
               DO 241 I=1,NACT
                  IND=ABS(IACT(I))
                  SIGNUM = SIGN(1.0D+00,DBLE(IACT(I)))
                  TMPMC3(J)=TMPMC3(J)+SIGNUM*(GAMMA*UM((J-1)*N+IND)-
     &                 SM((J-1)*N+IND))*TMPN1(I)
 241           CONTINUE
 240        CONTINUE
            
         END IF
         
*
*     Formulation of  TMPMA2 = TMPMAT - 1/GAMMA * W' A A' W, where
*     W=GAMMA UM - SM and TMPMAT is the middle matrix in compact SR1
*     formula.
*            

         IJ=0
         DO 250 I = 1,MN
            LI=I+IOLD-1
            IF (LI .GT. MCC) LI=LI-MCC
            DO 251 J=1,I
               LJ=J+IOLD-1
               IF (LJ .GT. MCC) LJ=LJ-MCC
               IJ=IJ+1
               
               TMPMA2(IJ)=0.0D+00
               DO 252 K = 1,NACT
                  IND = ABS(IACT(K))
                  TMPMA2(IJ) = TMPMA2(IJ) +
     &                 (GAMMA*UM((LJ-1)*N+IND) - SM((LJ-1)*N+IND)) *
     &                 (GAMMA*UM((LI-1)*N+IND) - SM((LI-1)*N+IND))
 252           CONTINUE
               TMPMA2(IJ)= - TMPMA2(IJ)/GAMMA + GAMMA*UMTUM(IJ) - R(IJ)
 251        CONTINUE
 250     CONTINUE
           
*     
*     Solving TMPMC3 from TMPMA2*TMPMC3 = (GAMMA*UM - SM)' A TMPN1 
*
         CALL CALQ(MN,MCC,IOLD,TMPMA2,TMPMC3,TMPMC3,SMALL
     &        ,IPRINT,IERR)
         IF (IERR .NE. 0) RETURN
         
         
         DO 260 I=1,NACT
            TMPN2(I)=0.0D+00
            IND=ABS(IACT(I))
            SIGNUM = SIGN(1.0D+00,DBLE(IACT(I)))
            IF (IOLD .EQ. 1) THEN
               DO 261 J=1,MN
                  TMPN2(I) = TMPN2(I) + SIGNUM * TMPMC3(J) *
     &                 (GAMMA*UM((J-1)*N+IND) - SM((J-1)*N+IND))
 261           CONTINUE
            ELSE
               DO 262 J = 1, INEW-1
                  TMPN2(I) = TMPN2(I) + SIGNUM * TMPMC3(J) *
     &                 (GAMMA*UM((J-1)*N+IND) - SM((J-1)*N+IND))
 262           CONTINUE
               DO 263 J = IOLD, MCC
                  TMPN2(I) = TMPN2(I) + SIGNUM * TMPMC3(J) *
     &                 (GAMMA*UM((J-1)*N+IND) - SM((J-1)*N+IND))
 263           CONTINUE

            END IF
            TMPN1(I) = (TMPN1(I) + TMPN2(I)/GAMMA) /GAMMA
 260     CONTINUE
            
      
*     
*     Computation of the search direction D = -DM(A TMPN1 + GA)
*
         IF (ICON .EQ. 1) THEN
c     Note that D=-DM GA

*     
*     Calculation of matrix TMPMAT = GAMMA*UMTUM - R - R' + C
*
            DO 300 I=1,MN*(MN+1)/2
               TMPMAT(I)= GAMMA * UMTUM(I) - R(I)
 300        CONTINUE
            
*     
*     Solving TMPMC3 from TMPMAT*TMPMC3= (GAMMA*UM-SM)'*A TMPN1
*     

            IF (IOLD .EQ. 1) THEN
               DO 310 J = 1, MN
                  TMPMC3(J)= 0.0D+00
                  DO 311 I=1,NACT
                     IND=ABS(IACT(I))
                     SIGNUM = SIGN(1.0D+00,DBLE(IACT(I)))
                     TMPMC3(J)=TMPMC3(J)+SIGNUM*(GAMMA*UM((J-1)*N+IND)-
     &                    SM((J-1)*N+IND))*TMPN1(I)
 311              CONTINUE
 310           CONTINUE
               
               CALL CALQ(MN,MN,IOLD,TMPMAT,TMPMC3,TMPMC3,SMALL,
     &              IPRINT,IERR)
               IF (IERR .NE. 0) RETURN
               
            ELSE 

               DO 320 J = 1, INEW-1
                  TMPMC3(J)= 0.0D+00
                  DO 321 I=1,NACT
                     IND=ABS(IACT(I))
                     SIGNUM = SIGN(1.0D+00,DBLE(IACT(I)))
                     TMPMC3(J)=TMPMC3(J)+SIGNUM*(GAMMA*UM((J-1)*N+IND)-
     &                    SM((J-1)*N+IND))*TMPN1(I)
 321              CONTINUE
 320           CONTINUE
               DO 330 J = IOLD, MCC
                  TMPMC3(J)= 0.0D+00
                  DO 331 I=1,NACT
                     IND=ABS(IACT(I))
                     SIGNUM = SIGN(1.0D+00,DBLE(IACT(I)))
                     TMPMC3(J)=TMPMC3(J)+SIGNUM*(GAMMA*UM((J-1)*N+IND)-
     &                    SM((J-1)*N+IND))*TMPN1(I)
 331              CONTINUE
 330           CONTINUE
               
               CALL CALQ(MN,MCC,IOLD,TMPMAT,TMPMC3,TMPMC3,SMALL,
     &              IPRINT,IERR)
               IF (IERR .NE. 0) RETURN
            END IF
            
*
*     Computation of the search direction D
*
            DO 340 I=1,NACT
               IND=ABS(IACT(I))
               SIGNUM = SIGN(1.0D+00,DBLE(IACT(I)))
               
               D(IND) = D(IND) - GAMMA*SIGNUM*TMPN1(I)
 340        CONTINUE

            
            IF (IOLD .EQ. 1 .OR. IOLD .EQ. 2) THEN
               CALL SCALEX(MN,GAMMA,TMPMC3(IOLD),TMPMC4(IOLD))
               CALL RECMAX(N,MN,SM((IOLD-1)*N+1),TMPMC3(IOLD),TMPN1)
               CALL RECMAX(N,MN,UM((IOLD-1)*N+1),TMPMC4(IOLD),TMPN2)
               DO 350 I=1,N
                  D(I) = D(I) + TMPN2(I) - TMPN1(I)
 350           CONTINUE
               
            ELSE
               CALL SCALEX(MCC,GAMMA,TMPMC3,TMPMC4)
               CALL RECMAX(N,INEW-1,SM,TMPMC3,TMPN1)
               CALL RECMAX(N,MN-INEW+1,SM((IOLD-1)*N+1),TMPMC3(IOLD),
     &              TMPN2)
               DO 360 I=1,N
                  D(I) = D(I) - TMPN1(I) - TMPN2(I)
 360           CONTINUE
               
               CALL RECMAX(N,INEW-1,UM,TMPMC4,TMPN1)
               CALL RECMAX(N,MN-INEW+1,UM((IOLD-1)*N+1),TMPMC4(IOLD),
     &              TMPN2)
               DO 370 I=1,N
                  D(I) = D(I) + TMPN1(I) + TMPN2(I)
 370           CONTINUE
            END IF
            
         ELSE
            
   
*     
*     Solving TMPMC3 from TMPMAT*TMPMC3= (GAMMA*UM-SM)'*A TMPN1
*     Note that TMPMAT is a factorization of the middle matrix in SR1
*     updating formula.
*     

            IF (IOLD .EQ. 1) THEN
               DO 400 J = 1, MN
                  TMPMC3(J)= 0.0D+00
                  DO 401 I=1,NACT
                     IND=ABS(IACT(I))
                     SIGNUM = SIGN(1.0D+00,DBLE(IACT(I)))
                     TMPMC3(J)=TMPMC3(J)+SIGNUM*(GAMMA*UM((J-1)*N+IND)-
     &                    SM((J-1)*N+IND))*TMPN1(I)
 401              CONTINUE
 400           CONTINUE
               
               CALL LINEQ(MN,MN,1,TMPMAT,TMPMC3,TMPMC3,SMALL,IERR)
               IF (IERR .NE. 0) RETURN
               
               
               DO 410 J = 1, MN
                  TMPMC3(J) = TMPMC3(J) + MWTGA(J)
 410           CONTINUE
               
            ELSE 

               DO 420 J = 1, INEW-1
                  TMPMC3(J)= 0.0D+00
                  DO 421 I=1,NACT
                     IND=ABS(IACT(I))
                     SIGNUM = SIGN(1.0D+00,DBLE(IACT(I)))
                     TMPMC3(J)=TMPMC3(J)+SIGNUM*(GAMMA*UM((J-1)*N+IND)-
     &                    SM((J-1)*N+IND))*TMPN1(I)
 421              CONTINUE
 420           CONTINUE
               DO 430 J = IOLD, MCC
                  TMPMC3(J)= 0.0D+00
                  DO 431 I=1,NACT
                     IND=ABS(IACT(I))
                     SIGNUM = SIGN(1.0D+00,DBLE(IACT(I)))
                     TMPMC3(J)=TMPMC3(J)+SIGNUM*(GAMMA*UM((J-1)*N+IND)-
     &                    SM((J-1)*N+IND))*TMPN1(I)
 431              CONTINUE
 430           CONTINUE

               CALL LINEQ(MN,MCC,IOLD,TMPMAT,TMPMC3,TMPMC3,SMALL,
     &              IERR)
               IF (IERR .NE. 0) RETURN

               DO 440 J = 1, INEW-1
                  TMPMC3(J) = TMPMC3(J) + MWTGA(J)
 440           CONTINUE
               DO 450 J = IOLD, MCC
                  TMPMC3(J) = TMPMC3(J) + MWTGA(J)
 450           CONTINUE
            END IF

*
*     Computation of the search direction D
*
            
            IF (IOLD .EQ. 1 .OR. IOLD .EQ. 2) THEN
               CALL SCALEX(MN,GAMMA,TMPMC3(IOLD),TMPMC4(IOLD))
               CALL RECMAX(N,MN,SM((IOLD-1)*N+1),TMPMC3(IOLD),D)
               CALL RECMAX(N,MN,UM((IOLD-1)*N+1),TMPMC4(IOLD),TMPN2)
               DO 460 I=1,N
                  D(I) = -GAMMA*GA(I) + TMPN2(I) - D(I)
 460           CONTINUE
               
               DO 470 I=1,NACT
                  IND=ABS(IACT(I))
                  SIGNUM = SIGN(1.0D+00,DBLE(IACT(I)))
                  
                  D(IND) = D(IND) - GAMMA*SIGNUM*TMPN1(I)
 470           CONTINUE
               
            ELSE

               CALL SCALEX(MCC,GAMMA,TMPMC3,TMPMC4)
               CALL RECMAX(N,INEW-1,SM,TMPMC3,D)
               CALL RECMAX(N,MN-INEW+1,SM((IOLD-1)*N+1),TMPMC3(IOLD),
     &              TMPN2)
               DO 480 I=1,N
                  D(I) = -GAMMA*GA(I) - D(I) - TMPN2(I)
 480           CONTINUE
               DO 490 I=1,NACT
                  IND=ABS(IACT(I))
                  SIGNUM = SIGN(1.0D+00,DBLE(IACT(I)))
                  
                  D(IND) = D(IND) - GAMMA*SIGNUM*TMPN1(I)

 490           CONTINUE
              
               CALL RECMAX(N,INEW-1,UM,TMPMC4,TMPN1)
               CALL RECMAX(N,MN-INEW+1,UM((IOLD-1)*N+1),TMPMC4(IOLD),
     &              TMPN2)
               DO 500 I=1,N
                  D(I) = D(I) + TMPN1(I) + TMPN2(I)
 500           CONTINUE
            END IF

         END IF
      END IF


 600  CONTINUE


*     
*     Backtrack to the feasible region if X + D violates bounds
*     (among free variables).
*

      IF (ITYPE .EQ. 1) THEN

         ALPHA = 1.0D+00
         DO 610 I = NACT+1,N
            IFREE=IACT(I)
            DFREE=D(IFREE)
            IF (IB(IFREE) .NE. 0) THEN
               IF (IB(IFREE) .LE. 2 .AND. 
     &              X(IFREE) + DFREE .LT. XL(IFREE)) THEN

*     Otherwise the lower bound is never met
 
                  IF (ABS(X(IFREE)-XCP(IFREE)+DFREE) .GT. SMALL) THEN
                     ALPHA=MIN(ALPHA,(XL(IFREE)-XCP(IFREE))/
     &                    (X(IFREE)-XCP(IFREE)+DFREE))
                  END IF

               ELSE IF (IB(IFREE) .GE. 2 .AND. 
     &                 X(IFREE) + DFREE  .GT. XU(IFREE)) THEN

*     Otherwise the upper bound is never met
 
                  IF (ABS(X(IFREE)-XCP(IFREE)+DFREE) .GT. SMALL) THEN
                     ALPHA=MIN(ALPHA,(XU(IFREE)-XCP(IFREE))/
     &                    (X(IFREE)-XCP(IFREE)+DFREE))
                  END IF
                  
               END IF
               
            END IF
 610     CONTINUE


         IF (ALPHA .LT. 1.0D+00) THEN
            
            DO 620 I=NACT+1,N
               IFREE=IACT(I)
               D(IFREE) = XCP(IFREE) - X(IFREE) +
     &              ALPHA*(X(IFREE) - XCP(IFREE) + D(IFREE))
 620        CONTINUE
         END IF

         ALPHA=0.0D+00
         J=0
         DO 630 I=1,NACT
            IND=ABS(IACT(I))
            D(IND)=XCP(IND)-X(IND)
 630     CONTINUE

      END IF

      
      RETURN
      END
      
      
      
************************************************************************
*
*     * SUBROUTINE CPSR1 *
*
*      
*     * Purpose *
*      
*     Computation of generalized Cauchy point by the limited memory
*     SR1 update.
*
*      
*     * Calling sequence *
*     
*     CALL CPSR1(N,MC,MN,IOLD,IFLAG,IPRINT,
*    &     X,XCP,XL,XU,IB,IT,GA,DC,T,
*    &     SM,UM,SMTGA,UMTGA,SMTSM,LM,C,TMPM2,
*    &     PAUX,CAUX,VAUX,WAUX,GAMMA,SMALL,IERR)
*
*      
*     * Parameters *
*     
*     II  N               Number of variables.
*     II  MC              Declared Number of stored corrections.
*     II  MN              Current number of stored corrections.
*     II  IOLD            Index for circular arrays.
*     II  IFLAG           Index for adaptive version:
*                             0  - Maximum number of stored corrections
*                                    has not been changed at previous
*                                    iteration.
*                             1  - Maximum number of stored corrections
*                                    has been changed at previous
*                                    iteration.
*     RI  X(N)            Vector of variables. 
*     RO  XCP(N)          Generalized Cauchy point.
*     RI  XL(N)           Lower bounds for variables. 
*     RI  XU(N)           Upper bounds for variables. 
*     II  IB(N)           Type of bound constraints:
*                             0  - X(I) is unbounded,
*                             1  - X(I) has only a lower bound,
*                             2  - X(I) has both lower and upper bounds,
*                             3  - X(I) has only an upper bound. 
*     RI  GA(N)           Current aggregate subgradient of the objective
*                           function.
*     RI  SM(N*(MN+1))    Matrix whose columns are stored corrections.
*     RI  UM(N*(MN+1))    Matrix whose columns are stored subgradient
*                           differences.
*     RI  SMTGA(MN+1)     Vector SMTGA = SM'*GA.
*     RI  UMTGA(MN+1)     Vector UMTGA = UM'*GA.
*     RI  SMTSM((MN+1)*MN/2)  Matrix SMTSM = SM' * SM.
*     RI  LM((MN+1)*MN/2) Lower triangular matrix stored rowwise
*                           in the one-dimensional array.
*     RI  CDIAG(MN+1)     Diagonal matrix.
*     RI  GAMMA           Scaling parameter.
*     RA  TMPM2((MN+1)*MN/2) Matrix TMPM2= LM+LM+CDIAG - 1/GAMMA SMTSM or
*                           its factorization.
*     RA  DC(N)           Auxiliary array used to store Cauchy
*                           direction P(X-T*GA)-X.
*     RA  T(N)            Auxiliary array used to store the
*                           breakpoints.
*     IA  IT(N)           Index set of breakpoints:
*                             for I=1,...,NLEFT, IT(I) are the indices
*                                 of breakpoints that have not been
*                                 encountered,
*                             for i=NLEFT+1,...,NBREAK, IT(I) are the
*                                 indices of encountered breakpoints and
*                             for i=IFREE,...,N, IT(I) are the indices
*                                 of variables that have no bound
*                                 constraints along the search direction.
*     RA  PAUX(MN)        Auxiliary array used to store the vector
*                           (UM-1/GAMMA*(SM))'*DC.
*     RA  CAUX(MN)        Auxiliary array used to store the vector
*                           (UM-1/GAMMA*(SM))'*(XCP-X).
*     RA  VAUX(MN)        Auxiliary array.
*     RA  WAUX(MN)        Auxiliary array.
*     RI  SMALL           Small positive value.
*     II  IPRINT          Printout specification.
*     IO  IERR            Error indicador: 
*                             0  - Everything is ok.
*                            -2  - Error in LINEQ.
*                            -7  - division with zero.
*     
*
*     * Subprograms used *
*      
*     RF  VDOT            Dot product of two vectors.
*     S   SCSUM           Sum of a vector and the scaled vector.
*     S   HPSRT           Heapsort algorithm.
*     S   CALQ            Solving X from linear equation A*X=Y.
*     S   LINEQ           Solving X from linear equation A*X=Y, where
*                           A is given in factorized form.
*
*     
*     The variable and subgradient differences and all the input
*     (MN+1)-vectors are stored in a circular order controlled by the
*     parameter point IOLD.
*
*
      
      SUBROUTINE CPSR1(N,MC,MN,IOLD,IFLAG,IPRINT,
     &     X,XCP,XL,XU,IB,IT,GA,DC,T,
     &     SM,UM,SMTGA,UMTGA,SMTSM,LM,CDIAG,TMPM2,
     &     PAUX,CAUX,VAUX,WAUX,GAMMA,SMALL,IERR)
      
      
*     Scalar Arguments
      INTEGER N,MC,MN,IOLD,IFLAG,IPRINT,IERR
      DOUBLE PRECISION GAMMA,SMALL
      
*     Array Arguments
      INTEGER IB(*),IT(*)
      DOUBLE PRECISION GA(*),X(*),XCP(*),XL(*),XU(*),DC(*),T(*),
     &     SM(*),UM(*),LM(*),CDIAG(*),SMTSM(*),TMPM2(*),
     &     SMTGA(*),UMTGA(*),PAUX(*),CAUX(*),VAUX(*),WAUX(*)
      
*     Local Scalars
      INTEGER I,II,J,L,MCC,NBREAK,IFREE,ITMIN,NINT,NLEFT,ISBP,
     &     IBOUND
      DOUBLE PRECISION F1,F2,TMIN,DLTMIN,TOLD,DELTAT,GISBP,GISBP2,
     &     ZISBP,WMC,WMP,WMW,SMALL2
      
*     External Functions
      DOUBLE PRECISION VDOT
      EXTERNAL VDOT

*     Intrinsic Functions
      INTRINSIC MAX,ABS,SQRT

*     External Subroutines
      EXTERNAL CALQ,LINEQ,HPSRT,SCSUM
      
*
*     Initialization
*      
      NBREAK = 0
      NLEFT  = 0
      IFREE  = N+1
      F1     = 0.0D+00
      F2     = 0.0D+00
      TMIN   = 0.0D+00
      ITMIN  = 0
      IBOUND = 0
      IERR   = 0

      SMALL2=SMALL
c      SMALL2=SQRT(SMALL)

      
      IF (IOLD .EQ. 1) THEN
         MCC = MN
      ELSE IF (IFLAG .EQ. 0) THEN
         MCC = MC + 1
      ELSE
         MCC = MC
      END IF
      

*     
*     Temporarily set PAUX = -(UM - 1/GAMMA*SM)'*GA
*     and CAUX = (UM - 1/GAMMA*SM)'*(XCP-X) = 0
*      
      DO 10 I=1,MN
         L = I + IOLD - 1
         IF (L .GT. MCC) L=L-MCC
         PAUX(I) = -UMTGA(L)+SMTGA(L)/GAMMA

         CAUX(I)= 0.0D+00
 10   CONTINUE


*
*     Computation of the Cauchy direction DC, the first derivative
*     F1= -(DC)'*DC, and the breakpoints T in each coordinate
*     direction. Identification of the smallest breakpoint TMIN. 
*      

      DO 20 I=1,N

         IF (IB(I) .NE. 0) THEN
            IF (IB(I).LE.2 .AND. ABS(GA(I)) .LE. SMALL2 .AND. 
     &           X(I)-XL(I) .LE. SMALL2) THEN
               DC(I) = 0.0D+00
               XCP(I) = X(I)
               GOTO 20
            END IF

            IF (IB(I).GE.2 .AND. ABS(GA(I)) .LE. SMALL2 .AND. 
     &           XU(I)-X(I) .LE. SMALL2) THEN
               DC(I) = 0.0D+00
               XCP(I) = X(I)
               GOTO 20
            END IF
         END IF


         IF (IB(I).NE.0 .AND. IB(I).LE.2 .AND. GA(I).GT.0.0D+00) THEN


*
*     Breakpoint in lower bound
*

            T(NBREAK+1) = (X(I) - XL(I))/GA(I)
            
            IF (T(NBREAK+1) .LE. SMALL2) THEN
               DC(I) = 0.0D+00
               

*     
*     Correction to PAUX = (UM - 1/GAMMA*SM)'*DC 
*
               DO 30 J=1,MN
                  L = J + IOLD - 1
                  IF (L .GT. MCC) L=L-MCC
                  PAUX(J) = PAUX(J) +
     &                 (UM(N*(L-1)+I)-SM(N*(L-1)+I)/GAMMA)*GA(I)
 30            CONTINUE

            ELSE

               NBREAK = NBREAK + 1
               IT(NBREAK) = I
               DC(I) = -GA(I)
               F1 = F1 - DC(I)*DC(I)


*     
*     Determination of the smallest breakpoint
*               
               IF (NBREAK .EQ. 1 .OR. T(NBREAK) .LT. TMIN) THEN
                  TMIN=T(NBREAK)
                  ITMIN=NBREAK
               END IF
            END IF
         ELSE IF (IB(I) .GE. 2 .AND. GA(I) .LT. 0.0D+00) THEN


*
*     Breakpoint in upper bound
*               

            T(NBREAK+1) = (X(I) - XU(I))/GA(I)
            
            IF (T(NBREAK+1) .LE. SMALL2) THEN
               DC(I) = 0.0D+00


*     
*     Correction to P = (UM - 1/GAMMA*SM)'*DC 
*

               DO 40 J=1,MN
                  L = J + IOLD - 1
                  IF (L .GT. MCC) L=L-MCC
                  PAUX(J) = PAUX(J) +
     &                 (UM(N*(L-1)+I)-SM(N*(L-1)+I)/GAMMA)*GA(I)
 40            CONTINUE

            ELSE

               NBREAK = NBREAK + 1
               IT(NBREAK) = I
               DC(I) = -GA(I)
               F1 = F1 - DC(I)*DC(I)


*
*     Determination of the smallest breakpoint
*               

               IF (NBREAK .EQ. 1 .OR. T(NBREAK) .LT. TMIN) THEN
                  TMIN=T(NBREAK)
                  ITMIN=NBREAK
               END IF
            END IF
            
         ELSE


*
*     No breakpoint 
*
            IFREE = IFREE - 1
            IT(IFREE) = I
            DC(I) = -GA(I)
            F1 = F1 - DC(I)*DC(I) 

            IF (ABS(GA(I)) .GT. 0.0D+00) IBOUND=-1

         END IF


*     
*     Initialization of Cauchy point
*

         XCP(I) = X(I)
           
 20   CONTINUE


*
*     The indices corresponding to free variables are located in
*     IT(1),...,IT(NBREAK) and IT(IFREE),...,IT(N). The smallest of the
*     NBREAK breakpoints is in T(ITMIN)=TMIN.
*
      
*
*     DC is a zero vector. Return with the initial XCP.
*

      IF (NBREAK .EQ. 0 .AND. IBOUND .EQ. 0) GOTO 999


*
*     Initialization of derivative F2 = (DC)' BM DC, where BM is 
*     representated by compact limited memory SR1 formula.
*

      F2 =  -F1/GAMMA


*
*     Calculation of matrix (LM+LM+CDIAG-SMTSM)
*      

      DO 50 I=1,MN*(MN+1)/2
         TMPM2(I)= LM(I) - SMTSM(I)/GAMMA
 50   CONTINUE
      
      II = 0
      DO 51 I = 1, MN
         II=II+I
         J=I+IOLD-1
         IF (J .GT. MCC) J=J-MCC
      	 TMPM2(II) = TMPM2(II) + CDIAG(J)
 51   CONTINUE
      

c     The product of the inverse of mc x mc middle matrix in the compact
c     L-SR1 formula of BM and a mc vector PAUX; The product is returned
c     in VAUX. Note that TMPM2 will be factorized during the process.

      CALL CALQ(MN,MN,1,TMPM2,VAUX,PAUX,SMALL,IPRINT,IERR)
      IF (IERR .NE. 0) RETURN


      F2 = F2 + VDOT(MN,PAUX,VAUX)


      IF (F2 .LE. SMALL) THEN
         IERR = -7
         RETURN
      END IF


      DLTMIN = -F1/F2

      TOLD = 0.0D+00
      NINT = 1

      NLEFT = NBREAK
 
     
*
*     There is no breakpoints, locate the GCP and return. 
* 

      IF (NBREAK .EQ. 0) GOTO 888

      
*
*     Begining of the loop.
*      

 100  CONTINUE


      IF (NINT .EQ. 1) THEN


*     the smallest breakpoint is TMIN=T(ITMIN) and its index is
*     in IT(ITMIN)
         
         ISBP = IT(ITMIN)

      ELSE

         
*     the smallest breakpoint is chosen by heapsort algorithm

         IF (NINT .EQ. 2) THEN

            
*     Remove the already used smallest breakpoint before heapsort call.
            
            IF (ITMIN .NE. NBREAK) THEN
               T(ITMIN) = T(NBREAK)
               IT(ITMIN) = IT(NBREAK)
            END IF

     
*     Heapsort with initialization of the heap.

            CALL HPSRT(NLEFT,T,IT,0)

         ELSE

            
*     Heapsort with updating the heap.
            
            CALL HPSRT(NLEFT,T,IT,1)

         END IF


         TMIN = T(NLEFT)
         ISBP = IT(NLEFT)

      END IF

      DELTAT = TMIN - TOLD
      
      IF (DELTAT .LT. 0.0D+00) THEN
         PRINT*,' no nyt voit jo hirtt ihtes SR1.'
      END IF


*
*     Minimizer is within this interval, locate the GCP and return.
*

      IF (DLTMIN .LT. DELTAT) GOTO 888


*      
*     Examination of subsequent seqment
*

      NLEFT = NLEFT - 1
      NINT = NINT + 1
      GISBP = GA(ISBP)
      DC(ISBP) = 0.0D+00

      IF (GISBP .LT. 0.0D+00) THEN
         XCP(ISBP) = XU(ISBP)
         ZISBP = XU(ISBP) - X(ISBP)
      ELSE 
         XCP(ISBP) = XL(ISBP)
         ZISBP = XL(ISBP) - X(ISBP)
      END IF


*         
*     All  variables are fixed, return with current XCP as GCP.
*         

      IF (NLEFT .EQ. 0 .AND. NBREAK .EQ. N) GOTO 999


*      
*     Update F1 and F2
*      

      GISBP2=GISBP*GISBP


*
*     Initial updating of F1 and F2
*      

      F1 = F1 + DELTAT*F2 + GISBP2 + GISBP*ZISBP/GAMMA
      F2 = F2 - GISBP2/GAMMA


*         
*     Update CAUX = CAUX + DELTAT*PAUX.
*         

      CALL SCSUM(MN,DELTAT,PAUX,CAUX,CAUX)


*     
*     Selection of WAUX, the ISBP:th row of (UM - 1/GAMMA*SM). 
*
         
      DO 110 J=1,MN
         L = J + IOLD - 1
         IF (L .GT. MCC) L=L-MCC
         WAUX(J)= UM(N*(L-1)+ISBP)-SM(N*(L-1)+ISBP)/GAMMA
 110  CONTINUE


*
*     Computation of products WMC, WMP and WMW
*

*     The product of the inverse of mc x mc middle matrix in the compact
*     L-SR1 formula of BM and an MC vector WAUX; The product is returned
*     in VAUX. Note that TMPM2 is already in the factorized form.

      CALL LINEQ(MN,MN,1,TMPM2,VAUX,WAUX,SMALL,IERR)
      IF (IERR .NE. 0) RETURN
         
      WMC = VDOT(MN,CAUX,VAUX)
      WMP = VDOT(MN,PAUX,VAUX)
      WMW = VDOT(MN,WAUX,VAUX)


*         
*     Update PAUX = PAUX + GISBP*WAUX. 
*

      CALL SCSUM(MN,GISBP,WAUX,PAUX,PAUX)

      
*
*     Complete updating of F1 and F2
*      

      F1 = F1 + GISBP*WMC
      F2 = F2 + 2.0D+00*GISBP*WMP + GISBP2*WMW

      IF (F2 .LE. SMALL) THEN
         IERR = -7
         RETURN
      END IF


      DLTMIN = -F1/F2


*
*     Repeat the loop for unsearched intervals.          
*

      TOLD = TMIN
      
      IF (NLEFT .GT. 0) GOTO 100
      IF (IBOUND .EQ. 0) DLTMIN = 0.0D+00   
      
*     
*     End of the loop.
*


 888  CONTINUE

      DLTMIN = MAX(0.0D+00,DLTMIN)
      TOLD = TOLD + DLTMIN


*      
*     Update free variables 
* 

      DO 60 I=1,NLEFT
         XCP(IT(I))=X(IT(I))+TOLD*DC(IT(I))
 60   CONTINUE

      DO  80 I=IFREE,N
         XCP(IT(I))=X(IT(I))+TOLD*DC(IT(I))
 80   CONTINUE
     

 999  CONTINUE

      
      RETURN
      END
      


************************************************************************
*
*     * SUBROUTINE AGBFGS *
*
*      
*     * Purpose *
*      
*     Computation of aggregate values by the limited memory BFGS update.
*
*      
*     * Calling sequence *
*     
*     CALL AGBFGS(N,MC,MCC,INEW,IBFGS,IFLAG,G,GP,GA,U,SM,UM,R,C,UMTUM,
*     &      SMTGP,ALFN,ALFV,GAMMA,TMPMC1,TMPMC2,TMPMC3,IC,RHO,SMALL,IERR)
*     
*     
*     * Parameters *
*
*     II  N               Number of variables.
*     II  MC              Declared number of stored corrections.
*     II  MCC             Current number of stored corrections.
*     II  INEW            Index for circular arrays.
*     II  IBFGS           Index of the type of BFGS update.
*     IU  IFLAG           Index for adaptive version:
*                             0  - Maximum number of stored corrections
*                                    has not been changed at this
*                                    iteration.
*                             1  - Maximum number of stored corrections
*                                    has been changed at this iteration.
*     RI  ALFN            Locality measure.
*     RO  ALFV            Aggregate locality measure.
*     RI  D(N)            Search direction.
*     RI  G(N)            Current (auxiliary) subgradient of the
*                           objective function.
*     RI  GP(N)           Previous subgradient and also aggregate
*                           subradient of the objective function
*     RO  GA(N)           Next aggregate subgradient of the objective
*                           function.
*     RI  U(N)            Difference of trial and aggregate gradients.
*     RI  SM(N*(MC+1))    Matrix whose columns are stored corrections.
*     RI  UM(N*(MC+1))    Matrix whose columns are stored subgradient
*                           differences.
*     RI  R((MC+2)*(MC+1)/2)  Upper triangular matrix.
*     RI  UMTUM((MC+2)*(MC+1)/2)  Matrix UMTUM = UM' * UM.
*     RI  SMTGP(MC+1)     Vector SMTGP = SM'*GP.
*     RI  C(MC+1)         Diagonal matrix.
*     RI  GAMMA           Scaling parameter.
*     RA  TMPMC#(MC+1)    Auxiliary arrays; # = 1,3.
*     II  IC              Correction indicator.
*     RI  RHO             Correction parameter.
*     RI  SMALL           Small positive value.
*     IO  IERR            Error indicador: 
*                             0  - Everything is ok.
*                            -3  - Error in TRLIEQ; Restart.
*     
*     
*     * Local variables *
*      
*     I   MN              Current size of vectors.
*     I   IOLD            Index of the oldest corrections.
*     R   P               P = GP'*DM*U - ALFN where DM is the inverse
*                            approximation of the hessian calculated
*                            by using the L-BFGS formula.
*     R   Q               Q = U'*DM*U.
*     R   LAM             Multiplier used to calculate aggregate
*                            values.
*     R   W               Correction.
*     R   V               Correction.
*      
*     
*     * Subprograms used *
*      
*     RF  VDOT            Dot product of two vectors.
*     S   SYMAX           Multiplication of a dense symmetric matrix
*                           by a vector.
*     S   RWAXV2          Multiplication of two rowwise stored dense 
*                           rectangular matrices A and B by vectors X 
*                           and Y.
*     S   TRLIEQ          Solving X from linear equation L*X=Y or
*                           L'*X=Y.
*     
*
*     The variable and subgradient differences and all the MC-vectors are
*     stored in a circular order controlled by the parameter point INEW.
*
*      
      
      SUBROUTINE AGBFGS(N,MC,MCC,INEW,IBFGS,IFLAG,G,GP,GA,U,SM,UM,
     &     R,C,UMTUM,SMTGP,ALFN,ALFV,GAMMA,TMPMC1,TMPMC2,TMPMC3,
     &     IC,RHO,SMALL,IERR)


*     Scalar Arguments
      INTEGER N,MC,MCC,INEW,IBFGS,IFLAG,IC,IERR
      DOUBLE PRECISION GAMMA,ALFN,ALFV,RHO,SMALL
      
*     Array Arguments
      DOUBLE PRECISION G(*),GP(*),GA(*),U(*),SM(*),UM(*),
     &     R(*),C(*),UMTUM(*),SMTGP(*),TMPMC1(*),TMPMC2(*),TMPMC3(*)

*     Local Scalars
      INTEGER I,MN,IOLD
      DOUBLE PRECISION P,Q,LAM,W,V

*     External Functions
      DOUBLE PRECISION VDOT
      EXTERNAL VDOT

*     Intrinsic Functions
      INTRINSIC MAX,MIN,SIGN

*     External Subroutines
      EXTERNAL RWAXV2,SYMAX,TRLIEQ
      
      IERR = 0
      
      IF (MCC .LT. MC) THEN
         IF (IBFGS .EQ.2) THEN
            MN = MCC + 1

         ELSE
            MN = MCC
         END IF

         IOLD = 1

      ELSE
         IF (IFLAG .EQ. 0) THEN
            IF (IBFGS .EQ. 2) THEN
               MN = MC + 1
            ELSE
               MN = MC
            END IF

            IOLD = INEW + 1
            IF (IOLD .GT. MC+1) IOLD = 1

         ELSE
            IF (IBFGS .EQ. 2) THEN
               MN = MC
            ELSE
               MN = MC - 1
            END IF

            IOLD = INEW + 1
            IF (IOLD .GT. MC) IOLD = 1
         END IF 

      END IF 
      
      
*     
*     Initialization
*

      P = -VDOT(N,GP,U)
      Q = VDOT(N,U,U)

      IF (IC .EQ. 1) THEN
         W = RHO * Q
         V = RHO * P
      ELSE
         W = 0.0D+00
         V = 0.0D+00
      END IF
         
        
*     
*     Computation of the product U'*DM*U and -GP'*DM*U
*

      IF (MCC .GT. 0 .OR. IBFGS .EQ. 2) THEN

         IF (IOLD .EQ. 1 .OR. IBFGS .EQ. 2) THEN
            CALL RWAXV2(N,MN,SM,UM,U,U,TMPMC1,TMPMC2)
            CALL TRLIEQ(MN,MN,IOLD,R,TMPMC1,TMPMC1,1,SMALL,IERR)
            CALL TRLIEQ(MN,MN,IOLD,R,TMPMC3,SMTGP,1,SMALL,IERR)
            IF (IERR .NE. 0) RETURN
            
            Q = Q - 2*VDOT(MN,TMPMC2,TMPMC1)
            Q = GAMMA*Q

            P = P + 2*VDOT(MN,TMPMC2,TMPMC3)
            P = GAMMA*P
            
            DO 10 I=1,MN
               TMPMC2(I) = C(I)*TMPMC1(I)
 10         CONTINUE

            Q = Q + VDOT(MN,TMPMC1,TMPMC2)
            P = P - VDOT(MN,TMPMC2,TMPMC3)

            CALL SYMAX(MN,MN,IOLD,UMTUM,TMPMC1,TMPMC2)

            Q = Q + GAMMA*VDOT(MN,TMPMC1,TMPMC2)
            P = P - GAMMA*VDOT(MN,TMPMC2,TMPMC3)

         ELSE
            CALL RWAXV2(N,INEW-1,SM,UM,U,U,TMPMC1,TMPMC2)
            CALL RWAXV2(N,MCC-INEW,SM((IOLD-1)*N+1),UM((IOLD-1)*N+1),
     &           U,U,TMPMC1(IOLD),TMPMC2(IOLD))
            CALL TRLIEQ(MN,MCC,IOLD,R,TMPMC1,TMPMC1,1,SMALL,IERR)
            CALL TRLIEQ(MN,MCC,IOLD,R,TMPMC3,SMTGP,1,SMALL,IERR)
            IF (IERR .NE. 0) RETURN

            Q = Q - 2*(VDOT(MCC-INEW,TMPMC2(IOLD),TMPMC1(IOLD)) + 
     &           VDOT(INEW-1,TMPMC2,TMPMC1))
            Q = GAMMA*Q

            P = P + 2*(VDOT(MCC-INEW,TMPMC2(IOLD),TMPMC3(IOLD)) +
     &           VDOT(INEW-1,TMPMC2,TMPMC3))
            P = GAMMA*P

            DO 20 I=1,MCC
               TMPMC2(I) = C(I)*TMPMC1(I)
 20         CONTINUE

            Q = Q + VDOT(MCC-INEW,TMPMC1(IOLD),TMPMC2(IOLD)) + 
     &           VDOT(INEW-1,TMPMC1,TMPMC2)
            P = P - VDOT(MCC-INEW,TMPMC2(IOLD),TMPMC3(IOLD)) -
     &           VDOT(INEW-1,TMPMC2,TMPMC3)

            CALL SYMAX(MN,MCC,IOLD,UMTUM,TMPMC1,TMPMC2)

            Q = Q + GAMMA*(VDOT(MCC-INEW,TMPMC1(IOLD),TMPMC2(IOLD)) + 
     &           VDOT(INEW-1,TMPMC1,TMPMC2))
            P = P - GAMMA*(VDOT(MCC-INEW,TMPMC2(IOLD),TMPMC3(IOLD)) +
     &           VDOT(INEW-1,TMPMC2,TMPMC3))
         END IF

      END IF


      Q = Q + W
      P = P + V - ALFN
      
      LAM = 0.5D+00 + SIGN(0.5D+00,P)
      IF (Q .GT. 0.0D+00) LAM = MIN(1.0D+00,MAX(0.0D+00,P/Q))

      
*
*     Computation of the aggregate values
*      

      P = 1.0D+00 - LAM
      DO 30 I=1,N
         GA(I)=LAM*G(I) + P*GP(I)
 30   CONTINUE
      
      ALFV = LAM*ALFN
      

      RETURN
      END

      

************************************************************************
*
*     * SUBROUTINE AGGSR1 *
*
*      
*     * Purpose *
*      
*     Computation of aggregate values by the limited memory SR1 update.
*
*      
*     * Calling sequence *
*     
*     CALL AGGSR1(N,MC,MCC,INEW,IFLAG,G,GP,GA,D,ALFN,ALFV,TMPMAT,UMTUM,
*    &     R,GAMMA,SMTGP,UMTGP,SMTGA,UMTGA,SM,UM,TMPMC3,TMPMC4,TMPMC5,
*    &     TMPN2,TMPN3,TMPN4,ICN,RHO,SMALL,IERR)
*
*      
*     * Parameters *
*
*     II  N               Number of variables.
*     II  MC              Declared number of stored corrections.
*     II  MCC             Current number of stored corrections.
*     II  INEW            Index for circular arrays.
*     II  IFLAG           Index for adaptive version:
*                             0  - Maximum number of stored corrections
*                                    has not been changed at this
*                                    iteration.
*                             1  - Maximum number of stored corrections
*                                    has been changed at this iteration.
*     RI  G(N)            Current (auxiliary) subgradient of the
*                           objective function.
*     RI  GP(N)           Basic subgradient of the objective function.
*     RU  GA(N)           Current aggregate subgradient.
*     RI  ALFN            Locality measure.
*     RU  ALFV            Aggregate locality measure.
*     RI  SM(N*(MC+1))    Matrix whose columns are stored corrections.
*     RI  UM(N*(MC+1))    Matrix whose columns are stored subgradient.
*     RI  UMTUM((MC+2)*(MC+1)/2)  Matrix UMTUM = UM'*UM.
*     RI  R((MC+2)*(MC+1)/2)   Upper triangular matrix.
*     RI  SMTGP(MC+1)     Vector SMTGP = SM'*GP.
*     RI  SMTGA(MC+1)     Vector SMTGA = SM'*GA.
*     RI  UMTGP(MC+1)     Vector UMTGP = UM'*GP.
*     RI  UMTGA(MC+1)     Vector UMTGA = UM'*GA.
*     RI  GAMMA           Scaling parameter.
*     RA  TMPMC#(MC+1)    Auxiliary arrays; # = 3,4.
*     RA  TMPN#(N)        Auxiliary arrays; # = 2,...,4.
*     RA  TMPMAT((MC+1)*(MC)/2)  Auxiliary matrix.
*     II  ICN             Correction indicator.
*     RI  RHO             Correction parameter.
*     RI  SMALL           Small positive value.
*     IO  IERR            Error indicador: 
*                             0  - Everything is ok.
*                            -2  - Error In LINEQ.
*
*     
*     * Local variables *
*
*     I   MN              Current size of vectors.
*     I   IOLD            Index of the oldest corrections.
*     R   LAM#            Multipliers: # = 1,2.
*     R   PR              PR = (GP-GA)'*DM*(GP-GA), where DM
*                           presents the L-SR1- approximation of Hessian.
*     R   RRP             RRP = (GP-GA)'*DM*GA - ALFV.
*     R   PRQR            PRQR = (GP-GA)'*DM*(G-GA).
*     R   RRQ             RRQ = (G-GA)'*DM*GA - ALFV + ALFN.
*     R   QR              QR = (G-GA)'*DM*(G-GA).
*     R   PQ              PQ = (G-GP)'*DM*(G-GP).
*     R   QQP             QQP = (G-GP)'*DM*G + ALFN.
*     R   TMP1            Auxiliary scalar.
*     R   TMP2            Auxiliary scalar.
*     R   W               Correction.
*     R   V               Correction.
*
*     
*     * Subprograms used *
*      
*     RF  VDOT            Dot product of two vectors.
*     S   XSUMY           Sum of two vectors.
*     S   RECMAX          Multiplication of a vector by a dense
*                           rectangular matrix.
*     S   RWAXV2          Multiplication of two rowwise stored dense 
*                           rectangular matrices A and B by vectors X 
*                           and Y.
*     S   SCALEX          Scaling a vector.
*     S   XDIFFY          Difference of two vectors.
*     S   SCSUM           Sum of a vector and the scaled vector.
*     S   SCDIFF          Difference of the scaled vector and a vector.
*     S   CALQ            Solving X from linear equation A*X=Y.
*     S   LINEQ           Solving X from linear equation A*X=Y, where
*                           A is in factorized form.
*     
*
*     The variable and subgradient differences and all the MC-vectors are
*     stored in a circular order controlled by the parameter point INEW.
*
*     
      
      SUBROUTINE AGGSR1(N,MC,MCC,INEW,IFLAG,G,GP,GA,ALFN,ALFV,TMPMAT,
     &     UMTUM,R,GAMMA,SMTGP,UMTGP,SMTGA,UMTGA,SM,UM,TMPMC3,TMPMC4,
     &     TMPMC5,TMPN2,TMPN3,TMPN4,ICN,RHO,SMALL)
      
*     Scalar Arguments
      INTEGER N,MC,MCC,INEW,IFLAG,ICN
      DOUBLE PRECISION ALFN,ALFV,GAMMA,RHO,SMALL
      
*     Array Arguments
      DOUBLE PRECISION G(*),GP(*),GA(*),TMPMAT(*),UMTUM(*),R(*),
     &     SMTGP(*),UMTGP(*),SMTGA(*),UMTGA(*),SM(*),UM(*),
     &     TMPMC3(*),TMPMC4(*),TMPMC5(*),TMPN2(*),TMPN3(*),TMPN4(*)
      
*     Local Scalars
      INTEGER I,IOLD,MN,IERR
      DOUBLE PRECISION LAM1,LAM2,PR,RRP,PRQR,RRQ,QR,PQ,QQP,W,TMP1,TMP2,V
      
*     External Functions
      DOUBLE PRECISION VDOT
      EXTERNAL VDOT

*     External Subroutines
      EXTERNAL XDIFFY,SCALEX,XSUMY,RECMAX,RWAXV2,CALQ,SCDIFF,SCSUM,
     &     LINEQ

*     Intrinsic Functions
      INTRINSIC MIN,MAX

      
      IF (MCC .LT. MC) THEN
         IOLD = 1
         MN = MCC

      ELSE IF (IFLAG .EQ. 0) THEN
         MN = MC
         IOLD = INEW + 1
         IF (IOLD .GT. MC+1) IOLD = 1

      ELSE
         MN = MC - 1
         IOLD = INEW + 1
         IF (IOLD .GT. MC) IOLD = 1
      END IF 

      
      CALL XDIFFY(N,GP,GA,TMPN2)

      
*      
*     Calculation of TMPN3 = (GP - GA)'*DM
*

      IF (MCC .GT. 0) THEN

         DO 10 I=1,MN*(MN+1)/2
            TMPMAT(I)= GAMMA * UMTUM(I) - R(I)
 10      CONTINUE

         IF (IOLD .EQ. 1) THEN
            CALL XDIFFY(MN,UMTGP,UMTGA,TMPMC4)
            CALL SCDIFF(MN,GAMMA,TMPMC4,SMTGP,TMPMC4)
            CALL XSUMY(MN,TMPMC4,SMTGA,TMPMC4)

            CALL CALQ(MN,MN,IOLD,TMPMAT,TMPMC3,TMPMC4,SMALL,0,
     &           IERR)
            CALL SCALEX(MN,GAMMA,TMPMC3,TMPMC4)

            CALL RECMAX(N,MN,SM,TMPMC3,TMPN4)
            CALL SCSUM(N,GAMMA,TMPN2,TMPN4,TMPN3)
            CALL RECMAX(N,MN,UM,TMPMC4,TMPN4)
            CALL XDIFFY(N,TMPN3,TMPN4,TMPN3)

         ELSE
            CALL XDIFFY(MCC,UMTGP,UMTGA,TMPMC4)
            CALL SCDIFF(MCC,GAMMA,TMPMC4,SMTGP,TMPMC4)
            CALL XSUMY(MCC,TMPMC4,SMTGA,TMPMC4)

            CALL CALQ(MN,MCC,IOLD,TMPMAT,TMPMC3,TMPMC4,SMALL,0,IERR)
            CALL SCALEX(MCC,GAMMA,TMPMC3,TMPMC4)

            CALL RECMAX(N,INEW-1,SM,TMPMC3,TMPN4)
            CALL SCSUM(N,GAMMA,TMPN2,TMPN4,TMPN3)
            CALL RECMAX(N,MN-INEW+1,SM((IOLD-1)*N+1),TMPMC3(IOLD)
     &           ,TMPN4)
            CALL XSUMY(N,TMPN3,TMPN4,TMPN3)
            CALL RECMAX(N,INEW-1,UM,TMPMC4,TMPN4)
            CALL XDIFFY(N,TMPN3,TMPN4,TMPN3)
            CALL RECMAX(N,MN-INEW+1,UM((IOLD-1)*N+1),TMPMC4(IOLD)
     &           ,TMPN4)
            CALL XDIFFY(N,TMPN3,TMPN4,TMPN3)
         END IF
         
         IF (ICN .EQ. 1) THEN
            CALL SCSUM(N,RHO,TMPN2,TMPN3,TMPN3)
         END IF
         
         PR = VDOT(N,TMPN3,TMPN2)
         RRP = VDOT(N,TMPN3,GA) 
         CALL XDIFFY(N,G,GA,TMPN4)
         PRQR = VDOT(N,TMPN3,TMPN4)

      ELSE

         PR = VDOT(N,TMPN2,TMPN2)
         RRP = VDOT(N,TMPN2,GA) 
         CALL XDIFFY(N,G,GA,TMPN4)
         PRQR = VDOT(N,TMPN2,TMPN4)
      END IF

      
*
*     Calculation of QR = (G-GA)'*DM*(G-GA) and RRQ = (G-GA)'*DM*GA
*

      QR = VDOT(N,TMPN4,TMPN4)
      RRQ = VDOT(N,TMPN4,GA)
      IF (ICN .EQ. 1) THEN
         W = RHO*QR
         V = RHO*RRQ
      ELSE
         W = 0.0D+00
         V = 0.0D+00
      END IF
      
      IF (MCC .GT. 0) THEN
         QR = GAMMA*QR
         RRQ = GAMMA*RRQ

         IF (IOLD .EQ. 1) THEN
            CALL RWAXV2(N,MN,SM,UM,TMPN4,TMPN4,TMPMC4,TMPMC3)
            CALL SCSUM(MN,-GAMMA,TMPMC3,TMPMC4,TMPMC4)
            CALL SCSUM(MN,-GAMMA,UMTGA,SMTGA,TMPMC5)
            CALL LINEQ(MN,MN,IOLD,TMPMAT,TMPMC3,TMPMC4,SMALL,IERR)
            
            QR = QR - VDOT(MN,TMPMC4,TMPMC3) + W
            RRQ = RRQ - VDOT(MN,TMPMC5,TMPMC3) + V

         ELSE
            CALL RWAXV2(N,INEW-1,SM,UM,TMPN4,TMPN4,TMPMC4,TMPMC3)
            CALL RWAXV2(N,MN-INEW+1,SM((IOLD-1)*N+1),
     &           UM((IOLD-1)*N+1),TMPN4,TMPN4,TMPMC4(IOLD),TMPMC3(IOLD))
            CALL SCSUM(MCC,-GAMMA,TMPMC3,TMPMC4,TMPMC4)
            CALL SCSUM(MCC,-GAMMA,UMTGA,SMTGA,TMPMC5)
            CALL LINEQ(MN,MCC,IOLD,TMPMAT,TMPMC3,TMPMC4,SMALL,IERR)
            QR = QR - VDOT(MCC-INEW,TMPMC4(IOLD),TMPMC3(IOLD)) -
     &           VDOT(INEW-1,TMPMC4,TMPMC3) + W
            RRQ = RRQ - VDOT(MCC-INEW,TMPMC5(IOLD),TMPMC3(IOLD)) -
     &           VDOT(INEW-1,TMPMC5,TMPMC3) + V
         END IF

      END IF
      
      PQ = QR - 2*PRQR + PR
      QQP = PQ + PRQR + RRQ - PR - RRP + ALFN
      RRP = RRP - ALFV
      RRQ = RRQ + ALFN - ALFV

      
*     
*     Computation of multipliers LAM1 and LAM2
*

      IF (PR .LE. 0.0D+00 .OR. QR .LE. 0.0D+00) GOTO 100
      TMP1 = RRQ/QR
      TMP2 = PRQR/QR
      W = PR - PRQR*TMP2
      IF (W .EQ. 0.0D+00) GOTO 100
      LAM1 = (TMP1*PRQR - RRP)/W
      LAM2 = -TMP1 - LAM1*TMP2
      IF (LAM1*(LAM1 - 1.0D+00) .LT. 0.0D+00 .AND.
     &     LAM2*(LAM1 + LAM2 - 1.0D+00) .LT. 0.0D+00) GOTO 200

      
*
*     Minimum on the boundary
*

 100  CONTINUE

      LAM1 = 0.0D+00
      LAM2 = 0.0D+00
      IF (ALFN .LE. ALFV) LAM2 = 1.0D+00
      IF (QR .GT. 0.0D+00) LAM2 = MIN(1.0D+00,MAX(0.0D+00,-RRQ/QR))
      W = (LAM2*QR + 2.0D+00*RRQ)*LAM2
      TMP1 = 0.0D+00
      IF (ALFV .GE. 0.0D+00) TMP1 = 1.0D+00
      IF (PR .GT. 0.0D+00) TMP1 = MIN(1.0D+00,MAX(0.0D+00,-RRP/PR))
      TMP2 = (TMP1*PR + 2.0D+00*RRP)*TMP1
      IF (TMP2 .LT. W) THEN
         W = TMP2
         LAM1 = TMP1
         LAM2 = 0.0D+00
      END IF
      
      IF (QQP*(QQP - PQ) .GE. 0.0D+00) GOTO 200

      IF (QR + 2.0D+00*RRQ - QQP*QQP/PQ .GE. W) GOTO 200
      LAM1 = QQP/PQ
      LAM2 = 1.0D+00 - LAM1
      
 200  CONTINUE

      IF (LAM1 .EQ. 0.0D+00 .AND. LAM2*(LAM2 - 1.0D+00) .LT. 0.0D+00
     &     .AND. -RRP - LAM2*PRQR .GT. 0.0D+00 .AND. PR .GT. 0.0D+00)
     &     LAM1 = MIN(1.0D+00 - LAM2, (-RRP-LAM2*PRQR)/PR)

      
*
*     Computation of the aggregate values
*      

      TMP1 = 1.0D+00 - LAM1 - LAM2
      DO 30 I=1,N
         GA(I)=LAM1*GP(I)+LAM2*G(I)+TMP1*GA(I)
 30   CONTINUE
      
      ALFV = LAM2*ALFN + TMP1*ALFV

      
      RETURN
      END

      

************************************************************************
*
*     * SUBROUTINE LLS2 *
*
*      
*     * Purpose *
*      
*     Special line search for LMBM-B.
*
*      
*     * Calling sequence *
*     
*     CALL LLS2(N,X,G,D,XO,T,FO,F,P,ALFN,TMIN,DNORM,WK,THETA,EPSL,
*    &     EPSR,ETA,MOS,ITERS,NFE,NNK,SMALL,ITERM)
*     
*     
*     * PARAMETERS *
*      
*     II  N             Number of variables.
*     RU  X(N)          Vector of variables.
*     RI  XO(N)         Previous vector of variables.
*     RO  G(N)          Subgradient of the objective function.
*     RI  D(N)          Direction vector.
*     RU  T             Stepsize.
*     RO  F             Value of the objective function.
*     RI  FO            Previous value of the objective function.
*     RU  P             Directional derivative.
*     RI  TMIN          Minimum value of the stepsize.
*     RI  DNORM         Euclidean norm of the direction vector.
*     RI  WK            Stopping parameter.
*     RI  EPSL          Termination tolerance for line search (in test
*                         on the change of the function value).
*     RI  EPSR          Termination tolerance for line search (in test
*                         on the directional derivative).
*     RI  ETA           Distance measure parameter.
*     RO  ALFN          Locality measure.
*     RI  THETA         Scaling parameter.
*     II  MOS           Locality measure parameter.
*     IO  ITERS         Null step indicator.
*                            0  - Null step.
*                            1  - Serious step.
*     II  NNK           Number of consequtive null steps.
*     IU  NFE           Number of function evaluations.
*     RI  SMALL         A small positive value.
*     IO  ITERM         Cause of termination:
*                          0  - Everything is ok.
*                         -3  - Failure in function or subgradient
*                               calculations.
*     
*      
*     * Local parameters *
*
*     I   MAXINT        Maximum number of interpolations.
*
*      
*     * Local variables *
*
*     I   NIN           Number of interpolations.
*     R   TL,TU         Lower and upper limits for T used in
*                         interpolation.
*     R   FL            Value of the objective function for T=TL.
*     R   FU            Value of the objective function for T=TU.
*     R   EPSA          Line search parameter.
*     R   EPST          Line search parameter.
*     R   EPSLK         Line search parameter.
*     R   EPSRK         Line search parameter.
*      
*
*     * Subprograms used *
*      
*     RF  VDOT          Dot product of two vectors.
*     S   QINT          Quadratic interpolation for line search
*                         with one directional derivative.
*     S   SCSUM         Sum of a vector and the scaled vector.
*
*      
*     * EXTERNAL SUBROUTINES *
*      
*     SE  FUNDER        Computation of the value and the subgradient of
*                       the objective function. Calling sequence:
*                       CALL FUNDER(N,X,F,G,ITERM), where N is a number
*                       of variables, X(N) is a vector of variables, F
*                       is the value of the objective function, G(N)
*                       is the subgradient of the objective function,
*                       and ITERM is the error indicator.
*      
*      
*

      SUBROUTINE LLS2(N,X,G,D,XO,T,FO,F,P,ALFN,TMIN,DNORM,WK,THETA,
     &     EPSL,EPSR,ETA,MOS,ITERS,NFE,NNK,SMALL,ITERM)

*     Scalar Arguments
      INTEGER N,MOS,ITERS,NFE,NNK,ITERM
      DOUBLE PRECISION T,FO,F,P,ALFN,TMIN,DNORM,WK,THETA,EPSL,
     &     EPSR,ETA,SMALL

*     Array Arguments
      DOUBLE PRECISION X(*),G(*),D(*),XO(*)

*     Local Scalars
      INTEGER NIN
      DOUBLE PRECISION FL,FU,TL,TU,EPSA,EPST,EPSLK,EPSRK

*     Parameters
      INTEGER MAXINT
      PARAMETER(MAXINT = 200)

*     External Functions
      DOUBLE PRECISION VDOT
      EXTERNAL VDOT

*     Intrinsic Functions
      INTRINSIC ABS,MAX

*     External Subroutines
      EXTERNAL FUNDER,SCSUM,QINT

      
*
*     Initialization
*      

      NIN=0

      EPST = 2.0D+00 * EPSL
      EPSA = (EPSR - EPST)/2.0D+00
      
      TL = 0.0D+00
      TU = T
      FL = FO

      IF (THETA .LT. 1.0D+00) THEN
         EPST  = THETA*EPST
         EPSA  = THETA*EPSA
         EPSLK = EPSL
         EPSL  = THETA*EPSL
         EPSRK = EPSR
         EPSR  = THETA*EPSR
      END IF

      
*     
*     Function and gradient evaluation at a new point
*

 10   CONTINUE

      CALL SCSUM(N,THETA*T,D,XO,X)
      CALL FUNDER(N,X,F,G,ITERM)
      NFE = NFE + 1

      IF (ITERM .NE. 0) RETURN
      

      P = THETA*VDOT(N,G,D)
      ALFN = MAX(ABS(FO-F+P*T),ETA*(T*THETA*DNORM)**MOS)
      

*     
*     Null/descent step test (ITERS=0/1)
*

      ITERS = 1
      IF (F .LE. FO - T*EPST*WK) THEN
         TL = T
         FL = F

      ELSE
         TU = T
         FU = F
      END IF

      
*
*     Serious step
*      

      IF (F .LE. FO - T*EPSL*WK .AND. (T .GE. TMIN .OR.
     &     ALFN .GT. EPSA*WK)) GO TO 40

      
*
*     Additional interpolation
*      

      IF (F .GT. FO .AND. TU-TL .GE. TMIN*1.0D-01
     &     .AND. NNK .GE. 1 .AND. NIN .LT. MAXINT) THEN
         GO TO 20
      ENDIF

      
*
*     Null step
*      

      IF (P-ALFN .GE. -EPSR*WK .OR. TU-TL .LT. TMIN*1.0D-01 .OR.
     &     NIN .GE. MAXINT) GO TO 30


*
*     Interpolation
*      

 20   CONTINUE
      
      NIN=NIN+1
      IF (TL.EQ.0.0D+00 .AND. WK.GT.0.0D+00) THEN
         CALL QINT(TU,FL,FU,WK,T,1.0D+00-0.5D+00/(1.0D+00-EPST),SMALL)

      ELSE
         T = 5.0D-01* (TU+TL)
      END IF

      GO TO 10

 30   CONTINUE
      ITERS = 0
      
 40   CONTINUE
      IF (THETA .LT. 1.0D+00) THEN
         EPSL=EPSLK
         EPSR=EPSRK
      END IF

      RETURN
      END


************************************************************************
*
*     * SUBROUTINE QINT *
*
*      
*     * Purpose *
*      
*     Quadratic interpolation for line search with one directional
*     derivative.
*
*
*     * Calling sequence *
*     
*     CALL QINT(TU,FL,FU,WK,T,KAPPA,SMALL)
*     
*     
*     * Parameters *
*
*     RI  TU            Upper value of the stepsize parameter.
*     RI  FL            Value of the objective function.
*     RI  FU            Value of the objective function for T=TU.
*     RI  WK            Directional derivative.
*     RO  T             Stepsize parameter.
*     RI  KAPPA         Interpolation parameter.
*     RI  SMALL         A small positive value.
*
*      
      
      SUBROUTINE QINT(TU,FL,FU,WK,T,KAPPA,SMALL)
      
*     Scalar Arguments
      DOUBLE PRECISION FL,FU,WK,T,TU,KAPPA,SMALL

*     Local Scalars
      DOUBLE PRECISION TMP1,TMP2

*     Intrinsic Functions 
      INTRINSIC MAX


      IF (WK*TU .LE. SMALL) THEN
         T = 0.50D+00 * TU
         RETURN
      END IF


      TMP1 = (FU-FL)/ (-WK*TU)

      
*     
*     Quadratic interpolation with one directional derivative
*

      TMP2 = 2.0D+00 * (1.0D+00 - TMP1)

      IF (TMP2 .GT. 1.0D+00) THEN

         
*     
*     Interpolation accepted
*

         T = MAX(KAPPA*TU,TU/TMP2)

         RETURN

      END IF

      
*     
*     Bisection
*

      T = 0.50D+00 * TU
      
      RETURN
      END


************************************************************************
*
*     * SUBROUTINE TINIT *
*
*      
*     * Purpose *
*      
*     Initial stepsize selection for LMBM-B.
*
*
*     * Calling sequence *
*     
*     CALL TINIT(N,NA,MAL,X,AF,AG,AY,IBUN,D,F,P,T,TMAX,TMIN,ETA,ETA9,
*    &     MOS,ITERS)
*     
*     
*     * Parameters *
*      
*     II  N             Number of variables.
*     II  NA            Maximum size of the bundle.
*     II  MAL           Current size of the bundle.
*     RI  X(N)          Vector of variables.
*     RU  AF(4*NA)      Vector of values of bundle functions.
*     RI  AG(N*NA)      Matrix whose columns are bundle subgradients.
*     RI  AY(N*NA)      Matrix whose columns are bundle points.
*     II  IBUN          Index for the circular arrays in bundle.
*     RI  D(N)          Direction vector.
*     RI  F             Value of the objective function.
*     RI  P             Directional derivative P = GA' .
*     RO  T             Value of the stepsize parameter.
*     RI  TMIN          Lower limit for stepsize parameter.
*     RI  TMAX          Upper limit for stepsize parameter.
*     RI  ETA           Distance measure parameter.
*     RI  ETA9          Maximum for real numbers.
*     RI  MOS           Locality measure parameter.
*     RI  SMALL         A small positive value.
*     II  ITERS         Null step indicator.
*                            0  - Null step.
*                            1  - Serious step.
*     
*
*     * Subprograms used *
*      
*     S   DESTEP        Stepsize determination for descent steps.
*     S   NULSTP        Stepsize determination for null steps.
*
*      
      
      SUBROUTINE TINIT(N,NA,MAL,X,AF,AG,AY,IBUN,D,F,P,T,TMAX,TMIN,ETA,
     &     ETA9,MOS,SMALL,ITERS)


*     Scalar Arguments
      INTEGER N,NA,MAL,IBUN,MOS,ITERS
      DOUBLE PRECISION P,ETA,ETA9,F,T,TMAX,TMIN,SMALL

*     Array Arguments
      DOUBLE PRECISION AF(*),AG(*),AY(*),D(*),X(*)

*     Local Scalars
      DOUBLE PRECISION SMALL2

*     External Subroutines
      EXTERNAL DESTEP,NULSTP

*     Intrinsic Functions
      INTRINSIC MAX,MIN,ABS,SQRT


      SMALL2 = SQRT(SMALL)
      T = MIN(1.0D+00,TMAX)

      IF (ABS(P) .LE. SMALL2) GO TO 10

      IF (ITERS.EQ.1) THEN
         CALL DESTEP(N,NA,MAL,X,AF,AG,AY,IBUN,D,F,P,T,ETA,ETA9,MOS)

      ELSE
         CALL NULSTP(N,NA,MAL,X,AF,AG,AY,IBUN,D,F,P,T,ETA,ETA9,MOS)

      END IF

 10   CONTINUE

      T = MIN(MAX(T,TMIN),TMAX)
      
      RETURN
      END


************************************************************************
*
*     * SUBROUTINE DESTEP *
*
*      
*     * Purpose *
*      
*     Stepsize selection using polyhedral approximation for descent step
*     in limited memory bundle method.
*
*
*     * Calling sequence *
*     
*     CALL DESTEP(N,NA,MAL,X,AF,AG,AY,IBUN,D,F,DF,T,ETA,ETA9,MOS)
*     
*     
*     * Parameters *
*      
*     II  N             Number of variables.
*     II  NA            Maximum size of the bundle.
*     II  MAL           Current size of the bundle.
*     RI  X(N)          Vector of variables.
*     RU  AF(4*NA)      Vector of values of bundle functions.
*     RI  AG(N*NA)      Matrix whose columns are bundle subgradients.
*     RI  AY(N*NA)      Matrix whose columns are bundle points.
*     II  IBUN          Index for the circular arrays in bundle.
*     RI  D(N)          Direction vector.
*     RI  F             Value of the objective function.
*     RI  DF            Directional derivative.
*     RO  T             Value of the stepsize parameter.
*     RI  ETA           Distance measure parameter.
*     RI  ETA9          Maximum for real numbers.
*     RI  MOS           Locality measure parameter.
*
*      

      SUBROUTINE DESTEP(N,MA,MAL,X,AF,AG,AY,IBUN,D,F,DF,T,ETA,ETA9,MOS)

*     Scalar Arguments
      INTEGER N,MA,MAL,MOS,IBUN
      DOUBLE PRECISION DF,ETA,ETA9,F,T

*     Array Arguments
      DOUBLE PRECISION AF(*),AG(*),AY(*),D(*),X(*)

*     Local Scalars
      INTEGER I,J,JN,K,L,LQ,IB
      DOUBLE PRECISION ALF,ALFL,ALFR,BET,BETL,BETR,DX,Q,R,W

*     Intrinsic Functions
      INTRINSIC ABS,DBLE,MAX,MIN,SQRT


      ALFL=0.0D+00
      BETL=0.0D+00
      
      W = DF*T* (1.0D+00-T*0.5D+00)
      
      
*     
*     Initial choice of possibly active lines
*      

      K = 0
      L = -1
      JN = (IBUN-1)*N
      BETR = -ETA9

      DO 20 J = 1,MAL - 1
         IB = IBUN - 1 + J
         IF (IB .GT. MAL) IB = IB - MAL
         IF (JN .GE. MAL*N) JN = JN - MAL*N
         R = 0.0D+00
         BET = 0.0D+00
         ALFL = AF(IB) - F

         DO 10 I = 1,N
            DX = X(I) - AY(JN+I)
            Q = AG(JN+I)
            R = R + DX*DX
            ALFL = ALFL + DX*Q
            BET = BET + D(I)*Q
 10      CONTINUE

         IF (MOS.NE.2) R = R** (DBLE(MOS)*0.5D0)
         ALF = MAX(ABS(ALFL),ETA*R)
         R = 1.0D+00 - BET/DF
         IF (R*R+ (ALF+ALF)/DF.GT.1.0D-6) THEN
            K = K + 1
            AF(MA+K) = ALF
            AF(MA+MA+K) = BET
            R = T*BET - ALF
            IF (R.GT.W) THEN
               W = R
               L = K
            END IF

         END IF

         BETR = MAX(BETR,BET-ALF)
         JN = JN + N
         
 20   CONTINUE

      LQ = -1
      IF (BETR.LE.DF*0.5D+00) RETURN
      LQ = 1
      IF (L.LT.0) RETURN
      BETR = AF(MA+MA+L)
      IF (BETR.LE.0.0D+00) THEN
         IF (T.LT.1.0D+00 .OR. BETR.EQ.0.0D+00) RETURN
         LQ = 2
      END IF

      ALFR = AF(MA+L)
      

*
*     Iteration loop
*

 30   IF (LQ.GE.1) THEN
         Q = 1.0D+00 - BETR/DF
         R = Q + SQRT(Q*Q+ (ALFR+ALFR)/DF)
         IF (BETR.GE.0.0D0) R = - (ALFR+ALFR)/ (DF*R)
         R = MIN(1.95D+00,MAX(0.0D+00,R))
         
      ELSE
         IF (ABS(BETR-BETL)+ABS(ALFR-ALFL).LT.-1.0D-4*DF) RETURN
         R = (ALFR-ALFL)/ (BETR-BETL)
      END IF

      IF (ABS(T-R).LT.1.0D-4) RETURN
      T = R
      AF(MA+L) = -1.0D+00
      W = T*BETR - ALFR
      L = -1

      DO 40 J = 1,K
         ALF = AF(MA+J)
         IF (ALF.LT.0.0D+00) GO TO 40
         BET = AF(MA+MA+J)
         R = T*BET - ALF
         IF (R.GT.W) THEN
            W = R
            L = J
         END IF

 40   CONTINUE
      IF (L.LT.0) RETURN
      BET = AF(MA+MA+L)
      IF (BET.EQ.0.0D+00) RETURN
      

*
*     New interval selection
*

      ALF = AF(MA+L)
      IF (BET.LT.0.0D+00) THEN
         IF (LQ.EQ.2) THEN
            ALFR = ALF
            BETR = BET
            
         ELSE
            ALFL = ALF
            BETL = BET
            LQ = 0
         END IF

      ELSE
         IF (LQ.EQ.2) THEN
            ALFL = ALFR
            BETL = BETR
            LQ = 0
         END IF

         ALFR = ALF
         BETR = BET
      END IF

      GO TO 30

      END


************************************************************************
*
*     * SUBROUTINE NULSTP *
*
*
*     * Purpose *
*      
*     Stepsize selection using polyhedral approximation for null step
*     in limited memory bundle method.
*
*      
*     * Calling sequence *
*     
*     CALL NULSTP(N,NA,MAL,X,AF,AG,AY,IBUN,D,F,DF,T,ETA,ETA9,MOS)
*     
*     
*     * Parameters *
*      
*     II  N             Number of variables.
*     II  NA            Maximum size of the bundle.
*     II  MAL           Current size of the bundle.
*     RI  X(N)          Vector of variables.
*     RU  AF(4*NA)      Vector of values of bundle functions.
*     RI  AG(N*NA)      Matrix whose columns are bundle subgradients.
*     RI  AY(N*NA)      Matrix whose columns are bundle points.
*     II  IBUN          Index for the circular arrays in bundle.
*     RI  D(N)          Direction vector.
*     RI  F             Value of the objective function.
*     RI  DF            Directional derivative.
*     RO  T             Value of the stepsize parameter.
*     RI  ETA           Distance measure parameter.
*     RI  ETA9          Maximum for real numbers.
*     RI  MOS           Locality measure parameter.
*
*      
      
      SUBROUTINE NULSTP(N,MA,MAL,X,AF,AG,AY,IBUN,D,F,DF,T,ETA,ETA9,MOS)

*     Scalar Arguments
      INTEGER MA,MAL,MOS,N,IBUN
      DOUBLE PRECISION DF,ETA,ETA9,F,T

*     Array Arguments
      DOUBLE PRECISION AF(*),AG(*),AY(*),D(*),X(*)

*     Local Scalars
      INTEGER I,J,JN,K,L,IB
      DOUBLE PRECISION ALF,ALFL,ALFR,BET,BETL,BETR,DX,Q,R,W

*     Intrinsic Functions
      INTRINSIC ABS,DBLE,MAX,MIN,SQRT


      W = DF*T


*     
*     Initial choice of possibly active parabolas
*

      K = 0
      L = -1
      JN = (IBUN-1)*N
      BETR = -ETA9

      DO 20 J = 1,MAL - 1
         IB = IBUN - 1 + J
         IF (IB .GT. MAL) IB = IB - MAL
         IF (JN .GE. MAL*N) JN = JN - MAL*N
         BET = 0.0D+00
         R = 0.0D+00
         ALFL = AF(IB) - F

         DO 10 I = 1,N
            DX = X(I) - AY(JN+I)
            R = R + DX*DX
            Q = AG(JN+I)
            ALFL = ALFL + DX*Q
            BET = BET + D(I)*Q
 10      CONTINUE

         IF (MOS.NE.2) R = R**(DBLE(MOS)*0.5D0)
         ALF = MAX(ABS(ALFL),ETA*R)
         BETR = MAX(BETR,BET-ALF)

         IF (ALF.LT.BET-DF) THEN
            K = K + 1
            R = T*BET - ALF
            AF(MA+K) = ALF
            AF(MA+MA+K) = BET

            IF (R.GT.W) THEN
               W = R
               L = K
            END IF

         END IF

         JN = JN + N
 20   CONTINUE

      IF (L.LT.0) RETURN
      BETR = AF(MA+MA+L)
      ALFR = AF(MA+L)
      ALF = ALFR
      BET = BETR
      ALFL = 0.0D+00
      BETL = DF


*     
*     Iteration loop
*

 30   W = BET/DF
      IF (ABS(BETR-BETL)+ABS(ALFR-ALFL).LT.-1.0D-4*DF) RETURN
      IF (BETR-BETL.EQ.0.0D+00) STOP 11
      R = (ALFR-ALFL)/ (BETR-BETL)

      IF (ABS(T-W).LT.ABS(T-R)) R = W
      Q = T
      T = R

      IF (ABS(T-Q).LT.1.0D-3) RETURN
      AF(MA+L) = -1.0D+00
      W = T*BET - ALF
      L = -1

      DO 40 J = 1,K
         ALF = AF(MA+J)
         IF (ALF.LT.0.0D+00) GO TO 40
         BET = AF(MA+MA+J)
         R = T*BET - ALF
         IF (R.GT.W) THEN
            W = R
            L = J
         END IF

 40   CONTINUE

      IF (L.LT.0) RETURN
      BET = AF(MA+MA+L)
      Q = BET - T*DF
      IF (Q.EQ.0.0D0) RETURN

      
*     
*     New interval selection
*

      ALF = AF(MA+L)
      IF (Q.LT.0.0D+00) THEN
         ALFL = ALF
         BETL = BET

      ELSE
         ALFR = ALF
         BETR = BET
      END IF

      GO TO 30

      END


************************************************************************
*
*     * SUBROUTINE DOBUN *
*
*
*     * Purpose *
*      
*     Bundle construction for limited memory bundle method
*
*      
*     * Calling sequence *
*     
*     CALL DOBUN(N,NA,MAL,X,G,F,AY,AG,AF,ITERS,IBUN)
*     
*     
*     * Parameters *
*      
*     II  N             Number of variables.
*     II  NA            Maximum size of the bundle.
*     II  MAL           Current size of the bundle.
*     RI  X(N)          Vector of variables.
*     RI  G(N)          Subgradient of the objective function.
*     RI  F             Value of the objective function.
*     RU  AY(N*NA)      Matrix whose columns are bundle points.
*     RU  AG(N*NA)      Matrix whose columns are bundle subgradients.
*     RU  AF(4*NA)      Vector of values of bundle functions.
*     IU  IBUN          Index for the circular arrays.
*     II  ITERS         Null step indicator.
*                            0  - Null step.
*                            1  - Serious step.
*
*      
*     * Subprograms used *
*
*     S   COPY2         Copying of two vectors.
*
*      

      SUBROUTINE DOBUN(N,MA,MAL,X,G,F,AY,AG,AF,ITERS,IBUN)

*     Scalar Arguments
      INTEGER ITERS,MA,MAL,N,IBUN
      DOUBLE PRECISION F
      
*     Array Arguments
      DOUBLE PRECISION AF(*),AG(*),AY(*),G(*),X(*)
      
*     Local Scalars
      INTEGER I,J
      
*     External Subroutines
      EXTERNAL COPY2


      IF (ITERS .EQ. 1) THEN


*     
*     Serious step
*      

         AF(IBUN) = F
         I = (IBUN-1)*N + 1
         CALL COPY2(N,G,AG(I),X,AY(I))

      ELSE


*
*     Null step
*      

         IF (MAL .LT. MA) THEN

            AF(IBUN) = AF(MAL)
            AF(MAL) = F

            I = MAL*N + 1
            CALL COPY2(N,AG(I-N),AG(I),AY(I-N),AY(I))
            CALL COPY2(N,G,AG(I-N),X,AY(I-N))
            
         ELSE
            I = IBUN-1
            IF (I .LT. 1) I = MAL
            AF(IBUN) = AF(I)
            AF(I) = F

            I = (IBUN-2)*N + 1
            IF (I .LT. 1) I = (MAL-1)*N + 1
            J = (IBUN-1)*N + 1
            CALL COPY2(N,AG(I),AG(J),AY(I),AY(J))
            CALL COPY2(N,G,AG(I),X,AY(I))
         END IF


      END IF
      
      MAL = MAL + 1
      IF (MAL .GT. MA) MAL = MA

      IBUN = IBUN + 1
      IF (IBUN .GT. MA) IBUN = 1

      RETURN
      END


      
************************************************************************
*
*     * SUBROUTINE PROJX *
*
*      
*     * Purpose *
*      
*     Projection of the initial X to the feasible region if necessary.
*
*      
*     * Calling sequence *
*     
*     CALL PROJX(N,X,XL,XU,IB,NACT,ITYPE)
*
*      
*     * Parameters *
*     
*     II  N               Number of variables.
*     RU  X(N)            Vector of variables. 
*     RI  XL(N)           Lower bounds for variables. 
*     RI  XU(N)           Upper bounds for variables.
*     II  IB(N)           Type of bound constraints:
*                             0  - X(I) is unbounded,
*                             1  - X(I) has only a lower bound,
*                             2  - X(I) has both lower and upper bounds,
*                             3  - X(I) has only an upper bound. 
*     IO  NACT            Number of active variables.
*     IO  ITYPE           Type of problem:
*                             0  - problem is unbounded,
*                             1  - problem is bounded.
*     
*      
      
      SUBROUTINE PROJX(N,X,XL,XU,IB,NACT,ITYPE)

      
*     Scalar Arguments
      INTEGER N,NACT,ITYPE

*     Array Arguments
      INTEGER IB(*)
      DOUBLE PRECISION X(*),XL(*),XU(*)

*     Local Scalars
      INTEGER I


*
*     Project the initial x to the feasible set if necessary.
*

      NACT = 0
      ITYPE = 0
      
      DO 10 I=1,N
         IF (IB(I) .GT. 0) THEN
            ITYPE = 1

            IF (IB(I) .LE. 2) THEN
               IF (X(I) .LE. XL(I)) THEN
                  X(I)=XL(I)
                  NACT = NACT+1
               END IF
            END IF

            IF (IB(I) .GE. 2) THEN
               IF (X(I) .GE. XU(I)) THEN
                  X(I)=XU(I)
                  NACT = NACT+1
               END IF
            END IF

         END IF
 10   CONTINUE


      RETURN
      END


      
************************************************************************
*
*     * SUBROUTINE PROJGR *
*
*      
*     * Purpose *
*      
*     Projection of the subgradient and calculation of Eucleidean norm.
*
*      
*     * Calling sequence *
*     
*     CALL PROJGR(N,X,XL,XU,G,GPROJ,GNORM,IB)
*
*      
*     * Parameters *
*     
*
*     II  N               Number of variables.
*     RI  X(N)            Vector of variables. 
*     RI  XL(N)           Lower bounds for variables. 
*     RI  XU(N)           Upper bounds for variables.
*     RI  G(N)            Current subgradient of the objective function.
*     RO  GPROJ(N)        Projected subgradient -(P(X-G)-X).
*     RO  GNORM           Eucleidean norm of projected subgradient.
*     II  IB(N)           Type of bound constraints:
*                             0  - X(I) is unbounded,
*                             1  - X(I) has only a lower bound,
*                             2  - X(I) has both lower and upper bounds,
*                             3  - X(I) has only an upper bound. 
*     
*      
      
      SUBROUTINE PROJGR(N,X,XL,XU,G,GPROJ,GNORM,IB)
      
*     Scalar Arguments
      INTEGER N
      DOUBLE PRECISION GNORM
      
*     Array Arguments
      INTEGER IB(*)
      DOUBLE PRECISION X(*),XL(*),XU(*),G(*),GPROJ(*)

*     Local Scalars
      INTEGER I
      
*     Intrinsic Functions
      INTRINSIC MAX,MIN


      GNORM = 0.0D+00

      DO 10 I=1,N
         GPROJ(I) = G(I)
         IF (IB(I) .GT. 0) THEN
            IF (IB(I) .LE. 2) THEN
               IF (X(I)-G(I) .LT. XL(I)) THEN
                  GPROJ(I) = MAX(0.0D+00,X(I)-XL(I))
               END IF
            END IF
               
            IF (IB(I) .GE. 2) THEN
               IF (X(I)-G(I) .GT. XU(I)) THEN
                  GPROJ(I) = MIN(0.0D+00,X(I)-XU(I))
               END IF
            END IF
         END IF
         GNORM = GNORM + GPROJ(I)*GPROJ(I)
 10   CONTINUE

      RETURN
      END
      

************************************************************************
*
*     * SUBROUTINE RESTAR *
*
*      
*     * Purpose *
*      
*     Initialization.
*
*      
*     * Calling sequence *
*     
*     CALL RESTAR(N,MC,MCC,MCINIT,INEW,IBUN,IBFGS,ITERS,GP,G,NNK,ALFV,
*    &     ALFN,GAMMA,IC,ICN,MAL,NCRES,IFLAG,IERR)
*     
*     
*     * Parameters *
*      
*     II  N               Number of variables.
*     IO  MC              Current maximum number of stored corrections.
*     IO  MCC             Current number of depositories used.
*     II  MCINIT          Initial maximum number of stored corrections.
*     IO  INEW            Index for the circular arrays.
*     IO  IBUN            Index for the circular arrays in bundle
*                           updating.
*     IO  IBFGS           Index of the type of BFGS update.
*     IU  ITERS           Null step indicator.
*                              0  - Null step.
*                              1  - Serious step.
*     IO  NNK             Consecutive null steps counter.
*     RI  GP(N)           Basic subgradient of the objective function.
*     RO  G(N)            Current (auxiliary) subgradient of the
*                           objective function.
*     RO  ALFN            Locality measure.
*     RO  ALFV            Aggregate locality measure.
*     RO  GAMMA           Scaling parameter.
*     IO  IC              Correction indicator.
*     IO  ICN             Correction indicator for null steps.
*     IO  MAL             Current size of the bundle.
*     IO  NCRES           Number of restarts.
*     IO  IFLAG           Index for adaptive version.
*     IO  IERR            Error indicator.
*
*      
*     * Subprograms used *
*      
*     S   COPY          Copying of a vector.
*     
*     
      
      SUBROUTINE RESTAR(N,MC,MCC,MCINIT,INEW,IBUN,IBFGS,ITERS,GP,G,NNK,
     &     ALFV,ALFN,GAMMA,IC,ICN,MAL,NCRES,IFLAG,IERR)

*     Scalar Arguments
      INTEGER N,MC,MCC,MCINIT,INEW,IBUN,IBFGS,ITERS,NNK,IC,ICN,MAL,
     &     NCRES,IFLAG,IERR
      DOUBLE PRECISION ALFN,ALFV,GAMMA

*     Array Arguments
      DOUBLE PRECISION G(*),GP(*)

*     External Subroutines
      EXTERNAL COPY


c      MC    = MCINIT
      MCC   = 0
      INEW  = 1
      IBUN  = 1
      IBFGS = 0
      IC    = 0
      ICN   = 0
      IFLAG = 0
      IERR  = 0
      MAL   = 0
      NCRES = NCRES + 1
      

      IF (ITERS .EQ. 0) THEN
         CALL COPY(N,GP,G)
         ITERS = 1
         NNK = 0
         ALFV=0.0D+00
         ALFN=0.0D+00
      END IF


      GAMMA=1.0D+00


      RETURN
      END

      
************************************************************************
*
*     * SUBROUTINE RPRINT *
*
*      
*     * Purpose *
*      
*     Printout the (final) results.
*
*      
*     * Calling sequence *
*     
*     SUBROUTINE RPRINT(N,NIT,NFE,NACT,X,F,WK,QK,ITERM,IPRINT)
*     
*     
*     * Parameters *
*      
*     II  N               Number of variables.
*     II  NIT             Number of used iterations.
*     II  NFE             Number of used function evaluations.
*     II  NACT            Number of active variables at solution.
*     RI  X(N)            Vector of variables.
*     RI  F               Value of the objective function.
*     RI  WK              Value of the first stopping criterion.
*     RI  QK              Value of the second stopping criterion.
*     II  ITERM           Cause of termination:
*                             1  - The problem has been solved.
*                                  with desired accuracy.
*                             2  - (F - FO) < TOLF in MTESF
*                                  subsequent iterations.
*                             3  - (F - FO) < TOLF*SMALL*MAX(|F|,|FO|,1).
*                             4  - Number of function calls > MFV.
*                             5  - Number of iterations > MIT.
*                             6  - Time limit exceeded. 
*                             7  - F < TOLB.
*                            -1  - Internal error: Two consecutive 
*                                  restarts.
*                            -2  - Internal error: Number of restarts 
*                                  > maximum number of restarts.
*                            -3  - Failure in function or subgradient
*                                  calculations (assigned by the user).
*                            -4  - Internal error: Failure in attaining 
*                                  the demanded accuracy.
*                            -5  - Invalid input parameters.
*                            -6  - Not enough working space.
*     II  IPRINT          Printout specification:
*                            -1  - No printout.
*                             0  - Only the error messages.
*                             1  - The final values of the objective
*                                  function.
*                             2  - The final values of the objective
*                                  function and the most serious
*                                  warning messages.
*                             3  - The whole final solution. 
*                             4  - At each iteration values of the
*                                  objective function.
*                             5  - At each iteration the whole
*                                  solution
*     
*      
      
      SUBROUTINE RPRINT(N,NIT,NFE,NACT,X,F,WK,QK,ITERM,IPRINT)

*     Scalar Arguments
      INTEGER N,NIT,NFE,NACT,ITERM,IPRINT,I
      DOUBLE PRECISION F,WK,QK
      
*     Array Arguments
      DOUBLE PRECISION X(*)

         
*
*     Intermediate results
*

      IF (ITERM .EQ. 0) THEN
         IF (IPRINT .GT. 3) WRITE (6,FMT='(1X,''NIT='',I5,2X,
     &        ''NFE='',I5,2X,''NACT='',I5,2X,''F='',D15.8,2X,''WK=''
     &        ,D11.2,2X,''QK ='',D11.2,2X)')
     &        NIT,NFE,NACT,F,WK,QK
         IF (IPRINT .EQ. 5) WRITE (6,FMT='(1X,''X ='',
     &        5D15.7:/(4X,5D15.7))')(X(I),I=1,N)
         RETURN
      END IF
         

*
*     Printout the final results
*

      IF (IPRINT .GT. 0) WRITE (6,FMT='(1X,''NIT='',I5,2X,
     &     ''NFE='',I5,2X,''NACT='',I5,2X,''F='',D15.8,2X,''WK=''
     &     ,D11.2,2X,''QK='',D11.2,2X,''ITERM='',I3)')
     &     NIT,NFE,NACT,F,WK,QK,ITERM
      IF (IPRINT .EQ. 3 .OR. IPRINT .EQ. 5)
     &     WRITE (6,FMT='(1X,''X ='',5D15.7:/(4X,5D15.7))')(X(I),I=1
     $     ,N)
      
      RETURN
      END

      
************************************************************************
*
*     * SUBROUTINE WPRINT *
*
*      
*     * Purpose *
*      
*     Printout the warning and error messages.
*
*      
*     * Calling sequence *
*     
*     SUBROUTINE WPRINT(ITERM,IPRINT,NOUT)
*     
*     
*     * Parameters *
*      
*     II  ITERM           Cause of termination:
*                             1  - The problem has been solved.
*                                  with desired accuracy.
*                             2  - (F - FO) < TOLF in MTESF subsequent 
*                                  iterations.
*                             3  - (F - FO) < TOLF*SMALL*MAX(|F|,|FO|,1).
*                             4  - Number of function calls > MFV.
*                             5  - Number of iterations > MIT.
*                             6  - Time limit exceeded. 
*                             7  - F < TOLB.
*                            -1  - Internal error: Two consecutive 
*                                  restarts.
*                            -2  - Internal error: Number of restarts 
*                                  > maximum number of restarts.
*                            -3  - Failure in function or subgradient
*                                  calculations (assigned by the user).
*                            -4  - Internal error: Failure in attaining 
*                                  the demanded accuracy.
*                            -5  - Invalid input parameters.
*                            -6  - Not enough working space.
*     II  IPRINT          Printout specification:
*                            -1  - No printout.
*                             0  - Only the error messages.
*                             1  - The final values of the objective
*                                  function.
*                             2  - The final values of the objective
*                                  function and the most serious
*                                  warning messages.
*                             3  - The whole final solution. 
*                             4  - At each iteration values of the
*                                  objective function.
*                             5  - At each iteration the whole
*                                  solution
*     II  NOUT            Auxilary printout specification.
*     
*      
      
      SUBROUTINE WPRINT(ITERM,IPRINT,NOUT)


*     Scalar Arguments
      INTEGER ITERM,IPRINT,NOUT



      IF (IPRINT .GE. 0) THEN


*     
*     Initial error messages
*

         IF (ITERM .LE. -5) THEN
            IF (ITERM .EQ. -5) THEN
               IF (NOUT .EQ. 1) WRITE (6,FMT='(1X,''Error: ''
     &              ''Number of variables (N) is too small. ITERM='',I3)
     $              ')ITERM
               IF (NOUT .EQ. 2) WRITE (6,FMT='(1X,''Error: ''
     &              ''The maximum number of stored corrections (MCU) ''
     &              ''is too small. ITERM='',I3)')ITERM
               IF (NOUT .EQ. 3) WRITE (6,FMT='(1X,''Error: ''
     &              ''The size of the bundle (NA) is too small. ITERM=''
     &              ,I3)')ITERM
               IF (NOUT .EQ. 4) WRITE (6,FMT='(1X,''Error: ''
     &              ''Line search parameter RPAR(5) >= 0.25. ITERM=''
     &              ,I3)')ITERM
            ELSE IF (ITERM .EQ. -6) THEN
               WRITE (6,FMT='(1X,''Error: ''
     &              ''Not enough working space. ITERM='',I3)')ITERM
               
            END IF
            RETURN
         END IF

         
*
*     Warning messages
*

         IF (IPRINT .EQ. 2) THEN
            IF (ITERM .EQ. 0) THEN
               IF (NOUT .GT. 0) WRITE (6,FMT='(1X,''Warning: ''
     &              ''The initial X is infeasible. ''
     &              ''Restart with its projection.'')')
               IF (NOUT .EQ. -1) WRITE (6,FMT='(1X,''Warning: ''
     &              ''MC > MCU. Assigned MC = MCU.'')')
               IF (NOUT .EQ. -2) WRITE (6,FMT='(1X,''Warning: ''
     &              ''A line search parameter EPSR >= 0.5.'')')
               IF (NOUT .EQ. -3) WRITE (6,FMT='(1X,''Warning: ''
     &              ''A nondescent search direction occured. Restart.'')
     &              ')
               IF (NOUT .EQ. -4) WRITE (6,FMT='(1X,''Warning: ''
     &              ''Does not converge.'')')
               IF (NOUT .EQ. -5) WRITE (6,FMT='(1X,''Warning: ''
     &              ''TMAX < TMIN. Restart.'')')
               IF (NOUT .EQ. -6) WRITE (6,FMT='(1X,''Warning: ''
     &              ''An indefinite matrix detected. Restart.'')
     &              ')

               RETURN
            END IF

         
*
*     Printout the final results
*
        
            IF (ITERM .EQ. 6) WRITE (6,FMT='(1X,''Abnormal exit: ''
     &           ''Time is up.'')')
            IF (ITERM .EQ. 7) WRITE (6,FMT='(1X,''Abnormal exit: ''
     &           ''F < TOLB.'')')
            IF (ITERM .EQ. 2) WRITE (6,FMT='(1X,''Abnormal exit: ''
     &           ''Too many steps without significant progress.'')
     &           ')
            IF (ITERM .EQ. 3) WRITE (6,FMT='(1X,''Abnormal exit: ''
     &           ''The value of the function does not change.'')')
            IF (ITERM .EQ. 5) WRITE (6,FMT='(1X,''Abnormal exit: ''
     &           ''Number of iterations > '',I5)') NOUT
            IF (ITERM .EQ. 4) WRITE (6,FMT='(1X,''Abnormal exit: ''
     &           ''Number of function evaluations > '',I5)') NOUT
            IF (ITERM .EQ. -1) THEN
               IF (NOUT .EQ. -1) THEN
                  WRITE (6,FMT='(1X,''Abnormal exit: ''
     &                 ''Two consecutive restarts.'')')
               ELSE
                  WRITE (6,FMT='(1X,''Abnormal exit: ''
     &           ''TMAX < TMIN in two subsequent iterations.'')')
               END IF
            END IF
            IF (ITERM .EQ. -2) WRITE (6,FMT='(1X,''Abnormal exit: ''
     &                 ''Number of restarts > '',I5''.'')') NOUT
            IF (ITERM .EQ. -3) WRITE (6,FMT='(1X,''Abnormal exit: ''
     &           ''Failure in function or subgradient calculations.'')')
            IF (ITERM .EQ. -4) WRITE (6,FMT='(1X,''Abnormal exit: ''
     &           ''Failure in attaining the demanded accuracy.'')')
         END IF

      END IF

      
      RETURN
      END

      

************************************************************************
*
*     * DOUBLE PRECISION FUNCTION SCLPAR *
*
*      
*     * Purpose *
*      
*     Calculation of the scaling parameter appointed by parameter ISCALE.
*
*      
*     * Calling sequence *
*      
*      GAMMA = SCLPAR(MN,ISCALE,STS,STU,UTU,SMALL)
*
*      
*     * Parameters *
*
*     II  MN              Current number of depositories used.
*     RI  STS             STS = S'*S. 
*     RI  STU             STU = S'*U. 
*     RI  UTU             UTU = U'*U. 
*     RI  SMALL           Small positive value.
*     II  ISCALE          Selection of the scaling:
*                           0  - Scaling at every iteration
*                                with STU/UTU.
*                           1  - Scaling at every iteration
*                                with STS/STU.
*                           2  - Interval scaling with STU/UTU.
*                           3  - Interval scaling with STS/STU.
*                           4  - Preliminary scaling with STU/UTU.
*                           5  - Preliminary scaling with STS/STU.
*                           6  - No scaling.      
*
*      

      DOUBLE PRECISION FUNCTION SCLPAR(MN,ISCALE,STS,STU,UTU,SMALL)
      
*     Scalar Arguments
      INTEGER MN,ISCALE
      DOUBLE PRECISION STS,STU,UTU,SMALL
      
*     Intrinsic Functions
      INTRINSIC SQRT


*        
*     Computation of scaling parameter.
*

      SCLPAR =1.0D+00


*
*     Scaling parameter = STU/UTU
*            

      IF (ISCALE .EQ. 0 .OR. ISCALE .EQ. 2 .OR. ISCALE .EQ. 4)
     &     THEN
         IF (UTU .LT. SQRT(SMALL)) THEN
            SCLPAR = 1.0D+00
            GO TO 80
         ELSE
            SCLPAR = STU/UTU
         END IF

         
*     
*     Scaling parameter = STS/STU
*               

      ELSE IF (ISCALE .EQ. 1 .OR. ISCALE .EQ. 3 .OR.
     &        ISCALE .EQ. 5) THEN
         IF (STU .LT. SQRT(SMALL)) THEN
            SCLPAR = 1.0D+00
            GO TO 80
         ELSE
            SCLPAR = STS/STU
         END IF
      ELSE


*     
*     No scaling
*               

         SCLPAR = 1.0D+00
         GO TO 80

      END IF
               

*     
*     Scaling
*               
            
      IF (MN .EQ. 0) THEN
         IF (SCLPAR .LT. 0.01D+00) SCLPAR=0.01D+00
         IF (SCLPAR .GT. 100.0D+00) SCLPAR=100.0D+00


*               
*     Interval scaling
*               

      ELSE IF (ISCALE .EQ. 2) THEN
         IF (SCLPAR .LT. 0.6D+00 .OR. SCLPAR .GT. 6.0D+00) 
     &        SCLPAR = 1.0D+00

      ELSE IF (ISCALE .EQ. 3) THEN
         IF (SCLPAR .LT. 0.5D+00 .OR. SCLPAR .GT. 5.0D+00) 
     &        SCLPAR = 1.0D+00
         
               
*     
*     Preliminary scaling
*     

      ELSE IF (ISCALE .EQ. 4 .OR. ISCALE .EQ. 5) THEN
         SCLPAR = 1.0D+00
               

*     
*     Scaling at every iteration
*              
 
      ELSE

         CONTINUE

      END IF

 80   CONTINUE


*
*     Lower bound
*

      IF (SCLPAR .LE. SQRT(SMALL)) SCLPAR = SQRT(SMALL)

         
      RETURN
      END


************************************************************************
*
*     * SUBROUTINE INDIC1 *
*
*      
*     * Purpose *
*      
*     Initialization of indices.
*     
*      
*     * Calling sequence *
*     
*     CALL INDIC1(MC,MCC,MN,INEW,IOLD,IFLAG,IFLAG2,ITYPE)
*
*      
*     * Parameters *
*
*     II  MC              Declared number of stored corrections.
*     II  MCC             Current number of depositories used.
*     IO  MN              Current size of vectors.
*     II  ITYPE           Type of Initialization:
*                             1  - corrections are stored,
*                             2  - corrections are not stored,
*                             3  - update is skipped.
*     IU  INEW            Index for circular arrays.
*     IO  IOLD            Index of the oldest corrections.
*     IU  IFLAG           Index for adaptive version:
*                             0  - Maximum number of stored corrections
*                                    has not been changed at previous
*                                    iteration.
*                             1  - Maximum number of stored corrections
*                                    has been changed at previous
*                                    iteration.
*     IU  IFLAG2          Index for adaptive version:
*                             0  - IFLAG has not been changed.
*                             1  - IFLAG has been changed.
*     
*      

      SUBROUTINE INDIC1(MC,MCC,MN,INEW,IOLD,IFLAG,IFLAG2,ITYPE)

      
*     Scalar Arguments
      INTEGER MC,MCC,INEW,IFLAG,MN,IOLD,IFLAG2,ITYPE
      

      IF (ITYPE .EQ. 1) THEN
         IF (MCC .LT. MC) THEN
            MN = MCC + 1
            IOLD = 1
            IFLAG = 0

         ELSE
            IF (IFLAG .EQ. 0) THEN
               MN = MC
               IOLD = INEW + 2
               IF (IOLD .GT. MC+1) IOLD = IOLD - MC - 1

            ELSE
               IF (INEW .EQ. 1) THEN
                  INEW = MC + 1
                  MN = MC
                  IOLD = 2
                  IFLAG = 0
                  IFLAG2 = 1

               ELSE IF (INEW .EQ. MC) THEN
                  MN = MC
                  IOLD = 1
                  IFLAG = 0
                  IFLAG2 = 1

               ELSE
                  MN = MC - 1
                  IOLD = INEW + 2
                  IF (IOLD .GT. MC) IOLD = IOLD - MC

               END IF

            END IF

         END IF
      

      ELSE IF (ITYPE .EQ. 2) THEN

         IF (MCC .LT. MC) THEN
            MN = MCC + 1
            IOLD = 1
            IFLAG = 0

         ELSE
            IF (IFLAG .EQ. 0) THEN
               MN = MC + 1
               IOLD = INEW + 1
               IF (IOLD .GT. MC + 1) IOLD = 1

            ELSE
               MN = MC
               IOLD = INEW + 1
               IF (IOLD .GT. MC) IOLD = 1
            END IF
               
         END IF
      

      ELSE 
         
         IF (MCC .LT. MC) THEN
            MN = MCC
            IOLD = 1
            IFLAG = 0

         ELSE
            IF (IFLAG .EQ. 0) THEN
               MN = MC
               IOLD = INEW + 1
               IF (IOLD .GT. MC + 1) IOLD = 1
               
            ELSE
               MN = MC - 1
               IOLD = INEW + 1
               IF (IOLD .GT. MC) IOLD = 1
            END IF
            
         END IF

      END IF
      

      RETURN
      END



************************************************************************
*
*     * SUBROUTINE ACTVAR *
*
*      
*     * Purpose *
*      
*     Finding the index set of free and active variables at the
*     generalized Cauchy point.
*
*      
*     * Calling sequence *
*     
*     CALL ACTVAR(N,NACT,IACT,IB,XCP,XL,XU,SMALL)
*
*      
*     * Parameters *
*     
*     II  N               Number of variables.
*     IO  NACT            Number of active variables.
*     IO  IACT(N)         Index set of active and free variables:
*                           for I=1,...,NACT, IACT(I) are the indices
*                             of active variables (-J, if
*                             XCP(J)=XL(J) and +J, otherwise)
*                           for i=NACT+1,...,N, IACT(I) are the indices
*                             of free variables.
*     II  IB(N)           Type of bound constraints:
*                           0  - X(I) is unbounded,
*                           1  - X(I) has only a lower bound,
*                           2  - X(I) has both lower and upper bounds,
*                           3  - X(I) has only an upper bound. 
*     RI  XCP(N)          Generalized Cauchy point.
*     RI  XL(N)           Lower bounds for variables. 
*     RI  XU(N)           Upper bounds for variables. 
*     RI  SMALL           Small positive value.
*     
*      

      SUBROUTINE ACTVAR(N,NACT,IACT,IB,XCP,XL,XU,SMALL)
      

*     Scalar Arguments
      INTEGER N,NACT
      DOUBLE PRECISION SMALL
      
*     Array Arguments
      INTEGER IB(*),IACT(*)
      DOUBLE PRECISION XCP(*),XL(*),XU(*)

*     Local Scalars
      INTEGER I,IFREE
     

      NACT = 0
      IFREE = N+1

      DO 10 I = 1,N
         IF (IB(I) .EQ. 0) THEN
            IFREE = IFREE - 1
            IACT(IFREE) = I

         ELSE IF (IB(I) .LE. 2 .AND. XCP(I) .LE. XL(I)+SMALL) THEN
            NACT = NACT + 1
            IACT(NACT) = -I

         ELSE IF (IB(I) .GE. 2 .AND. XCP(I) .GE. XU(I)-SMALL) THEN
            NACT = NACT + 1
            IACT(NACT) = I

         ELSE
            IFREE = IFREE - 1
            IACT(IFREE) = I

         END IF

 10   CONTINUE
      
      RETURN
      END

      

************************************************************************
*
*     * SUBROUTINE GETIME *
*
*      
*     * Purpose *
*      
*     Execution time.
*     
*      
*     * Calling sequence *
*     
*     CALL GETIME(CTIM,RTIM)
*
*      
*     * Parameters *
*     
*     RO  CTIM          Current time. REAL argument
*     RA  RTIM(2)       Auxiliary array. REAL array.
*     
*     
*     * Subprograms used *
*      
*     RF  ETIME         Execution time.
*      
*

      SUBROUTINE GETIME(CTIM,RTIM)

      
*     Scalar Arguments
      REAL CTIM

*     Array arguments
      REAL RTIM(2)

*     Intrinsic Functions
      REAL ETIME
      INTRINSIC ETIME


      CTIM = ETIME(RTIM)
      CTIM = RTIM(1)

      RETURN
      END


      
************************************************************************
 
